/*
 Copyright (c) 1998 Burton Computer Corporation.
 All rights reserved.
 
 Redistribution and use of this file in source and binary forms,
 with or without modification, are permitted provided that the
 following conditions are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 
 
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 
 THIS SOFTWARE IS PROVIDED BY BURTON COMPUTER CORPORATION ``AS IS'' AND ANY
 EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL BURTON COMPUTER CORPOATION
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.
    
 **********
    
 This file contains the source code for BCC's simple interface to 
 the Unix functions setgid() and setuid().  The functions in this
 file implement the native methods defined in the BccUnix class.
   
 This file has been successfully compiled on Solaris 2.6 SPARC
 and Intel editions and Red Hat Linux 5.1.  The code is meant to
 be simple and portable but may require some modification for
 non-standard Unix platforms..
 
 */
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <assert.h>
#include "BccUnix.h"


static int name2uid(const char *name)
{
   struct passwd *pw = getpwnam(name);
   return (pw == NULL) ? -1 : pw->pw_uid;
}

static int name2gid(const char *name)
{
   struct group *grp = getgrnam(name);
   return (grp == NULL) ? -1 : grp->gr_gid;
}

/*
   We throw a simple Exception when any error is encountered.
   It didn't seem worthwile to define a new exception class
   for this little utility.
 */
static void throw_it(JNIEnv *env,
                     const char *message)
{
   jstring msg = env->NewStringUTF(message);
   jclass _class = env->FindClass("java/lang/Exception");
   jmethodID mid = env->GetMethodID(_class, "<init>", "(Ljava/lang/String;)V");
   jthrowable ex = (jthrowable)env->NewObject(_class, mid, msg);
   env->Throw(ex);
}

/*
   Build error message and throw an exception.
   Called when a system call fails.
 */
static void throw_error(JNIEnv *env,
                        jobject self)
{
   char buffer[4096];
   const char *errmsg = strerror(errno);
   sprintf(buffer, "errno=%d (%s)", errno, (errmsg ? errmsg : "<null>"));
   assert(strlen(buffer) < sizeof(buffer));
   throw_it(env, buffer);
}

/*
   Build error message and throw an exception.
   Called when a group or user name is unknown..
 */
static void throw_not_found(JNIEnv *env,
                            jobject self,
                            const char *type,
                            const char *name)
{
   char buffer[4096];

   if (strlen(name) > 256) {
      name = "<name too long>";
   }
   sprintf(buffer, "unknown %s: %s", type, name);
   assert(strlen(buffer) < sizeof(buffer));
   throw_it(env, buffer);
}

JNIEXPORT void JNICALL Java_BccUnix_setUserID(JNIEnv *env,
                                              jclass self, 
                                              jstring jname)
{
   const char *name = env->GetStringUTFChars(jname, NULL);
   int uid = name2uid(name);
   if (uid < 0) {
      throw_not_found(env, self, "user", name);
   } else if (setuid(uid) < 0) {
      throw_error(env, self);
   }
   env->ReleaseStringUTFChars(jname, name);
}

JNIEXPORT void JNICALL Java_BccUnix_setGroupID(JNIEnv *env,                           
                                               jclass self,
                                               jstring jname)
{
   const char *name = env->GetStringUTFChars(jname, NULL);
   int gid = name2gid(name);
   if (gid < 0) {
      throw_not_found(env, self, "group", name);
   } else if (setgid(gid) < 0) {
      throw_error(env, self);
   }
   env->ReleaseStringUTFChars(jname, name);
}
