///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.apps.clnews.configure;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import javax.swing.border.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.awt.*;
import org.cooldevtools.coffee_server.services.nntp.protocol.ServerConnection;

public class MainConfigPanel
extends JPanel
{
   public MainConfigPanel(ErrorHandler errors)
   {
      m_errors = errors;
      m_server = new ServerConnection();
      initDisplay();
   }

   private void initDisplay()
   {
      setLayout(new BorderLayout());
      
      m_modified = new ModifiedEventHelper();
      m_modified.addActionListener(new Dirty());

      m_tree = new OverviewTreePanel();
      m_buttonPanel = new DetailButtonPanel();
      m_buttonPanel.setBorder(new EmptyBorder(5, 5, 5, 5));
      m_userPanel   = new UserDetailPanel(m_modified);
      m_rolePanel   = new RoleDetailPanel(m_modified);
      m_groupPanel  = new NewsGroupDetailPanel(m_modified);

      m_selectedOptionIndex = -1;
      m_options = new ControlPanelInterface[3];
      m_options[0] = new ServerGeneralDetailPanel(m_modified);
      m_options[1] = new ServerLoggingDetailPanel(m_modified);
      m_options[2] = new ServerUnixDetailPanel(m_modified);
      if (Assert.DEBUG) Assert.pre(OPTION_CARDS.length == m_options.length);
      if (Assert.DEBUG) Assert.pre(OPTION_NAMES.length == m_options.length);

      m_cards = new CardLayout();

      m_detailPanel = new JPanel();
      m_detailPanel.setLayout(m_cards);
      m_detailPanel.setBorder(new EmptyBorder(5, 5, 0, 5));
      
      m_detailPanel.add(ROLE_CARD,  m_rolePanel);
      m_detailPanel.add(USER_CARD,  m_userPanel);
      m_detailPanel.add(GROUP_CARD, m_groupPanel);
      m_detailPanel.add(EMPTY_CARD, new JPanel());
      for (int i = 0; i < m_options.length; i++) {
         m_detailPanel.add(OPTION_CARDS[i], (JPanel)m_options[i]);
      }
      m_cards.show(m_detailPanel, EMPTY_CARD);

      m_tree.addSelectionChangedListener(new Selector());
      m_buttonPanel.addDeleteListener(new Delete());
      m_buttonPanel.addRevertListener(new Revert());
      m_buttonPanel.addApplyListener(new Apply());

      JPanel p = new JPanel();
      p.setLayout(new BorderLayout());
      p.add(m_detailPanel, "Center");
      p.add(m_buttonPanel, "South");

      JSplitPane splitter = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, m_tree, p);
      add(splitter, "Center");
   }

   public void setServer(ServerConnection server)
   {
      m_server = server;
   }

   public void setServerParams(String host,
                               int port,
                               String username,
                               String password)
   throws IOException
   {
      m_server.reset();
      m_server.setHost(host, port);
      m_server.setIdentity(username, password);
      reload();
   }

   private void clear()
   {
      m_modified.setModified(false);
      m_buttonPanel.setEmpty(true);
      m_cards.show(m_detailPanel, EMPTY_CARD);
      m_userName = null;
      m_groupName = null;
   }

   private boolean isPermissionDenied(IOException ex)
   {
      return ex.getMessage().startsWith("480 ") || ex.getMessage().startsWith("502 ");
   }

   /**
      Load the user, roles, and groups tree nodes.
      Gracefully degrade if a permission denied error is hit when
      asking for users or roles.
    */
   public void reload()
   throws IOException
   {
      clear();

      m_modified.setModified(false);

      try {
         m_roleNames = m_server.getRoleNames();
      } catch (IOException ex) {
         if (isPermissionDenied(ex)) {
            m_roleNames = new String[0];
         } else {
            throw ex;
         }
      }

      String[] real_users;
      try {
         real_users = m_server.getUserNames();
      } catch (IOException ex) {
         if (isPermissionDenied(ex)) {
            real_users = new String[0];
         } else {
            throw ex;
         }
      }

      String[] real_roles = m_roleNames;
      String[] roles = new String[real_roles.length + 1];
      System.arraycopy(real_roles, 0, roles, 0, real_roles.length);
      roles[real_roles.length] = NEW_ROLE;

      String[] users = new String[real_users.length + 1];
      System.arraycopy(real_users, 0, users, 0, real_users.length);
      users[real_users.length] = NEW_USER;
      
      String[] real_groups = m_server.getGroupNames();
      String[] groups = new String[real_groups.length + 1];
      System.arraycopy(real_groups, 0, groups, 0, real_groups.length);
      groups[real_groups.length] = NEW_GROUP;
      
      m_dontLoad = true;
      try {
         m_tree.reset(OPTION_NAMES, roles, users, groups);
      } finally {
         m_dontLoad = false;
      }
   }

   private void loadUser(String name)
   {
      try {
         if (name.equals(NEW_USER)) {
            m_userPanel.setUsername("");
            m_userPanel.setPassword("");
            m_userPanel.setEmail("");
            m_userPanel.setRoles(m_roleNames, new String[0]);
            m_buttonPanel.setCanDelete(false);
         } else {
            StringBuffer password = new StringBuffer();
            StringBuffer email = new StringBuffer();
            SuperVector roles = new SuperVector();
            m_server.getUser(name, password, email, roles);
            
            m_userPanel.setUsername(name);
            m_userPanel.setPassword(password.toString());
            m_userPanel.setEmail(email.toString());
            m_userPanel.setRoles(m_roleNames, roles.getStrings());
            m_buttonPanel.setCanDelete(true);
         }
   
         m_modified.setModified(false);
         m_buttonPanel.setEmpty(false);
         m_cards.show(m_detailPanel, USER_CARD);
         m_userName = name;
	 m_roleName = null;
         m_groupName = null;
         
         m_tree.selectUser(name);
      } catch (IOException ex) {
         m_errors.handleIOException(ex);
         clear();
      }
   }


   private void loadRole(String name)
   {
      try {
         if (name.equals(NEW_ROLE)) {
	    m_rolePanel.clear();
            m_buttonPanel.setCanDelete(false);
         } else {
	    m_rolePanel.loadFromServer(m_server, name);
            m_buttonPanel.setCanDelete(true);
         }
   
         m_modified.setModified(false);
         m_buttonPanel.setEmpty(false);
         m_cards.show(m_detailPanel, ROLE_CARD);
         m_roleName = name;
	 m_userName = null;
         m_groupName = null;
         
         m_tree.selectRole(name);
      } catch (IOException ex) {
         m_errors.handleIOException(ex);
         clear();
      }
   }

   private void loadGroup(String name)
   {
      try {
         if (name.equals(NEW_GROUP)) {
            m_groupPanel.setNewsGroupNameEditable(true);
            m_groupPanel.setNewsGroupName("");
            m_groupPanel.setOwnerEmail("");
            m_groupPanel.setDescription("");
            m_buttonPanel.setCanDelete(false);
         } else {
            StringBuffer email = new StringBuffer();
            StringBuffer description = new StringBuffer();
            m_server.getNewsGroup(name, email, description);
            
            m_groupPanel.setNewsGroupNameEditable(false);
            m_groupPanel.setNewsGroupName(name);
            m_groupPanel.setDescription(description.toString());
            m_groupPanel.setOwnerEmail(email.toString());
            m_buttonPanel.setCanDelete(true);
         }
         m_modified.setModified(false);
         m_buttonPanel.setEmpty(false);
         
         m_cards.show(m_detailPanel, GROUP_CARD);
         m_groupName = name;
         m_roleName = null;
         m_userName = null;
         
         m_tree.selectGroup(name);
      } catch (IOException ex) {
         m_errors.handleIOException(ex);
         clear();
      }
   }

   private int findOptionIndex(String name)
   {
      int index = -1;
      for (int i = 0; i < OPTION_NAMES.length; i++) {
         if (OPTION_NAMES[i].equals(name)) {
            index = i;
            break;
         }
      }
      return index;
   }

   private void loadOption(String name)
   {
      int index = findOptionIndex(name);
      if (index < 0) {
         return;
      }
      loadOption(index);
   }

   private void loadOption(int index)
   {
      m_selectedOptionIndex = index;
      m_userName = null;
      m_groupName = null;
      try {
         m_options[index].loadFromServer(m_server);
         m_cards.show(m_detailPanel, OPTION_CARDS[index]);
         m_modified.setModified(false);
         m_buttonPanel.setCanDelete(false);
         m_buttonPanel.setEmpty(false);
      } catch (IOException ex) {
         m_errors.handleIOException(ex);
         clear();
      }
   }

   private class Selector extends OverviewTreePanel.SelectionChangedListener
   {
      public void optionSelected(String name)
      {
         if (!m_dontLoad) {
            loadOption(name);
         }
      }

      public void groupSelected(String name)
      {
         if (!m_dontLoad) {
            loadGroup(name);
         }
      }
   
      public void userSelected(String name)
      {
         if (!m_dontLoad) {
            loadUser(name);
         }
      }
                 
      public void roleSelected(String name)
      {
         if (!m_dontLoad) {
            loadRole(name);
         }
      }
                 
      public boolean canChangeSelection()
      {
         return !m_modified.isModified();
      }
   }

   private class Dirty implements ActionListener
   {
      public void actionPerformed(ActionEvent ev)
      {
         m_buttonPanel.setDirty(true);
      }
   }

   private class Revert implements ActionListener
   {
      public void actionPerformed(ActionEvent ev)
      {
         if (m_userName != null) {
            loadUser(m_userName);
         } else if (m_roleName != null) {
            loadRole(m_roleName);
         } else if (m_groupName != null) {
            loadGroup(m_groupName);
         } else if (m_selectedOptionIndex >= 0) {
            loadOption(m_selectedOptionIndex);
         }
      }
   }

   private void defineUser()
   throws IOException
   {
      String name = m_userPanel.getUsername();
      if (name.length() == 0) {
         return;
      }

      String password = m_userPanel.getPassword();
      String email = m_userPanel.getEmail();
      String[] roles = m_userPanel.getRoles();
      if (name.equals(m_userName)) {
         m_server.defineUser(name, password, email, roles);
         loadUser(name);
      } else {
         m_server.defineUser(name, password, email, roles);
         if (!m_userName.equals(NEW_USER)) {
            m_server.deleteUser(m_userName);
         }
         reload();
         loadUser(name);
      }
   }

   private void defineRole()
   throws IOException
   {
      String name = m_rolePanel.getName();
      if (name.length() == 0) {
         return;
      }

      if (name.equals(m_roleName)) {
	 m_rolePanel.sendToServer(m_server);
	 loadRole(name);
      } else {
	 m_rolePanel.sendToServer(m_server);
	 if (!m_roleName.equals(NEW_ROLE)) {
	    m_server.deleteRole(m_roleName);
	 }
	 reload();
	 loadRole(name);
      }
   }

   private void defineGroup()
   throws IOException
   {
      String name  = m_groupPanel.getNewsGroupName();
      if (name.length() == 0) {
         return;
      }

      String email = m_groupPanel.getOwnerEmail();
      String desc  = m_groupPanel.getDescription();
      m_server.defineNewsGroup(name, email, desc);
      reload();
      loadGroup(name);
   }

   private void saveConfig()
   throws IOException
   {
      m_options[m_selectedOptionIndex].sendToServer(m_server);
      m_modified.setModified(false);
      m_buttonPanel.setEmpty(false);
   }

   private class Apply implements ActionListener
   {
      public void actionPerformed(ActionEvent ev)
      {
         try {
            if (m_userName != null) {
               defineUser();
            } else if (m_groupName != null) {
               defineGroup();
            } else if (m_roleName != null) {
               defineRole();
            } else if (m_selectedOptionIndex >= 0) {
               saveConfig();
            }
         } catch (IOException ex) {
            m_errors.handleIOException(ex);
            clear();
         }
      }
   }

   private boolean confirm(String title,
                           String message)
   {
      int result = JOptionPane.showConfirmDialog(MainConfigPanel.this, 
                                                 message,
                                                 title,
                                                 JOptionPane.YES_NO_OPTION);
      return result == JOptionPane.YES_OPTION;
   }

   private class Delete implements ActionListener
   {
      public void actionPerformed(ActionEvent ev)
      {
         try {
            if (m_userName != null) {
               if (!m_userName.equals(NEW_USER)) {
                  if (confirm("Delete User?", "Permanently delete user " + m_userName + "?")) {
                     m_server.deleteUser(m_userName);
                     reload();
                  }
               }
            } else if (m_groupName != null) {
               if (!m_groupName.equals(NEW_GROUP)) {
                  if (confirm("Delete Group?", "Permanently delete group " + m_groupName + "?")) {
                     m_server.deleteNewsGroup(m_groupName);
                     reload();
                  }
               }
            } else if (m_roleName != null) {
               if (!m_roleName.equals(NEW_ROLE)) {
                  if (confirm("Delete Role?", "Permanently delete role " + m_roleName + "?")) {
                     m_server.deleteRole(m_roleName);
                     reload();
                  }
               }
            }
         } catch (IOException ex) {
            m_errors.handleIOException(ex);
            clear();
         }
      }
   }

   public ServerConnection getServer()
   {
      return m_server;
   }

   public static void main(String[] args)
   throws IOException
   {
      MainConfigPanel m = new MainConfigPanel(null);
      m.setServerParams("einstein", 5250, "admin", "admin");

      JFrame f = new JFrame();
      f.getContentPane().add(m);
      f.pack();
      f.show();
   }

   private ErrorHandler         m_errors;
   private OverviewTreePanel    m_tree;
   private DetailButtonPanel    m_buttonPanel;
   private UserDetailPanel      m_userPanel;
   private RoleDetailPanel      m_rolePanel;
   private NewsGroupDetailPanel m_groupPanel;
   private CardLayout           m_cards;
   private JPanel               m_detailPanel;
   private ModifiedEventHelper  m_modified;
   private ServerConnection m_server;
   private boolean m_dontLoad;
   private int m_selectedOptionIndex;
   private ControlPanelInterface[] m_options;

   private String m_roleName;
   private String m_userName;
   private String m_groupName;
   private String[] m_roleNames;

   private static String[] OPTION_NAMES = {
      "General", "Logging", "Unix Only"
   };
   private static String[] OPTION_CARDS = OPTION_NAMES;
   private static String ROLE_CARD  = "role";
   private static String USER_CARD  = "user";
   private static String GROUP_CARD = "group";
   private static String EMPTY_CARD = "empty";
   private static String NEW_ROLE   = "<New Role>";
   private static String NEW_USER   = "<New User>";
   private static String NEW_GROUP  = "<New Group>";
}

