///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.apps.clnews.repository;

import java.io.*;
import java.text.*;
import org.cooldevtools.coffee_server.services.nntp.pub.*;
import org.cooldevtools.coffee_server.framework.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.utility.logging.*;

public class BackupThread extends Thread
{
   public BackupThread(FileArchive archive,
                       String directory_name,
                       String file_base_name,
		       FileMessageRepository message_repository)
   {
      m_archive = archive;
      File dir = new File(directory_name);
      m_directoryName = dir.getName();
      m_rootDirectoryName = dir.getParent();
      final String fileNamePattern = file_base_name + "{0,number,integer}.zip";
      m_fileSequence = new FileSequence(fileNamePattern, NUM_FILES);
      m_messageRepository = message_repository;
      setDaemon(true);
   }

   public void run()
   {
      long next_run = System.currentTimeMillis() + IDLE_SLEEP_TIME;
      int num_exceptions = 0;
      while (true) {
         try {
	    long sleep_time = next_run - System.currentTimeMillis();
	    sleep_time = sleep_time > MIN_IDLE_SLEEP_TIME ? sleep_time : MIN_IDLE_SLEEP_TIME;
	    sleep(sleep_time);
	    next_run = System.currentTimeMillis() + IDLE_SLEEP_TIME;

            makeBackup();
	    expireArticles();
            if (num_exceptions > 0) {
               num_exceptions--;
            }
         } catch (Exception ex) {
            Logger.s().logSilently(MessagePriority.LOG_WARNING,
				   "backup thread caught exception: {0}",
				   new Object[] { ex.toString() });
            if (num_exceptions++ > MAX_EXCEPTIONS) {
               Logger.s().logSilently(MessagePriority.LOG_EMERGENCY,
				      "backup thread caught too many exceptions, shutting down");
               break;
            }
         }
      }
   }

   private int expireSomeArticles(String max_date,
				  int max_expired)
      throws CoffeeServiceException
   {
      boolean commit = false;
      m_archive.acquireWriteLock();
      try {
	 FileMessage[] messages = m_archive.getDBMessages().loadMessagesOlderThan(max_date, max_expired);
	 for (int i = 0; i < messages.length; i++) {
	    Logger.s().logSilently(MessagePriority.LOG_INFO,
				   "removing old article {0}",
				   new Object[] { messages[i].messageId() });
	    m_messageRepository.removeOneMessage(messages[i]);
	 }
	 commit = true;
	 return messages.length;
      } finally {
	 m_archive.releaseWriteLock(commit);
      }
   }

   private void expireArticles()
      throws CoffeeServiceException, InterruptedException
   {
      int max_days = m_archive.getMaxMessageDays();
      if (max_days == 0) {
	 return;
      }
      String max_date = DateUtils.nDaysAgoString(max_days + 1);

      int expired_count = 0;
      for (int i = 0; i < MAX_EXPIRED_LOOPS; i++) {
	 int loop_count = expireSomeArticles(max_date, MAX_EXPIRED_PER_LOOP);
	 if (loop_count == 0) {
	    break;
	 }
	 expired_count += loop_count;
	 sleep(EXPIRE_SLEEP_TIME);
      }

      Logger.s().logSilently(MessagePriority.LOG_INFO,
			     "removed {0} old articles",
			     new Object[] { new Integer(expired_count) });
   }

   private void makeBackup()
      throws IOException, CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         m_fileSequence.shift();
         String filename = m_fileSequence.getFile().getPath();
         Logger.s().logSilently(MessagePriority.LOG_INFO,
				"writing backup file {0}",
				new Object[] { filename });
         Zipper.zip(filename, m_rootDirectoryName, m_directoryName);
         Logger.s().logSilently(MessagePriority.LOG_NOTICE,
				"backup file {0} complete",
				new Object[] { filename });
      } finally {
         m_archive.releaseReadLock();
      }
   }

   private FileArchive m_archive;
   private String m_rootDirectoryName;
   private String m_directoryName;
   private FileSequence m_fileSequence;
   private FileMessageRepository m_messageRepository;

   private static final int  NUM_FILES            = 3;
   private static final int  MAX_EXCEPTIONS       = 20;
   private static final int  MAX_EXPIRED_PER_LOOP = 20;
   private static final int  MAX_EXPIRED_LOOPS    = 150;
   private static final long IDLE_SLEEP_TIME      = 8 * 3600 * 1000; // 8 hours
   private static final long MIN_IDLE_SLEEP_TIME  = IDLE_SLEEP_TIME / 2;
   private static final long EXPIRE_SLEEP_TIME    = 2000; // 2 seconds
}
