///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.apps.clnews.repository;

import java.io.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.utility.storage.*;
import org.cooldevtools.coffee_server.framework.*;
import org.cooldevtools.coffee_server.services.nntp.pub.*;

public class DBGroups
{
   public DBGroups(String directory,
		   FileArchive archive,
		   Schema schema,
		   CommitManager commit_manager)
   throws IOException, StorageException
   {
      m_db = new SimpleDatabase(schema, directory, DB_NAME);
      commit_manager.addCommitable(m_db);
      m_archive  = archive;
   }

   public static class GroupCounters {
      GroupCounters(Record r)
      throws StorageException
      {
         nextIndex = r.getLong(COL_NextNum);
         lowIndex  = r.getLong(COL_LowNum);
         highIndex = r.getLong(COL_HighNum);
         count     = r.getLong(COL_MsgCount);
      }

      void store(Record r)
      throws StorageException
      {
         r.setLong(COL_NextNum,  nextIndex);
         r.setLong(COL_LowNum,   lowIndex);
         r.setLong(COL_HighNum,  highIndex);
         r.setLong(COL_MsgCount, count);
      }

      public long nextIndex;
      public long lowIndex;
      public long highIndex;
      public long count;
   }

   private FileNewsGroup loadGroup(Record r)
   throws IOException, StorageException, CoffeeServiceException
   {
      return new FileNewsGroup(r.getString(COL_Name),
                               r.getLong(COL_DBKey),
                               DateUtils.stringToDate(r.getString(COL_Date)),
                               m_archive);
   }

   public GroupCounters loadGroupCounters(long group_key)
   throws CoffeeServiceException
   {
      try {
         Record r = findGroupRecord(group_key);
         return new GroupCounters(r);
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public String loadGroupDescription(long group_key)
   throws CoffeeServiceException
   {
      try {
         Record r = findGroupRecord(group_key);
         return r.getString(COL_Description);
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public String loadGroupEmailAddress(long group_key)
   throws CoffeeServiceException
   {
      try {
         Record r = findGroupRecord(group_key);
         return r.getString(COL_Email);
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public void storeGroupCounters(GroupCounters counters,
                                  long group_key)
   throws CoffeeServiceException
   {
      try {
         Record r = findGroupRecord(group_key);
         counters.store(r);
         m_db.update(r);
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public FileNewsGroup loadGroup(String name)
   throws CoffeeServiceException
   {
      try {
         Record r = findGroupRecord(name);
         return loadGroup(r);
      } catch (UnknownKeyException ex) {
         return null;
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public FileNewsGroup loadGroup(long key)
   throws CoffeeServiceException
   {
      try {
         Record r = findGroupRecord(key);
         return loadGroup(r);
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   private Record findGroupRecord(String name)
   throws IOException, StorageException, CoffeeServiceException
   {
      Record key = m_db.getBlank();
      key.setString(COL_Name, name);

      Record r = m_db.find(NDX_ByName, key);
      return r;
   }

   private Record findGroupRecord(long db_key)
   throws IOException, StorageException, CoffeeServiceException
   {
      Record key = m_db.getBlank();
      key.setLong(COL_DBKey, db_key);

      Record r = m_db.find(NDX_ByKey, key);
      return r;
   }

   public FileNewsGroup createGroup(User creator,
                                    String name,
                                    String email,
                                    String description)
   throws CoffeeServiceException
   {
      try {
         Record r = m_db.getBlank();
         r.setString(COL_Name, name);
         r.setString(COL_Email, email);
         r.setString(COL_Description, description);
         r.setString(COL_Date, DateUtils.now());
         r.setLong(COL_NextNum, 1);
         r.setLong(COL_LowNum,  0);
         r.setLong(COL_HighNum, 0);
         r.setLong(COL_MsgCount, 0);
         r.setLong(COL_DBKey, m_archive.nextInSequence(SEQ_NAME));
         m_db.insert(r);
         return loadGroup(r);
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public void changeGroupAttributes(long db_key,
                                     String email,
                                     String description)
   throws CoffeeServiceException
   {
      try {
         Record r = findGroupRecord(db_key);
         if (Assert.DEBUG) Assert.pre(r != null);
         r.setString(COL_Email, email);
         r.setString(COL_Description, description);
         m_db.update(r);
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public void removeGroup(FileNewsGroup g)
   throws CoffeeServiceException
   {
      try {
         Record r = findGroupRecord(g.getKey());
         if (Assert.DEBUG) Assert.pre(r != null);
         m_db.remove(r);
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public FileNewsGroup[] loadGroups()
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setString(COL_Name, "");

         SuperVector vec = new SuperVector(1000,0);
         DatabaseNavigator nav = m_db.navigate(NDX_ByName, key);
         try {
            while (nav.isValid()) {
               vec.addElement(loadGroup(nav.getRecord()));
               nav.next();
            }
            return FileNewsGroup.toArray(vec);
         } finally {
            nav.close();
         }
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public void sync()
   throws IOException, StorageException
   {
      m_db.sync();
   }

   public void abort()
   {
      m_db.abort();
   }

   private FileArchive m_archive;
   private Database    m_db;

   private static final String COL_DBKey       = "DBKey";
   private static final String COL_Name        = "Name";
   private static final String COL_Email       = "Email";
   private static final String COL_Description = "Desc";
   private static final String COL_Date        = "Date";
   private static final String COL_NextNum     = "NextNum";
   private static final String COL_LowNum      = "LowNum";
   private static final String COL_HighNum     = "HighNum";
   private static final String COL_MsgCount    = "MsgCount";

   private static final String NDX_ByName      = "ByName";
   private static final String NDX_ByDate      = "ByDate";
   private static final String NDX_ByKey       = "ByKey";

   private static final String DB_NAME         = "groups";
   private static final String SEQ_NAME        = DB_NAME;
}
