///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.apps.clnews.repository;

import java.io.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.utility.storage.*;
import org.cooldevtools.coffee_server.framework.*;
import org.cooldevtools.coffee_server.services.nntp.pub.*;

public class DBMessages
{
   public DBMessages(String directory,
		     FileArchive archive,
		     Schema schema,
		     CommitManager commit_manager)
   throws IOException, StorageException
   {
      m_db = new SimpleDatabase(schema, directory, DB_NAME);
      commit_manager.addCommitable(m_db);
      m_archive = archive;
   }

   public void insertMessage(FileMessage m)
   throws CoffeeServiceException
   {
      try {
         if (Assert.DEBUG) Assert.pre(!messageExists(m.messageId()));
         Record r = m_db.getBlank();
         r.setString(COL_MessageID, m.messageId());
         r.setLong(COL_MessageNum,   m.getNumber());
         r.setLong(COL_RefCount,     0);
         r.setString(COL_Date,      DateUtils.dateToString(m.getDate()));
         r.setBlob(COL_Header,      m.getHeaderAsString());
         r.setBlob(COL_Body,        m.getBodyAsString());
         r.setBlob(COL_Overview,    m.messageOverview());
         m_db.insert(r);
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public boolean messageExists(String message_id)
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setString(COL_MessageID, message_id);
         return m_db.exists(NDX_ById, key);
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public FileMessage loadMessage(String message_id)
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setString(COL_MessageID, message_id);
         try {
            Record r = m_db.find(NDX_ById, key);
            return readMessageFromRecord(r);
         } catch (UnknownKeyException ex) {
            return null;
         }
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public FileMessage loadMessage(Long message_num)
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setLong(COL_MessageNum, message_num.longValue());
         try {
            Record r = m_db.find(NDX_ByNumber, key);
            return readMessageFromRecord(r);
         } catch (UnknownKeyException ex) {
            return null;
         }
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public void removeMessageRef(long message_num)
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setLong(COL_MessageNum, message_num);
         try {
            Record r = m_db.find(NDX_ByNumber, key);
            long new_ref = r.getLong(COL_RefCount) - 1;
            r.setLong(COL_RefCount, new_ref);
            if (Assert.DEBUG) Assert.pre(new_ref >= 0);
            if (new_ref <= 0) {
               m_db.remove(r);
            } else {
               m_db.update(r);
            }
         } catch (UnknownKeyException ex) {
            // shouldn't happen, but safe to ignore it
            if (Assert.DEBUG) Assert.pre(false);
         }
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public void addMessageRef(long message_num)
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setLong(COL_MessageNum, message_num);
         try {
            Record r = m_db.find(NDX_ByNumber, key);
            long new_ref = r.getLong(COL_RefCount) + 1;
            if (Assert.DEBUG) Assert.pre(new_ref >= 1);
            r.setLong(COL_RefCount, new_ref);
            m_db.update(r);
         } catch (UnknownKeyException ex) {
            // shouldn't happen, but safe to ignore it
            if (Assert.DEBUG) Assert.pre(false);
         }
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public String loadMessageOverview(Long message_num)
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setLong(COL_MessageNum, message_num.longValue());
         try {
            Record r = m_db.find(NDX_ByNumber, key);
            return r.getBlobAsString(COL_Overview);
         } catch (UnknownKeyException ex) {
            return null;
         }
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public void removeMessage(FileMessage m)
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setLong(COL_MessageNum, m.getNumber());
         try {
            Record r = m_db.find(NDX_ByNumber, key);
            m_db.remove(r);
         } catch (UnknownKeyException ex) {
            // shouldn't happen, but safe to ignore it
            if (Assert.DEBUG) Assert.pre(false);
         }
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   private String loadMessageBlob(long message_num,
                                  String column_name)
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setLong(COL_MessageNum, message_num);
         Record r = m_db.find(NDX_ByNumber, key);
         return r.getBlobAsString(column_name);
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   String loadMessageHeader(long message_num)
   throws CoffeeServiceException
   {
      return loadMessageBlob(message_num, COL_Header);
   }

   String loadMessageBody(long message_num)
   throws CoffeeServiceException
   {
      return loadMessageBlob(message_num, COL_Body);
   }

   private FileMessage readMessageFromRecord(Record r)
   throws IOException, StorageException
   {
      FileMessage m = new FileMessage(m_archive, r.getLong(COL_MessageNum));
      m.setMessageID(r.getString(COL_MessageID));
      return m;
   }

   public FileMessage[] loadMessagesNewerThan(String date)
   throws CoffeeServiceException
   {
      return loadMessagesByDate(date, DateUtils.INFINITE_DATE_STRING, Integer.MAX_VALUE);
   }

   public FileMessage[] loadMessagesOlderThan(String date,
					      int max_to_return)
   throws CoffeeServiceException
   {
      return loadMessagesByDate(DateUtils.ZERO_DATE_STRING, date, max_to_return);
   }

   public FileMessage[] loadMessagesByDate(String min_date,
					   String max_date,
					   int max_to_return)
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setString(COL_Date, min_date);
         DatabaseNavigator nav = m_db.navigate(NDX_ByDate, key);
         try {
            long counter = 0;
            SuperVector vec = new SuperVector(max_to_return == Integer.MAX_VALUE ? 500 : max_to_return);
            while (nav.isValid() && counter < max_to_return) {
               Record r = nav.getRecord();
	       String msg_date = r.getString(COL_Date);
	       if (max_date.compareTo(msg_date) <= 0) {
		  break;
	       }

               FileMessage m = readMessageFromRecord(r);
               vec.addElement(m);
               nav.next();

	       counter++;
               if ((counter % 300) == 0) {
                  Thread.yield();
               }
            }
            return FileMessage.toArray(vec);
         } finally {
            nav.close();
         }
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public void sync()
   throws IOException, StorageException
   {
      m_db.sync();
   }

   public void abort()
   {
      m_db.abort();
   }

   private Database m_db;
   private FileArchive m_archive;

   private static final String COL_MessageID  = "MessageID";
   private static final String COL_MessageNum = "MessageNum";
   private static final String COL_RefCount   = "RefCount";
   private static final String COL_Date       = "Date";
   private static final String COL_Header     = "Header";
   private static final String COL_Body       = "Body";
   private static final String COL_Overview   = "Overview";

   private static final String NDX_ById       = "ById";
   private static final String NDX_ByDate     = "ByDate";
   private static final String NDX_ByNumber   = "ByNumber";

   private static final String DB_NAME        = "messages";
}
