///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.apps.clnews.repository;

import java.io.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.utility.storage.*;
import org.cooldevtools.coffee_server.framework.*;
import org.cooldevtools.coffee_server.services.nntp.pub.*;

public class DBPosts
{
   public DBPosts(String directory,
		  FileArchive archive,
		  Schema schema,
		  CommitManager commit_manager,
		  DBMessages messages)
   throws IOException, StorageException
   {
      if (Assert.DEBUG) Assert.pre(messages != null);

      m_db = new SimpleDatabase(schema, directory, DB_NAME);
      commit_manager.addCommitable(m_db);
      m_archive  = archive;
      m_messages = messages;
   }

   public void postMessage(FileMessage m,
                           FileNewsGroup g,
                           String date,
                           long index)
   throws CoffeeServiceException
   {
      try {
         Record r = m_db.getBlank();
         r.setString(COL_Date, date);
         r.setLong(COL_GroupKey, g.getKey());
         r.setLong(COL_Number, index);
         r.setLong(COL_MessageKey, m.getNumber());
         m_db.insert(r);
         m_messages.addMessageRef(m.getNumber());
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public Long[] cancelMessage(FileMessage m)
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setLong(COL_MessageKey, m.getNumber());

         Record r = null;
         SuperHashtable hash = new SuperHashtable();
         SuperVector vec = new SuperVector();
         DatabaseNavigator nav = m_db.navigate(NDX_ById, key);
         try {
            while (nav.isValid()) {
               r = nav.getRecord();
               if (r.getLong(COL_MessageKey) != m.getNumber()) {
                  break;
               }

               vec.addElement(r);
               Long group_key = new Long(r.getLong(COL_GroupKey));
               hash.put(group_key, group_key);
               nav.next();
            }
         } finally {
            nav.close();
         }

         Record[] recs = Record.toArray(vec);
         for (int i = 0; i < recs.length; i++) {
            m_db.remove(recs[i]);
         }

         return hash.getLongs();
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public int countPostsForGroup(FileNewsGroup g)
   throws CoffeeServiceException
   {
      try {
         return getSomePostsForGroup(g, 0, Long.MAX_VALUE).length;
      } catch (Exception ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public int listGroup(FileNewsGroup g,
                        long start_index,
                        long[] indexes,
                        long max_to_return)
   throws CoffeeServiceException
   {
      if (Assert.DEBUG) Assert.pre(indexes.length >= max_to_return);
      try {
         Record[] recs = getSomePostsForGroup(g, start_index, max_to_return);
         for (int i = 0; i < recs.length; i++) {
            indexes[i] = recs[i].getLong(COL_Number);
         }
         return recs.length;
      } catch (Exception ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   private Record[] getSomePostsForGroup(FileNewsGroup g,
                                         long start_index,
                                         long max_posts_to_return)
   throws IOException, StorageException
   {
      Record key = m_db.getBlank();
      key.setLong(COL_GroupKey, g.getKey());
      key.setLong(COL_Number, start_index);

      long num_found = 0;
      SuperVector vec = new SuperVector(10000,0);
      DatabaseNavigator nav = m_db.navigate(NDX_ByNumber, key);
      try {
         while (nav.isValid() && (num_found < max_posts_to_return)) {
            Record r = nav.getRecord();
            if (r.getLong(COL_GroupKey) != g.getKey()) {
               break;
            }
            vec.addElement(r);
            nav.next();
            num_found++;
         }

         return Record.toArray(vec);
      } finally {
         nav.close();
      }
   }

   private void removePosts(Record[] recs)
   throws IOException, StorageException, CoffeeServiceException
   {
      for (int i = 0; i < recs.length; i++) {
         m_messages.removeMessageRef(recs[i].getLong(COL_MessageKey));
         m_db.remove(recs[i]);
      }
   }

   public long removePostsForGroup(FileNewsGroup g)
   throws CoffeeServiceException
   {
      return removeSomePostsForGroup(g, Long.MAX_VALUE);
   }

   public long removeSomePostsForGroup(FileNewsGroup g,
                                       long max_to_remove)
   throws CoffeeServiceException
   {
      try {
         Record[] recs = getSomePostsForGroup(g, 0, max_to_remove);
         removePosts(recs);
         return recs.length;
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public Long[] getNewsGroupsForMessage(FileMessage m)
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setLong(COL_MessageKey, m.getNumber());

         SuperVector vec = new SuperVector();
         DatabaseNavigator nav = m_db.navigate(NDX_ById, key);
         try {
            while (nav.isValid()) {
               Record r = nav.getRecord();
               if (r.getLong(COL_MessageKey) != m.getNumber()) {
                  break;
               }
               vec.addElement(new Long(r.getLong(COL_GroupKey)));
               nav.next();
            }
            return vec.getLongs();
         } finally {
            nav.close();
         }
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public Long getMessageKeyForPost(FileNewsGroup g,
                                    long index)
   throws CoffeeServiceException
   {
      try {
         Record key = m_db.getBlank();
         key.setLong(COL_GroupKey, g.getKey());
         key.setLong(COL_Number, index);

         try {
            Record r = m_db.find(NDX_ByNumber, key);
            return new Long(r.getLong(COL_MessageKey));
         } catch (UnknownKeyException ex) {
            return null;
         }
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public void sync()
   throws IOException, StorageException
   {
      m_db.sync();
   }

   public void abort()
   {
      m_db.abort();
   }

   private FileArchive m_archive;
   private Database    m_db;
   private DBMessages  m_messages;

   private static final String COL_MessageKey = "MessageKey";
   private static final String COL_Number     = "Number";
   private static final String COL_Date       = "Date";
   private static final String COL_GroupKey   = "GroupKey";

   private static final String NDX_ByNumber   = "ByNumber";
   private static final String NDX_ById       = "ById";
   private static final String NDX_ByDate     = "ByDate";

   private static final String DB_NAME        = "posts";
}
