///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.apps.clnews.repository;

import java.io.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.utility.storage.*;
import org.cooldevtools.coffee_server.framework.*;
import org.cooldevtools.coffee_server.services.nntp.pub.*;
import org.cooldevtools.coffee_server.services.nntp.protocol.WildMat;

public class DBUsers
{
   public DBUsers(String directory,
		  FileArchive archive,
		  Schema schema,
		  CommitManager commit_manager)
   throws IOException, StorageException
   {
      m_db = new SimpleDatabase(schema, directory, DB_NAME);
      commit_manager.addCommitable(m_db);
      m_archive  = archive;
   }

   public String[] getRoles(String user_name)
   throws CoffeeServiceException
   {
      try {
         try {
            Record key = m_db.getBlank();
            key.setString(COL_Name, user_name);

            Record r = m_db.find(NDX_ByName, key);
            return Misc.splitString(r.getBlobAsString(COL_Roles), SEP);
         } catch (UnknownKeyException ex) {
            return null;
         }
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   private String getField(String user_name,
                           String field_name)
   throws CoffeeServiceException
   {
      try {
         try {
            Record key = m_db.getBlank();
            key.setString(COL_Name, user_name);

            Record r = m_db.find(NDX_ByName, key);
            return r.getString(field_name);
         } catch (UnknownKeyException ex) {
            return null;
         }
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public String getEmail(String user_name)
   throws CoffeeServiceException
   {
      return getField(user_name, COL_Email);
   }

   public String getPassword(String user_name)
   throws CoffeeServiceException
   {
      return getField(user_name, COL_Password);
   }

   public void deleteUser(String user_name)
   throws CoffeeServiceException
   {
      try {
         try {
            Record key = m_db.getBlank();
            key.setString(COL_Name, user_name);

            Record r = m_db.find(NDX_ByName, key);
            m_db.remove(r);
         } catch (UnknownKeyException ex) {
            // ignored
            
         }
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public String[] getUserDefinition(String user_name)
   throws CoffeeServiceException
   {
      try {
         try {
            Record key = m_db.getBlank();
            key.setString(COL_Name, user_name);

            Record r = m_db.find(NDX_ByName, key);

            SuperVector v = new SuperVector(500,0);

            String field = r.getString(COL_Password);
            if (field.length() > 0) {
               v.addElement("PASSWORD " + field);
            }

            field = r.getString(COL_Email);
            if (field.length() > 0) {
               v.addElement("EMAIL " + field);
            }

            String[] tmp = Misc.splitString(r.getBlobAsString(COL_Roles), SEP);
            for (int i = 0; i < tmp.length; i++) {
               v.addElement("ROLE " + tmp[i]);
            }

            return v.getStrings();
         } catch (UnknownKeyException ex) {
            return null;
         }
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   /**
      Ensure that the admin user always belongs to the admin role.
    */
   private String[] validateUserRoles(String user_name,
                                      String[] roles)
   {
      if (!user_name.equals(FileRole.ADMIN_NAME)) {
         return roles;
      }

      for (int i = 0; i < roles.length; i++) {
         if (roles[i].equals(FileRole.ADMIN_NAME)) {
            return roles;
         }
      }

      SuperVector vec = new SuperVector(roles);
      vec.addElement(FileRole.ADMIN_NAME);
      return vec.getStrings();
   }

   public void defineUser(String user_name,
                          String password,
                          String email,
                          String[] roles)
   throws CoffeeServiceException
   {
      try {
         roles = validateUserRoles(user_name, roles);

         Record r = m_db.getBlank();
         r.setString(COL_Name, user_name);
         r.setLong(COL_DBKey, m_archive.nextInSequence(SEQ_NAME));
         r.setString(COL_Email, email);
         r.setString(COL_Password, password);
         r.setBlob(COL_Roles, Misc.joinStrings(roles, SEP));

         deleteUser(user_name);
         m_db.insert(r);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public String[] getUserNames()
   throws CoffeeServiceException
   {
      try {
         DatabaseNavigator nav = m_db.navigate(NDX_ByName, "");
         try {
            SuperVector v = new SuperVector(10000,0);

            while (nav.isValid()) {
               String key = nav.getKey();
               v.addElement(key.substring(0, key.length() - 1));
               nav.next();
            }

            return v.getStrings();
         } finally {
            nav.close();
         }
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   public void sync()
   throws IOException, StorageException
   {
      m_db.sync();
   }

   public void abort()
   {
      m_db.abort();
   }

   private static char SEP = (char)0;

   private FileArchive m_archive;
   private Database    m_db;

   private static final String COL_Name        = "Name";
   private static final String COL_Email       = "Email";
   private static final String COL_Password    = "Password";
   private static final String COL_DBKey       = "DBKey";
   private static final String COL_Roles       = "Roles";

   private static final String NDX_ByName      = "ByName";
   private static final String NDX_ByKey       = "ByKey";

   private static final String DB_NAME         = "users";
   private static final String SEQ_NAME        = DB_NAME;
}
