///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.apps.clnews.repository;

import java.io.*;
import java.util.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.utility.storage.*;
import org.cooldevtools.coffee_server.framework.*;
import org.cooldevtools.coffee_server.services.nntp.pub.*;

public class FileNewsGroup
implements NewsGroup
{
   FileNewsGroup(FileArchive archive)
   throws CoffeeServiceException
   {
      this("", -1, new Date(), archive);
   }

   public FileNewsGroup(String name,
                        long dbkey,
                        Date created,
                        FileArchive archive)
   throws CoffeeServiceException
   {
      m_name        = name;
      m_dbKey       = dbkey;
      m_archive     = archive;
      m_date        = created;
   }

   public void changeAttributes(String name,
                                String email,
                                String description)
   throws CoffeeServiceException
   {
      boolean commit = false;
      m_archive.acquireWriteLock();
      try {
         m_archive.getDBGroups().changeGroupAttributes(m_dbKey, email, description);
         commit = true;
      } finally {
         m_archive.releaseWriteLock(commit);
      }
   }

   public long getAgeInSeconds()
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         return (new Date().getTime() - m_date.getTime()) / 1000;
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public String getDescription()
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         return m_archive.getDBGroups().loadGroupDescription(m_dbKey);
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public String getEmailAddress()
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         return m_archive.getDBGroups().loadGroupEmailAddress(m_dbKey);
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public boolean isNewerThan(Date when)
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         return (when == null) || m_date.after(when);
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public void postMessage(User user,
                           Message message)
   throws CoffeeServiceException
   {
      boolean commit = false;
      m_archive.acquireWriteLock();
      try {
         DBGroups.GroupCounters c = m_archive.getDBGroups().loadGroupCounters(m_dbKey);

         FileMessage msg = (FileMessage)message;
         m_archive.getDBPosts().postMessage(msg,
                                 this,
                                 DateUtils.now(),
                                 c.nextIndex);

         c.highIndex = c.nextIndex;
         if (c.count == 0) {
            c.lowIndex = c.nextIndex;
         }
         c.count++;
         c.nextIndex++;

         m_archive.getDBGroups().storeGroupCounters(c, m_dbKey);

         commit = true;
      } finally {
         m_archive.releaseWriteLock(commit);
      }
   }

   void messageRemoved()
   throws CoffeeServiceException
   {
      // NOTE: we should already have a write lock (caller acquired it)

      // ensure that low index is still valid
      // shouldn't require more than a few iterations in real world
      DBGroups.GroupCounters c = m_archive.getDBGroups().loadGroupCounters(m_dbKey);

      while (c.lowIndex < c.highIndex) {
         if (isValidMessageIndex(c.lowIndex)) {
            break;
         }
         c.lowIndex++;
      }

      // ensure that high index is still valid
      while (c.highIndex >= c.lowIndex) {
         if (isValidMessageIndex(c.highIndex)) {
            break;
         }
         c.highIndex--;
      }

      c.count--;

      // sanity check that count is reasonable
      if (Assert.DEBUG) Assert.post(m_archive.getDBPosts().countPostsForGroup(this) == c.count);
      if (Assert.DEBUG) Assert.post((c.highIndex - c.lowIndex + 1) >= c.count);
      
      m_archive.getDBGroups().storeGroupCounters(c, m_dbKey);
   }

   public long getLowMessageIndex()
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         DBGroups.GroupCounters c = m_archive.getDBGroups().loadGroupCounters(m_dbKey);
         return c.lowIndex;
      } finally {
         m_archive.releaseReadLock();
      }
   }

   long getNextMessageIndex()
   throws CoffeeServiceException
   {
      DBGroups.GroupCounters c = m_archive.getDBGroups().loadGroupCounters(m_dbKey);
      return c.nextIndex;
   }

   public long getHighMessageIndex()
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         DBGroups.GroupCounters c = m_archive.getDBGroups().loadGroupCounters(m_dbKey);
         return c.highIndex;
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public long getMessageCount()
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         DBGroups.GroupCounters c = m_archive.getDBGroups().loadGroupCounters(m_dbKey);
         return c.count;
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public String getMessageOverview(long index)
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         Long key = m_archive.getDBPosts().getMessageKeyForPost(this, index);
         return (key == null) ? null : m_archive.getDBMessages().loadMessageOverview(key);
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public boolean isValidMessageIndex(long index)
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         Long key = m_archive.getDBPosts().getMessageKeyForPost(this, index);
         return key != null;
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public int listGroup(long start_index,
                        long[] indexes,
                        long max_to_return)
   throws CoffeeServiceException
   {
      if (Assert.DEBUG) Assert.pre(indexes.length >= max_to_return);
      m_archive.acquireReadLock();
      try {
         return m_archive.getDBPosts().listGroup(this, start_index, indexes, max_to_return);
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public Message getMessage(long index)
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         Long key = m_archive.getDBPosts().getMessageKeyForPost(this, index);
         return (key == null) ? null : m_archive.getDBMessages().loadMessage(key);
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public String getName()
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         return m_name;
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public Date getDate()
   {
      return m_date;
   }

   public long getKey()
   {
      return m_dbKey;
   }

   public static FileNewsGroup[] toArray(Vector v)
   {
      synchronized (v) {
         FileNewsGroup[] answer = new FileNewsGroup[v.size()];
         v.copyInto(answer);
         return answer;
      }
   }

   private String m_name;
   private long   m_dbKey;
   private Date   m_date;
   private FileArchive m_archive;
}
