///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.apps.clnews.repository;

import java.io.*;
import java.util.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.utility.storage.*;
import org.cooldevtools.coffee_server.framework.*;
import org.cooldevtools.coffee_server.services.nntp.pub.*;

public class FileNewsGroupRepository
implements NewsGroupRepository
{
   public FileNewsGroupRepository(FileArchive archive)
   throws CoffeeServiceException
   {
      m_archive = archive;
   }

   public NewsGroup createNewsGroup(User user,
                                    String name,
                                    String email,
                                    String description)
   throws CoffeeServiceException
   {
      boolean commit = false;
      m_archive.acquireWriteLock();
      try {
         FileNewsGroup group = m_archive.getDBGroups().createGroup(user, name, email, description);
         commit = true;
         return group;
      } finally {
         m_archive.releaseWriteLock(commit);
      }
   }

   public NewsGroup findNewsGroup(String name)
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         return m_archive.getDBGroups().loadGroup(name);
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public NewsGroup[] getNewsGroups()
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         return m_archive.getDBGroups().loadGroups();
      } finally {
         m_archive.releaseReadLock();
      }
   }

   /**
      Remove some of the posts from a news group.
      Return false iff the number removed was smaller than requested
      which means that the rest of the posts and the group itself can
      be removed in one transaction.
    */
   private boolean removeSomePosts(String name)
   throws CoffeeServiceException
   {
      boolean commit = false;
      m_archive.acquireWriteLock();
      try {
         final long MAX_TO_REMOVE = 10;
         FileNewsGroup group = m_archive.getDBGroups().loadGroup(name);
         long num_removed = m_archive.getDBPosts().removeSomePostsForGroup(group, MAX_TO_REMOVE);
         commit = true;
         return num_removed == MAX_TO_REMOVE;
      } finally {
         m_archive.releaseWriteLock(commit);
      }
   }

   private void removeEntireGroup(String name)
   throws CoffeeServiceException
   {
      boolean commit = false;
      m_archive.acquireWriteLock();
      try {
         FileNewsGroup group = m_archive.getDBGroups().loadGroup(name);
         m_archive.getDBPosts().removePostsForGroup(group);
         m_archive.getDBGroups().removeGroup(group);
         commit = true;
      } finally {
         m_archive.releaseWriteLock(commit);
      }
   }

   public void removeNewsGroup(User user,
                               String name)
   throws CoffeeServiceException
   {
      while (removeSomePosts(name)) {
         // remove all/most of the messages
         
      }
      // now remove remaining messages and the group itself
      removeEntireGroup(name);
   }

   public NewsGroup[] getNewsGroupsNewerThan(Date when)
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         SuperVector vec = new SuperVector(1000,0);
         FileNewsGroup[] groups = m_archive.getDBGroups().loadGroups();
         for (int i = 0; i < groups.length; i++) {
            if (groups[i].isNewerThan(when)) {
               vec.addElement(groups[i]);
            }
         }

         return FileNewsGroup.toArray(vec);
      } finally {
         m_archive.releaseReadLock();
      }
   }

   private FileArchive m_archive;
}
