///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.apps.clnews.repository;

import java.io.*;
import java.util.*;
import java.text.*;
import org.cooldevtools.hdl.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.utility.storage.*;
import org.cooldevtools.coffee_server.config.*;
import org.cooldevtools.coffee_server.framework.*;
import org.cooldevtools.coffee_server.util.*;
import org.cooldevtools.coffee_server.services.nntp.pub.*;
import org.cooldevtools.apps.clnews.configure.Constants;

public class FileNewsRepositoryFactory implements NewsRepositoryFactory, FileArchive
{
   public FileNewsRepositoryFactory(ServiceConfigInfo config,
                                    String db_directory,
                                    Statement schema_hdl)
   throws CoffeeServiceException
   {
      if (Assert.DEBUG) Assert.pre(config != null);
      if (Assert.DEBUG) Assert.pre(db_directory != null);
      if (Assert.DEBUG) Assert.pre(schema_hdl != null);
      if (Assert.DEBUG) Assert.pre(new File(db_directory).exists());

      m_config      = config;
      m_dbDirectory = db_directory;
      m_schemaHDL   = schema_hdl;
      m_lock        = new RWLock();
      m_lockFile    = new LockFile(m_dbDirectory, LOCK_FILE_NAME);

      reset();
   }

   public int getMaxMessageSize()
   {
      Token msg_size_prop = m_config.getProperty(NewsGroup.PROP_MAX_MESSAGE_SIZE);
      int max_msg_size    = msg_size_prop == null ? 60 : msg_size_prop.intValue();
      return max_msg_size * 1024;
   }

   public int getMaxMessageDays()
   {
      Token msg_days_prop = m_config.getProperty(NewsGroup.PROP_MAX_MESSAGE_DAYS);
      int max_msg_days    = msg_days_prop == null ? 0 : msg_days_prop.intValue();
      return max_msg_days;
   }

   public synchronized NewsGroupRepository getNewsGroupRepository()
   {
      return m_groups;
   }

   public synchronized MessageRepository getMessageRepository()
   {
      return m_messages;
   }

   public synchronized UserRepository getUserRepository()
   {
      return m_users;
   }

   public synchronized RoleRepository getRoleRepository()
   {
      return m_roles;
   }

   private void createLockFile()
   throws CoffeeServiceException
   {
      try {
	 if (!m_lockFile.create(LOCK_FILE_TEXT)) {
	    String msg = MessageFormat.format(LOCK_EXISTS_ERROR_MESSAGE,
					      new Object[] { m_lockFile.getPath() });
	    throw new CoffeeServiceException(msg);
	 }
      } catch (IOException ex) {
         String msg = MessageFormat.format(LOCK_WRITE_ERROR_MESSAGE,
					   new Object[] { m_lockFile.getPath() });
         throw new CoffeeServiceException(msg);
      }
   }

   private void deleteLockFile()
   {
      m_lockFile.delete();
   }

   private synchronized void openDB()
   throws CoffeeServiceException
   {
      if (Assert.DEBUG) Assert.pre(m_commitManager == null);
      try {
         createLockFile();
         HdlSchema schema = new HdlSchema(m_schemaHDL);
         m_commitManager  = new CommitManager(m_dbDirectory, COMMIT_LOG_NAME);
         m_sequences      = new Sequence(m_dbDirectory, SEQUENCE_DB_NAME);
         m_dbGroups       = new DBGroups(m_dbDirectory, this, schema, m_commitManager);
         m_dbMessages     = new DBMessages(m_dbDirectory, this, schema, m_commitManager);
         m_dbRoles        = new DBRoles(m_dbDirectory, this, schema, m_commitManager);
         m_dbUsers        = new DBUsers(m_dbDirectory, this, schema, m_commitManager);
         m_dbPosts        = new DBPosts(m_dbDirectory, this, schema, m_commitManager, m_dbMessages);
         m_commitManager.addCommitable(m_sequences);
         m_commitManager.initialize();
      } catch (HDLException ex) {
         throw new CoffeeServiceException(ex);
      } catch (StorageException ex) {
         throw new CoffeeServiceException(ex);
      } catch (IOException ex) {
         throw new CoffeeServiceException(ex);
      }
   }

   private synchronized void reset()
   throws CoffeeServiceException
   {
      if (Assert.DEBUG) Assert.pre(m_dbDirectory != null);

      openDB();

      m_groups   = new FileNewsGroupRepository(this);
      m_messages = new FileMessageRepository(this);
      m_users    = new FileUserRepository(this);
      m_roles    = new FileRoleRepository(this);

      m_backup   = new BackupThread(this, m_dbDirectory, m_dbDirectory + "_", m_messages);
      m_backup.start();
   }

   public void acquireReadLock()
   throws CoffeeServiceException
   {
      try {
         m_lock.acquireReadLockInt();
      } catch (InterruptedException ex) {
         throw new CoffeeServiceException("thread interrupted");
      }
   }

   public void acquireWriteLock()
   throws CoffeeServiceException
   {
      try {
         m_lock.acquireWriteLockInt();
         if (m_lock.isOuterWriteLock()) {
            m_commit = true;
         }
      } catch (InterruptedException ex) {
         throw new CoffeeServiceException("thread interrupted");
      }
   }

   public void shutdown()
   throws CoffeeServiceException
   {
      acquireWriteLock();
      if (Assert.DEBUG) Assert.post(m_lock.isOuterWriteLock());
      abort();
      deleteLockFile();
   }

   public void releaseReadLock()
   {
      m_lock.releaseReadLock();
   }

   private void sync()
   throws CoffeeServiceException
   {
      // Precommit problems aren't fatal since we can simply
      // abort the changes and keep going.
      try {
         m_commitManager.precommit();
      } catch (IOException ex) {
         if (Assert.DEBUG) ex.printStackTrace();
         abort();
         throw new CoffeeServiceException("precommit failed: " + ex.getMessage());
      }

      // But commit failures indicate that the server has some deep problem
      // like out of disk space so we have to terminate immediately.
      try {
         m_commitManager.commit();
      } catch (IOException ex) {
         if (Assert.DEBUG) ex.printStackTrace();
         throw new FatalCoffeeServiceException("commit failed: " + ex.getMessage());
      }
   }

   private void abort()
   {
      m_sequences.abort();
      m_dbPosts.abort();
      m_dbGroups.abort();
      m_dbMessages.abort();
      m_dbUsers.abort();
      m_dbRoles.abort();
   }

   public void releaseWriteLock()
   throws CoffeeServiceException
   {
      releaseWriteLock(true);
   }

   public void releaseWriteLock(boolean commit)
   throws CoffeeServiceException
   {
      m_commit = m_commit && commit;
      if (m_lock.isOuterWriteLock()) {
         if (m_commit) {
            sync();
         } else {
            abort();
         }
      }
      m_lock.releaseWriteLock();
   }

   public DBGroups getDBGroups()
   {
      return m_dbGroups;
   }

   public DBPosts getDBPosts()
   {
      return m_dbPosts;
   }

   public DBRoles getDBRoles()
   {
      return m_dbRoles;
   }

   public DBUsers getDBUsers()
   {
      return m_dbUsers;
   }

   public DBMessages getDBMessages()
   {
      return m_dbMessages;
   }

   public long nextInSequence(String seq_name)
      throws CoffeeServiceException
   {
      try {
	 return m_sequences.next(seq_name);
      } catch (IOException ex) {
	 throw new CoffeeServiceException(ex);
      }
   }

   private String m_dbDirectory;
   private FileNewsGroupRepository m_groups;
   private FileMessageRepository m_messages;
   private FileUserRepository m_users;
   private FileRoleRepository m_roles;
   private RWLock m_lock;
   private boolean m_commit;
   private Statement m_schemaHDL;
   private ServiceConfigInfo m_config;
   private LockFile m_lockFile;
   private BackupThread m_backup;

   // keeps the singletons alive
   private DBGroups      m_dbGroups;
   private DBPosts       m_dbPosts;
   private DBRoles       m_dbRoles;
   private DBUsers       m_dbUsers;
   private DBMessages    m_dbMessages;
   private Sequence      m_sequences;
   private CommitManager m_commitManager;

   private static final String SEQUENCE_DB_NAME = "sequences";
   private static final String COMMIT_LOG_NAME  = "commit.dat";
   private static final String LOCK_FILE_NAME   = "lock.txt";
   private static final String LOCK_EXISTS_ERROR_MESSAGE = 
      Constants.PRODUCT_NAME + " failed to start because the lock file \"{0}\"" +
      " already exists.  This generally happens for one of two reasons:\n" +
      " \n" +
      Misc.joinStrings(Misc.wrapString(
        "1. There is another instance of " + Constants.PRODUCT_NAME + 
        " running using the same database.  Only one instance can use" +
        " a database at a time.  Running more than one at a time would" +
        " cause the database to become corrupted.", 70, 0, 3), '\n') +
      " \n" +
      Misc.joinStrings(Misc.wrapString(
        "2. The server was shut down improperly the last time it was run." +
        " If that is the case, simply delete the lock file and restart the" +
        " server.  DO NOT DO THIS IF ANOTHER INSTANCE OF THE SERVER IS" +
        " CURRENTLY USING THE DATABASE!", 70, 0, 3), '\n');
   private static final String LOCK_WRITE_ERROR_MESSAGE = 
      "Server failed to start because the lock file {0} could not be created.";
   private static final String LOCK_FILE_TEXT = 
      "This file is used to ensure that two instances of " + Constants.PRODUCT_NAME +
      " are never run simultaneously with the same database directory.";
}
