///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.apps.clnews.repository;

import java.io.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.coffee_server.framework.*;
import org.cooldevtools.coffee_server.services.nntp.pub.*;

public class FileRole
extends NoneRole
{
   public FileRole(String name,
                   FileArchive archive)
   {
      m_name = name;
      m_archive = archive;
   }

   public String getName()
   throws CoffeeServiceException
   {
      return m_name;
   }

   public boolean canShutdown()
   throws CoffeeServiceException
   {
      return m_name.equals(ADMIN_NAME);
   }

   public boolean canCreateNewsGroup(String group_name)
   throws CoffeeServiceException
   {
      return checkPermission(group_name, CAN_ADMIN);
   }

   public boolean canRemoveNewsGroup(NewsGroup group)
   throws CoffeeServiceException
   {
      return checkPermission(group.getName(), CAN_ADMIN);
   }

   private boolean checkPermission(String group_name,
                                   String[] perm_array)
   throws CoffeeServiceException
   {
      if (Assert.DEBUG) {
         System.out.print("checkpermission(" + m_name + ", " + group_name + ", [");
         for (int i = 0; i < perm_array.length; ++i) {
            if (i > 0) {
               System.out.print(", ");
            }
            System.out.print(perm_array[i]);
         }
         System.out.print("]) = ");
      }

      m_archive.acquireReadLock();
      try {
         String perm = m_archive.getDBRoles().getGroupPerm(m_name, group_name);
         boolean ok = Misc.stringInArray(perm, perm_array);

         if (Assert.DEBUG) {
            System.out.print(ok);
            System.out.println(" perm = " + perm);
         }

         return ok;
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public boolean canReadNewsGroup(NewsGroup group)
   throws CoffeeServiceException
   {
      return checkPermission(group.getName(), CAN_READ);
   }

   public boolean canPostToNewsGroup(NewsGroup group)
   throws CoffeeServiceException
   {
      return checkPermission(group.getName(), CAN_POST);
   }

   public boolean hasPermission(String[] perm_array)
   throws CoffeeServiceException
   {
      if (Assert.DEBUG) {
         System.out.print("hasPermission(" + m_name + ", ");
         for (int i = 0; i < perm_array.length; ++i) {
            if (i > 0) {
               System.out.print(", [");
            }
            System.out.print(perm_array[i]);
         }
         System.out.print("]) = ");
      }

      m_archive.acquireReadLock();
      try {
         String perm = m_archive.getDBRoles().getAdminPerm(m_name);
         boolean ok = m_archive.getDBRoles().hasGroupPerm(m_name, perm_array);

         if (Assert.DEBUG) {
            System.out.println(ok);
         }

         return ok;
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public boolean canRead()
   throws CoffeeServiceException
   {
      return hasPermission(CAN_READ);
   }

   public boolean canPost()
   throws CoffeeServiceException
   {
      return hasPermission(CAN_POST);
   }

   public boolean canCancelMessage(Message m)
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         NewsGroup[] groups = m.getNewsGroups();
         for (int i = 0; i < groups.length; i++) {
            if (!checkPermission(groups[i].getName(), CAN_CANCEL)) {
               return false;
            }
         }
         return true;
      } finally {
         m_archive.releaseReadLock();
      }
   }

   private boolean checkAdminLevel(String[] perm_array)
   throws CoffeeServiceException
   {
      if (Assert.DEBUG) {
         System.out.print("CheckAdminLevel(" + m_name + ", ");
         for (int i = 0; i < perm_array.length; ++i) {
            if (i > 0) {
               System.out.print(", [");
            }
            System.out.print(perm_array[i]);
         }
         System.out.print("]) = ");
      }

      m_archive.acquireReadLock();
      try {
         String perm = m_archive.getDBRoles().getAdminPerm(m_name);
         boolean ok = Misc.stringInArray(perm, perm_array);

         if (Assert.DEBUG) {
            System.out.print(ok);
            System.out.println(" perm = " + perm);
         }

         return ok;
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public boolean canDefineRole(String rolename)
   throws CoffeeServiceException
   {
      return checkAdminLevel(ADMIN_ROLES) && !isRequired(rolename);
   }

   public boolean canListRole(String rolename)
   throws CoffeeServiceException
   {
      return checkAdminLevel(ADMIN_ROLES);
   }

   public boolean canDeleteRole(String rolename)
   throws CoffeeServiceException
   {
      return checkAdminLevel(ADMIN_ROLES) && canDefineRole(rolename) && !isRequired(rolename);
   }

   public boolean canListRoles()
   throws CoffeeServiceException
   {
      return checkAdminLevel(ADMIN_USERS);
   }

   /**
      Can define (modify) a user if all of the user's roles are in our role
      membership authority.
    */
   public boolean canDefineUser(String username)
   throws CoffeeServiceException
   {
      return canListUser(username);
   }

   public boolean canListUser(String username)
   throws CoffeeServiceException
   {
      
      String[] user_roles = m_archive.getDBUsers().getRoles(username);
      if (user_roles != null && !m_archive.getDBRoles().canUseRoles(m_name, user_roles)) {
         return false;
      }

      return checkAdminLevel(ADMIN_USERS);
   }

   public boolean canDeleteUser(String username)
   throws CoffeeServiceException
   {
      return checkAdminLevel(ADMIN_USERS) && !isRequired(username);
   }

   public boolean canListUsers()
   throws CoffeeServiceException
   {
      return checkAdminLevel(ADMIN_USERS);
   }

   public boolean canChangeUserPassword(String user_to_change)
   throws CoffeeServiceException
   {
      return canDefineUser(user_to_change);
   }

   public boolean canUseRoles(String[] role_names)
   throws CoffeeServiceException
   {
      m_archive.acquireReadLock();
      try {
         return m_archive.getDBRoles().canUseRoles(m_name, role_names);
      } finally {
         m_archive.releaseReadLock();
      }
   }

   public static boolean isRequired(String name)
   {
      return name.equals("admin") || name.equals("default");
   }

   private String m_name;
   private FileArchive m_archive;

   private static final String[] CAN_READ = {
      RoleRepository.PERM_READ,
      RoleRepository.PERM_POST,
      RoleRepository.PERM_CANCEL,
      RoleRepository.PERM_ADMIN
   };
   private static final String[] CAN_POST = {
      RoleRepository.PERM_POST,
      RoleRepository.PERM_CANCEL,
      RoleRepository.PERM_ADMIN
   };
   private static final String[] CAN_CANCEL = {
      RoleRepository.PERM_CANCEL,
      RoleRepository.PERM_ADMIN
   };
   private static final String[] CAN_ADMIN = {
      RoleRepository.PERM_ADMIN
   };

   private static final String[] ADMIN_USERS = {
      RoleRepository.ADMIN_USERS,
      RoleRepository.ADMIN_ROLES
   };
   private static final String[] ADMIN_ROLES = {
      RoleRepository.ADMIN_ROLES
   };
   public static final String ADMIN_NAME = "admin";
}
