///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.coffee_server;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.util.*;
import javax.swing.*;
import org.cooldevtools.awt.*;
import org.cooldevtools.hdl.*;
import org.cooldevtools.coffee_server.framework.*;
import org.cooldevtools.coffee_server.config.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.utility.logging.*;

/**
   Run any service using info in file written by Configure.
 */
public class Run
{
   private static class GuiServer extends CoffeeServer implements Runnable
   {
      public void run()
      {
         System.exit(0);
      }

      public void shutdownAndExit()
      {
         Thread t = new ShutdownThread(false);
         t.start();
         try {
            t.join(MAX_SHUTDOWN_MILLIS);
         } catch (InterruptedException ex) {
         }
         SwingUtilities.invokeLater(this);
      }
   }

   public Run(String server_name,
              String[] args)
   throws Exception
   {
      m_serverName = server_name;
      m_started = false;

      Logger.setDefaultLogger(new FileLogger("-"));

      CommandLine cmd = new CommandLine(false);
      cmd.addStringOption(OPT_USER);
      cmd.addStringOption(OPT_GROUP);
      cmd.addFlagOption(OPT_GUI);

      int index = 0;
      try {
         index = cmd.parse(args);
      } catch (Failed ex) {
         System.err.println("invalid command line option: " + ex.getMessage());
         System.exit(1);
      }

      if ((args.length - index) != 1) {
         System.err.println("usage: Run [options] filename");
         System.exit(1);
      }

      String install_root = Misc.getInstallRoot();
      String config_filename = install_root + args[index];

      m_config = CoffeeServerConfigInfo.readFromFile(config_filename);
      m_config.loadLibraries();

      applyOptions(cmd);

      m_server = (m_frame == null) ? new CoffeeServer() : new GuiServer();
      m_server.loadConfig(m_config);

      m_config.changeIdentity();
   }

   public void go()
   {
      m_server.startup();
      m_started = true;
   }

   private class ClearLogHandler implements ActionListener {
      public void actionPerformed(ActionEvent ev) {
         m_guiLogger.clear();
      }
   }

   private class ShutdownHandler implements ActionListener,Runnable {
      public synchronized void run() {
         if (m_started) {
            m_server.shutdownAndExit();
         } else {
            System.exit(0);
         }
      }
      public synchronized void actionPerformed(ActionEvent ev) {
         SwingUtilities.invokeLater(this);
      }
   }

   private class WindowHandler extends WindowAdapter implements Runnable
   {
      public synchronized void run() {
         if (m_started) {
            m_server.shutdownAndExit();
         } else {
            System.exit(0);
         }
      }
      public synchronized void windowClosing(WindowEvent ev)
      {
         SwingUtilities.invokeLater(this);
      }
   }

   private void startGUI()
   throws LoggingException
   {
      s_isGuiBased = true;

      m_frame = new JFrame(getServerName());
      m_frame.addWindowListener(new WindowHandler());

      m_config.setGuiBased(true);
      m_config.setParentFrame(m_frame);

      JTextArea text = new JTextArea(20, 40);
      text.setEditable(false);

      m_guiLogger = new DocumentLogger(text.getDocument());
      Logger.setDefaultLogger(m_guiLogger);
      if (Logger.isToConsole()) {
         LoggerInterface old_logger = Logger.s();
         Logger.initSingleton(m_guiLogger);
         old_logger.close();
      }

      JPanel p = new JPanel();
      p.setLayout(new GridLayout(2, 1, 5, 5));
      JButton b = new JButton("Clear");
      b.addActionListener(new ClearLogHandler());
      p.add(b);
      b = new JButton("Exit");
      b.addActionListener(new ShutdownHandler());
      p.add(b);

      JPanel p2 = new JPanel();
      p2.add(p);

      m_frame.getContentPane().setLayout(new BorderLayout());
      m_frame.getContentPane().add(new JScrollPane(text), "Center");
      m_frame.getContentPane().add(p2, "East");
      m_frame.pack();

      Dimension paneSize = m_frame.getSize();
      Dimension screenSize = m_frame.getToolkit().getScreenSize();
      m_frame.setLocation((screenSize.width - paneSize.width) / 6,
                          (screenSize.height - paneSize.height) / 6);

      m_frame.show();
   }

   private void applyOptions(CommandLine cmd)
   throws LoggingException
   {
      String value = cmd.getString(OPT_USER, null);
      if (value != null) {
         m_config.setUserName(value);
      }

      value = cmd.getString(OPT_GROUP, null);
      if (value != null) {
         m_config.setGroupName(value);
      }

      if (cmd.getFlag(OPT_GUI, false)) {
         startGUI();
      }
   }

   private static String[] buildErrorMessage(String error_desc)
   {
      SuperVector vec = new SuperVector();
      vec.addElement("Server terminated with fatal error:\n");
      vec.addElement(" \n");
      String[] lines = Misc.wrapString(error_desc, 75, 2, 2);
      for (int i = 0; i < lines.length; i++) {
         vec.addElement(lines[i] + "\n");
      }
      vec.addElement(" \n");
      return Misc.wrapStrings(vec.getStrings(), 75, 0, 0);
   }

   public static void main(String[] args)
   {
      extended_main(null, args);
   }

   public static void extended_main(String server_name,
                                    String[] args)
   {
      s_isGuiBased = false;
      try {
         Run r = new Run(server_name, args);
         r.go();
      } catch (Exception ex) {
         if (Assert.DEBUG) {
            ex.printStackTrace();
         }
         String[] lines = buildErrorMessage(ex.getMessage());
         if (!s_isGuiBased) {
            for (int i = 0; i < lines.length; i++) {
               System.out.println(lines[i]);
            }
         } else {
            JOptionPane.showMessageDialog(null,
                                          lines,
                                          "Server Failure",
                                          JOptionPane.ERROR_MESSAGE);
         }
         for (int i = 0; i < lines.length; i++) {
            Logger.s().logSilently(MessagePriority.LOG_ERROR,
                                   lines[i]);
         }
         System.exit(1);
      }
   }

   private String getServerName()
   {
      String name = m_serverName;
      if (name != null) {
         return name;
      }

      name = m_config.getServerName();
      if (name == null) {
         name = "";
      }

      return name;
   }

   private static boolean s_isGuiBased;
   private boolean m_started;
   private JFrame m_frame;
   private CoffeeServer m_server;
   private CoffeeServerConfigInfo m_config;
   private DocumentLogger m_guiLogger;
   private String m_serverName;

   private static final String OPT_USER  = "user";
   private static final String OPT_GROUP = "group";
   private static final String OPT_GUI   = "gui";
   private static final long   MAX_SHUTDOWN_MILLIS = 10000;
}
