///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//


package org.cooldevtools.coffee_server.config;

import java.io.*;
import java.util.*;
import java.awt.Component;
import org.cooldevtools.hdl.*;
import org.cooldevtools.utility.Assert;
import org.cooldevtools.coffee_server.framework.*;

public class ServiceConfigInfo
{
   ServiceConfigInfo()
   {
      initialize(null);
   }

   ServiceConfigInfo(CoffeeServerConfigInfo server_config)
   {
      initialize(server_config);
   }

   ServiceConfigInfo(String name)
   {
      m_name = name;
   }

   private void initialize(CoffeeServerConfigInfo server_config)
   {
      m_name = "";
      m_className = "";
      m_address = "";
      m_port = 0;
      m_alternatePort = 0;
      m_socketTimeout = 0;
      m_threadModel = SINGLE_THREADED;
      m_manager = null;
      m_bean = null;
      m_properties = new Hashtable();
      m_serverConfig = server_config;
   }

   public CoffeeServerConfigInfo getServerConfig()
   {
      return m_serverConfig;
   }

   public String getName()
   {
      return m_name;
   }

   public void setName(String value)
   {
      m_name = value;
   }

   public String getAddress()
   {
      return m_address;
   }

   public void setAddress(String value)
   {
      m_address = value;
   }

   public int getAlternatePort()
   {
      return m_alternatePort == 0 ? m_port : m_alternatePort;
   }

   public void setAlternatePort(int value)
   {
      m_alternatePort = value;
   }

   public int getPort()
   {
      return m_port;
   }

   public void setPort(int value)
   {
      m_port = value;
   }

   public String getClassName()
   {
      return m_className;
   }

   public String toString()
   {
      StringBuffer buffer = new StringBuffer();
      buffer.append(m_name);
      buffer.append(" (");
      buffer.append(m_className);
      buffer.append(")");
      return buffer.toString();
   }

   public boolean isValid(StringBuffer message)
   {
      if ((m_name == null) || (m_name.length() == 0)) {
         message.append("name not set");
         return false;
      }

      if ((m_port < 0) || (m_port > MAX_PORT)) {
         message.append("invalid port");
         return false;
      }

      if (m_manager == null) {
         message.append("invalid class name");
         return false;
      }

      return true;
   }

   private CoffeeServiceManager findClass(String name)
      throws BadClassNameException
   {
      if ((name == null) || (name.length() == 0)) {
	 throw new BadClassNameException("no class name specified");
      }
      try {
	 Class _class = (Class)Class.forName(name);
	 return (CoffeeServiceManager)_class.newInstance();
      } catch (ClassCastException ex) {
	 throw new BadClassNameException(name + " is not a CoffeeServiceManager");
      } catch (ClassNotFoundException ex) {
	 throw new BadClassNameException("cannot load " + name);
      } catch (InstantiationException ex) {
	 throw new BadClassNameException("cannot instantiate " + name);
      } catch (IllegalAccessException ex) {
	 throw new BadClassNameException("no public constructor for " + name);
      } catch (NoSuchMethodError ex) {
	 throw new BadClassNameException("no default constructor for " + name);
      }
   }

   public void setClassName(String class_name)
      throws BadClassNameException
   {
      try {
	 m_manager   = findClass(class_name);
	 m_bean      = null;
	 m_className = class_name;
      } catch (BadClassNameException ex) {
         m_manager   = null;
         m_bean      = null;
         m_className = null;
         throw ex;
      }
   }

   public CoffeeServiceManager getManager()
   {
      return m_manager;
   }

   public boolean hasBean()
   {
      if (m_manager == null) {
         return false;
      }

      return m_manager.hasConfigBean();
   }

   public CoffeeServiceConfigBean getBean()
   throws CoffeeServiceException
   {
      if (m_bean != null) {
         return m_bean;
      }

      if (m_manager == null) {
         return null;
      }

      m_bean = m_manager.newConfigBeanInstance();
      if ((m_bean != null) && (m_bean.getComponent() == null)) {
         throw new CoffeeServiceException("bean has no Component");
      }

      return m_bean;
   }

   public Statement getConfiguration()
   {
      return m_configuration;
   }

   public void setConfiguration(Statement value)
   {
      m_configuration = value;
   }

   public int getThreadModel()
   {
      return m_threadModel;
   }

   public void setThreadModel(int value)
   {
      m_threadModel = value;
   }

   public void copyFrom(ServiceConfigInfo other)
   {
      if (other != this) {
         m_name = other.m_name;
         m_port = other.m_port;
         m_alternatePort = other.m_alternatePort;
         m_className = other.m_className;
         m_manager = other.m_manager;
         m_bean = other.m_bean;
         m_configuration = other.m_configuration;
         m_threadModel = other.m_threadModel;
         m_properties = other.m_properties;
      }
   }

   public Token getProperty(String name)
   {
      return (Token)m_properties.get(name);
   }

   public void setProperty(String name,
                           Token value)
   {
      m_properties.put(name, value);
   }

   public static ServiceConfigInfo readFromHDL(CoffeeServerConfigInfo server_config,
                                               Statement stmt)
   throws IOException, BadClassNameException, HDLException
   {
      ServiceConfigInfo info = new ServiceConfigInfo(server_config);

      info.m_name = stmt.argument(0).strValue();
      
      int num_children = stmt.numChildren();
      for (int i = 0; i < num_children; i++) {
         Statement child = stmt.child(i);
         Token name = child.name();
         if (name.equals(TOK_PORT)) {
            info.m_port = child.argument(0).intValue();
         } else if (name.equals(TOK_ALT_PORT)) {
            info.m_alternatePort = child.argument(0).intValue();
         } else if (name.equals(TOK_ADDRESS)) {
            info.m_address = child.argument(0).strValue();
         } else if (name.equals(TOK_MODEL)) {
            info.m_threadModel = child.argument(0).intValue();
         } else if (name.equals(TOK_CONFIG)) {
            info.m_configuration = child;
         } else if (name.equals(TOK_TIMEOUT)) {
            info.m_socketTimeout = child.argument(0).intValue();
         } else if (name.equals(TOK_CLASS)) {
            String class_name = child.argument(0).strValue();
            info.setClassName(class_name);
         } else if (name.equals(TOK_PROPERTY)) {
            info.m_properties.put(child.argument(0).strValue(), child.argument(1));
         } else {
            Assert.impossible();
         }
      }

      return info;
   }

   public Statement writeToHDL()
   throws IOException
   {
      Statement root = new Statement(TOK_SERVICE, true);
      root.addArgument(Token.strToken(m_name));

      Statement stmt = new Statement(TOK_PORT, false);
      stmt.addArgument(Token.intToken(m_port));
      root.addChild(stmt);

      stmt = new Statement(TOK_ALT_PORT, false);
      stmt.addArgument(Token.intToken(getAlternatePort()));
      root.addChild(stmt);

      if (m_address.length() > 0) {
	 stmt = new Statement(TOK_ADDRESS, false);
	 stmt.addArgument(Token.strToken(m_address));
	 root.addChild(stmt);
      }

      stmt = new Statement(TOK_MODEL, false);
      stmt.addArgument(Token.intToken(m_threadModel));
      root.addChild(stmt);

      stmt = new Statement(TOK_TIMEOUT, false);
      stmt.addArgument(Token.intToken(m_socketTimeout));
      root.addChild(stmt);

      stmt = new Statement(TOK_CLASS, false);
      stmt.addArgument(Token.strToken(m_className));
      root.addChild(stmt);

      Enumeration en = m_properties.keys();
      while (en.hasMoreElements()) {
         String name = (String)en.nextElement();
         stmt = new Statement(TOK_PROPERTY, false);
         stmt.addArgument(Token.idToken(name));
         stmt.addArgument((Token)m_properties.get(name));
         root.addChild(stmt);
      }

      if (m_configuration != null) {
         root.addChild(m_configuration);
      }

      return root;
   }
   
   public int getSocketTimeout()
   {
      return m_socketTimeout * 1000;
   }

   public static ServiceConfigInfo[] toArray(Vector v)
   {
      synchronized (v) {
         ServiceConfigInfo[] answer = new ServiceConfigInfo[v.size()];
         v.copyInto(answer);
         return answer;
      }
   }

   private String m_name;
   private int m_port;
   private int m_alternatePort;
   private int m_threadModel;
   private String m_className;
   private Statement m_configuration;
   private int m_socketTimeout;
   private Hashtable m_properties;
   private String m_address;
   private CoffeeServerConfigInfo m_serverConfig;

   private CoffeeServiceManager m_manager;
   private CoffeeServiceConfigBean m_bean;

   public static final int SINGLE_THREADED = 0;
   public static final int MULTI_THREADED  = 1;
   public static final int THREAD_POOLED   = 2;

   private static final int MAX_PORT       = 65535;

   private static final Token TOK_SERVICE  = Token.idToken("service");
   private static final Token TOK_PORT     = Token.idToken("port");
   private static final Token TOK_ALT_PORT = Token.idToken("alternate_port");
   private static final Token TOK_CLASS    = Token.idToken("class");
   private static final Token TOK_ADDRESS  = Token.idToken("address");
   private static final Token TOK_CONFIG   = Token.idToken("configuration");
   private static final Token TOK_MODEL    = Token.idToken("thread_model");
   private static final Token TOK_TIMEOUT  = Token.idToken("socket_timeout");
   private static final Token TOK_PROPERTY = Token.idToken("property");
}
