///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//


package org.cooldevtools.coffee_server.framework;

import java.io.*;
import java.net.*;
import java.util.*;
import org.cooldevtools.utility.logging.*;
import org.cooldevtools.coffee_server.util.*;
import org.cooldevtools.coffee_server.config.*;

/**
   Main class that manages all of the objects and connections
   while the system is running.
 */
public class CoffeeServer
implements ExceptionHandler, ThreadPoolOwner
{
   public CoffeeServer()
   {
      m_numThreads  = DEFAULT_NUM_THREADS;
      clear();
   }

   public CoffeeServerConfigInfo getConfig()
   {
      return m_config;
   }

   public void loadConfig(CoffeeServerConfigInfo config)
   throws CoffeeServiceException, IOException
   {
      m_config = config;
      setNumThreads(m_config.getNumThreads());

      ServiceConfigInfo[] services = m_config.getServices();
      for (int i = 0; i < services.length; i++) {
	 String address = services[i].getAddress();
         int primary_port = services[i].getPort();
         int alternate_port = services[i].getAlternatePort();
         CoffeeServiceManager manager = services[i].getManager();
         manager.configure(services[i]);

         switch (services[i].getThreadModel()) {
         case ServiceConfigInfo.SINGLE_THREADED:
            addSingleThreadedService(address, primary_port, alternate_port, manager);
            break;

         case ServiceConfigInfo.MULTI_THREADED :
            addMultiThreadedService(address, primary_port, alternate_port, manager);
            break;

         case ServiceConfigInfo.THREAD_POOLED  :
            addPooledService(address, primary_port, alternate_port, manager);
            break;
         }
      }
   }

   public void writeConfigToFile()
   throws CoffeeServiceException, IOException
   {
      CoffeeServiceManager[] managers = getServiceManagers();
      for (int i = 0; i < managers.length; i++) {
         managers[i].updateConfiguration();
      }
      m_config.writeToFile();
   }

   public ThreadPool getThreadPool(Object key)
   {
      return m_pool;
   }

   public void startup()
   {
      m_pool = new ThreadPool();

      m_threads = new Vector();

      for (int i = 0; i < m_numThreads; i++) {
         Thread t = new ThreadPoolWorker(m_pool, this);
         m_threads.addElement(t);
         t.start();
      }

      ServiceAcceptor[] acceptors = getAcceptors();
      for (int i = 0; i < acceptors.length; i++) {
         acceptors[i].start();
      }
   }

   private ServerSocket bindSocketImpl(String host,
				       int primary_port,
				       int alternate_port)
   throws IOException
   {
      final int QUEUE_SIZE = 50;
      ServerSocket socket = null;
      InetAddress  address = (host.length() > 0) ? InetAddress.getByName(host) : null;
      try {
         socket = new ServerSocket(primary_port, QUEUE_SIZE, address);
      } catch (IOException ex) {
         if (primary_port == alternate_port) {
            throw ex;
         } else {
            socket = new ServerSocket(alternate_port, QUEUE_SIZE, address);
         }
      }
      Logger.s().logSilently(MessagePriority.LOG_NOTICE,
                             "listening on ip addr " + socket.getInetAddress()
                             + " port " + socket.getLocalPort());
      return socket;
   }

   private ServerSocket bindSocket(String host,
				   int primary_port,
                                   int alternate_port)
   throws IOException
   {
      ServerSocket socket;
      try {
	 socket = bindSocketImpl(host, primary_port, alternate_port);
      } catch (IOException ex) {
	 if (host.length() == 0) {
	    throw ex;
	 }

	 Logger.s().logSilently(MessagePriority.LOG_ERROR,
				"unable to bind host/address '" + host + "': " + ex.getMessage());
	 socket = bindSocketImpl("", primary_port, alternate_port);
      }
      return socket;
   }

   public void addSingleThreadedService(String host,
					int primary_port,
                                        int alternate_port, 
                                        CoffeeServiceManager manager)
   throws IOException
   {
      ServerSocket socket = bindSocket(host, primary_port, alternate_port);
      ServiceAcceptor acceptor = new SingleThreadServiceAcceptor(this, socket, manager, this);
      m_acceptors.addElement(acceptor);
      m_managers.addElement(manager);
   }

   public void addMultiThreadedService(String host,
				       int primary_port,
                                       int alternate_port, 
                                       CoffeeServiceManager manager)
   throws IOException
   {
      ServerSocket socket = bindSocket(host, primary_port, alternate_port);
      ServiceAcceptor acceptor = new MultiThreadServiceAcceptor(this, socket, manager, this);
      m_acceptors.addElement(acceptor);
      m_managers.addElement(manager);
   }

   public void addPooledService(String host,
				int primary_port,
                                int alternate_port, 
                                CoffeeServiceManager manager)
   throws IOException
   {
      ServerSocket socket = bindSocket(host, primary_port, alternate_port);
      ServiceAcceptor acceptor = new ThreadPoolServiceAcceptor(this,
                                                               socket,
                                                               manager,
                                                               this,
                                                               null,
                                                               this);
      m_acceptors.addElement(acceptor);
      m_managers.addElement(manager);
   }

   public class ShutdownThread extends Thread {
      public ShutdownThread(boolean exit)
      {
         m_exit = exit;
      }

      public void run()
      {
         m_pool.close();
         ServiceAcceptor[] acceptors = getAcceptors();
         for (int i = 0; i < acceptors.length; i++) {
            try {
               acceptors[i].shutdown();
            } catch (CoffeeServiceException ex) {
               ex.printStackTrace();
            }
         }

         CoffeeServiceManager[] managers = getServiceManagers();
         for (int i = 0; i < managers.length; i++) {
            try {
               managers[i].shutdown();
            } catch (CoffeeServiceException ex) {
               ex.printStackTrace();
            }
         }

         clear();
         
         if (m_exit) {
            System.exit(0);
         }
      }

      private boolean m_exit;
   }

   public void shutdown()
   {
      new ShutdownThread(false).start();
   }

   public void shutdownAndExit()
   {
      new ShutdownThread(true).start();
   }

   private void clear()
   {
      m_pool = null;
      m_threads = null;
      m_acceptors = new Vector();
      m_managers = new Vector();
   }

   private ServiceAcceptor[] getAcceptors()
   {
      ServiceAcceptor[] answer = new ServiceAcceptor[m_acceptors.size()];
      m_acceptors.copyInto(answer);
      return answer;
   }

   private CoffeeServiceManager[] getServiceManagers()
   {
      CoffeeServiceManager[] answer = new CoffeeServiceManager[m_managers.size()];
      m_managers.copyInto(answer);
      return answer;
   }

   public void setNumThreads(int value)
   {
      m_numThreads = value;
   }

   public void fatalAcceptorException(ServiceAcceptor acceptor,
                                      Exception ex)
   {
      ex.printStackTrace();
   }

   public void serviceFailure(ClientRequest request,
                              CoffeeService service,
                              Exception ex)
   {
      ex.printStackTrace();
      if (ex instanceof FatalCoffeeServiceException) {
         Logger.s().logSilently(MessagePriority.LOG_EMERGENCY,
                                "IMMEDIATE SERVER TERMINATION ON FATAL EXCEPTION: "
                                + ex);
         System.exit(1);
      } else {
         Logger.s().logSilently(MessagePriority.LOG_ERROR,
                                "ERROR: " + ex);
      }
   }

   private int m_numThreads;
   private ThreadPool m_pool;
   private Vector m_threads;
   private Vector m_acceptors;
   private Vector m_managers;
   private CoffeeServerConfigInfo m_config;

   public static final int DEFAULT_NUM_THREADS = 10;
}
