///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//


package org.cooldevtools.coffee_server.services.nntp.protocol;

import java.io.*;
import java.net.*;
import java.util.*;
import org.cooldevtools.coffee_server.config.*;
import org.cooldevtools.coffee_server.framework.*;
import org.cooldevtools.coffee_server.services.nntp.pub.*;
import org.cooldevtools.utility.logging.*;

public class NntpService
implements CoffeeService
{
   public NntpService()
   {
      m_commands = new Hashtable();
   }

   public void serviceRequest(ClientRequest req)
   throws IOException, CoffeeServiceException
   {
      NntpSession session = new NntpSession(req, m_NewsFactory);

      NntpReader reader = new NntpReader(new BufferedInputStream(req.getInputStream()));
      NntpWriter writer = new NntpWriter(new BufferedOutputStream(req.getOutputStream()));

      NntpService_MODE_READER.greetReader(session, writer);
      writer.flush();

      try {
         ServiceConfigInfo config = session.getRequest().getServiceManager().getConfiguration();

         // we loop forever or until we see 50 consecutive unknown commands
         int consecutive_unknowns = 0;
         while (true) {
            if (consecutive_unknowns == 50) {
               writer.writeResponse(501, "Goodbye.  We don't seem to speak the same language.");
               writer.flush();
               break;
            }

            reader.readNewLine();
            String command_name = reader.getNextCommand();

            long start = System.currentTimeMillis();
            try {
               NntpCommand command = (NntpCommand)m_commands.get(command_name);
               if (command == null) {
                  if (command_name.length() > 0) {
                     String msg = "Sorry, '" + command_name + "' not recognized";
                     writer.writeResponse(501, msg);
                     Logger.s().logSilently(MessagePriority.LOG_NOTICE, 501 + " " + msg);
                  }
                  consecutive_unknowns++;
               } else {
		  if (command.isAuthenticationRequired() && !session.isAuthenticated()) {
		     writer.writeResponse(ServerConnection.CODE_AUTH_REQUIRED, "authentication required");
		  } else {
		     if (!command.executeCommand(session, reader, writer)) {
			String msg = "Invalid '" + command_name + "' command";
			writer.writeResponse(501, msg);
			Logger.s().logSilently(MessagePriority.LOG_NOTICE, 501 + " " + msg);
		     }
		  }
                  consecutive_unknowns = 0;
               }
               writer.flush();
            } finally {
	       long finish = System.currentTimeMillis();
	       if (!Logger.s().isPriorityBelowMinimum(MessagePriority.LOG_INFO)) {
		  Logger.s().logSilently(MessagePriority.LOG_INFO, 
					 "Command ''{0}'' completed in {1} milliseconds for user {2}",
					 new Object[] { command_name,
							new Long(finish - start),
							session.getUser().getSenderAddress() });
	       }
            }
         }
      } catch (SocketException ex) {
         // this is probably a timeout
         Logger.s().logSilently(MessagePriority.LOG_NOTICE, ex.toString());
	 try {
	    writer.writeResponse(503, "Socket error or timeout - goodbye.");
	 } catch (Exception ignored) {
	 }
      } catch (EOFException ex) {
         // gracefully shutdown the connection.
      } finally {
         req.close();
      }
   }

   public void addCommand(String name,
                          NntpCommand command)
   {
      m_commands.put(name, command);
   }

   public NewsRepositoryFactory getNewsFactory()
   {
      return m_NewsFactory;
   }

   public void setNewsFactory(NewsRepositoryFactory value)
   {
      m_NewsFactory = value;
   }

   private Hashtable m_commands;
   private NewsRepositoryFactory m_NewsFactory;
}
