///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//


package org.cooldevtools.coffee_server.services.nntp.protocol;

import java.io.*;
import java.util.*;
import java.text.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.utility.logging.*;
import org.cooldevtools.coffee_server.util.*;
import org.cooldevtools.coffee_server.framework.*;
import org.cooldevtools.coffee_server.services.nntp.pub.*;

public class NntpService_POST extends NntpCommand implements Runnable
{
   public NntpService_POST()
   {
      basicInitialization();
      m_messagePool = new ThreadPool();
      m_thread = new Thread(this);
      m_thread.setDaemon(true);
      m_thread.start();
   }

   NntpService_POST(NntpService_POST master_poster)
   {
      basicInitialization();
      m_messagePool = master_poster.m_messagePool;
   }

   private void basicInitialization()
   {
      m_controls = new Hashtable();
      m_controls.put(CANCEL_COMMAND, new CancelMessage());
      m_controls.put(NEWGROUP_COMMAND, new NewGroupMessage());
      m_controls.put(RMGROUP_COMMAND, new RemoveGroupMessage());
   }

   private void backgroundPost(Tuple request)
   {
      NntpSession session = (NntpSession)request.getElement(0);
      Vector      header  = (Vector)request.getElement(1);
      Vector      body    = (Vector)request.getElement(2);

      Exception failure = null;
      String message_id = "<unknown>";
      try {
         // These commands involve multiple operations on the message
         // and group repositories.  All of these operations are
         // performed in a single transaction.
         session.getFactory().acquireWriteLock();
         try {
            Message message = session.getMessages().createMessage();
            message.setHeader(header);
            message.setBody(body);
            message_id = message.messageId();

            String control = message.getField(Message.CONTROL);
            if (control != null) {
               handleControlMessage(session, control, message);
            } else {
               postMessage(session, message);
            }
         } finally {
            session.getFactory().releaseWriteLock();
         }
      } catch (Failed ex) {
         failure = ex;
      } catch (CoffeeServiceException ex) {
         failure = ex;
      }

      if (failure != null) {
         Logger.s().logSilently(MessagePriority.LOG_WARNING,
                                "background post failed for message {0}: {1}",
                                new Object[] { message_id, failure.toString() });
      }
   }

   /**
      Shutdown the background thread after flushing all requests from queue.
    */
   public void shutdown()
   throws InterruptedException
   {
      Logger.s().logSilently(MessagePriority.LOG_NOTICE,
                             "background post thread shutdown begins with {0} messages in queue",
                             new Object[] { new Integer(m_messagePool.size()) });
      m_messagePool.close();
      m_thread.join();
      Logger.s().logSilently(MessagePriority.LOG_NOTICE,
                             "background post thread shutdown complete");
   }

   public void run()
   {
      try {
         while (true) {
            try {
               Tuple request = (Tuple)m_messagePool.get();

               long start = System.currentTimeMillis();
               backgroundPost(request);
               long elapsed = System.currentTimeMillis() - start;

               if (!Logger.s().isPriorityBelowMinimum(MessagePriority.LOG_INFO)) {
                  Logger.s().logSilently(MessagePriority.LOG_INFO,
                                         "background post time = {0} mills",
                                         new Object[] { new Long(elapsed) });
               }

               // ensure readers have time to do their thing
               // unless we are shutting down
               if (!m_messagePool.isClosed()) {
                  Thread.sleep(POST_THREAD_DELAY);
               }
            } catch (InterruptedException ex) {
               // ignored
            }
         }
      } catch (ThreadPoolClosedException ex) {
      }
   }

   private static abstract class ControlMessageStrategy {
      public abstract void execute(NntpSession session,
                                   String control_header,
                                   Message control_message)
      throws CoffeeServiceException, Failed;
   }

   private static class CancelMessage extends ControlMessageStrategy {
      public void execute(NntpSession session,
                          String control_header,
                          Message control_message)
      throws CoffeeServiceException, Failed
      {
         int index = control_header.indexOf('<');
         if (index == -1) {
            throw new Failed("invalid cancel command: no message id: " + control_header);
         }

         String message_id = control_header.substring(index);
         MessageRepository messages = session.getMessages();
         Message message = messages.findMessage(message_id);

         if (message == null) {
            throw new Failed("invalid cancel command: unknown message: " + control_header);
         }

         User user = session.getUser();
         if (!user.canCancelMessage(message)) {
            throw new Failed("invalid cancel command: permission denied: " + control_header);
         }

         messages.removeMessage(user, message);
         if (!Logger.s().isPriorityBelowMinimum(MessagePriority.LOG_INFO)) {
            Logger.s().logSilently(MessagePriority.LOG_INFO, "message {0} cancelled by user {1}",
                                   new Object[] { message_id, session.getUser().getSenderAddress() });
         }
      }
   }

   private static class NewGroupMessage extends ControlMessageStrategy {
      public void execute(NntpSession session,
                          String control_header,
                          Message control_message)
      throws CoffeeServiceException, Failed
      {
         String group_name;
         try {
            int start = NntpService_POST.NEWGROUP_COMMAND.length() + 1;
            int space = control_header.indexOf(' ', start);
            if (space == -1) {
               group_name = control_header.substring(start);
            } else {
               if (!control_header.substring(space).equals("unmoderated")) {
                  throw new Failed("only unmoderated groups allowed: " + control_header);
               }
               group_name = control_header.substring(start, space);
            }
         } catch (StringIndexOutOfBoundsException ex) {
            throw new Failed("invalid newgroup command: " + control_header);
         }

         NewsGroup group = session.getNewsGroups().findNewsGroup(group_name);
         if (group != null) {
            throw new Failed("group already exists: " + control_header);
         }

         User user = session.getUser();
         if (!NntpService_BCC_DEFINE_GROUP.canCreateNewsGroup(session, user, group_name, true)) {
            throw new Failed("invalid newgroup command: permission denied: " + control_header);
         }

         String[] body = control_message.getBody();
         String description = (body.length == 0) ? "" : body[0];
         session.getNewsGroups().createNewsGroup(user,
                                                 group_name,
                                                 user.getEmailAddress(),
                                                 description);
         if (!Logger.s().isPriorityBelowMinimum(MessagePriority.LOG_INFO)) {
            Logger.s().logSilently(MessagePriority.LOG_INFO, "group {0} created by user {1}",
                                   new Object[] { group_name, user.getSenderAddress() });
         }
      }
   }

   private static class RemoveGroupMessage extends ControlMessageStrategy {
      public void execute(NntpSession session,
                          String control_header,
                          Message control_message)
      throws CoffeeServiceException, Failed
      {
         String group_name;
         try {
            int start = NntpService_POST.RMGROUP_COMMAND.length() + 1;
            group_name = control_header.substring(start).trim();
         } catch (StringIndexOutOfBoundsException ex) {
            throw new Failed("invalid rmgroup command: " + control_header);
         }

         NewsGroup group = session.getNewsGroups().findNewsGroup(group_name);
         if (group == null) {
            throw new Failed("unknown group: " + control_header);
         }

         User user = session.getUser();
         if (!user.canRemoveNewsGroup(group)) {
            throw new Failed("invalid rmgroup command: permission denied: " + control_header);
         }

         session.getNewsGroups().removeNewsGroup(user, group_name);
         if (!Logger.s().isPriorityBelowMinimum(MessagePriority.LOG_INFO)) {
            Logger.s().logSilently(MessagePriority.LOG_INFO, "group {0} removed by user {1}",
                                   new Object[] { group_name, user.getSenderAddress() });
         }
      }
   }

   private void postMessage(NntpSession session,
                            Message message)
   throws CoffeeServiceException, Failed
   {
      String message_id = message.getField(Message.MESSAGE_ID);
      if (session.getMessages().messageExists(message_id)) {
         throw new Failed("error: duplicate message id: " + message_id);
      }

      SuperHashtable group_names = new SuperHashtable();
      parseGroupNames(message.getFieldAlways(Message.NEWSGROUPS), group_names);

      NewsGroup[] groups = new NewsGroup[group_names.size()];
      validateGroups(session, group_names.getKeyStrings(), groups);

      User user = session.getUser();
      session.getMessages().storeMessage(user, message);
      for (int i = 0; i < groups.length; i++) {
         if (groups[i] != null) {
            groups[i].postMessage(user, message);
         }
      }

      if (!Logger.s().isPriorityBelowMinimum(MessagePriority.LOG_INFO)) {
         Logger.s().logSilently(MessagePriority.LOG_INFO, "message {0} posted by user {1}",
                                new Object[] { message_id, user.getSenderAddress() });
      }
   }

   private static boolean underMaxMessagesLimit(NntpSession session,
                                                NewsGroup group,
                                                String limit_name,
                                                long limit)
   throws CoffeeServiceException
   {
      long current_number = group.getMessageCount();
      if (current_number < limit) {
         return true;
      }

      Logger.s().logSilently(MessagePriority.LOG_ERROR,
                             "post request denied: {0} servers limited to {1} messages per group",
                             new Object[] { limit_name, new Long(limit) });
      return false;
   }

   public static boolean canPostToNewsGroup(NntpSession session,
                                            User user,
                                            NewsGroup group)
   throws CoffeeServiceException
   {
      if (!user.canPostToNewsGroup(group)) {
         Logger.s().logSilently(MessagePriority.LOG_ERROR,
                                "user {0} not authoized to post to group {1}",
                                new Object[] { user.getUsername(), group.getName() });
         return false;
      }

      return true;
   }

   private void validateGroups(NntpSession session,
                               String[] group_names,
                               NewsGroup[] groups)
   throws CoffeeServiceException, Failed
   {
      User user = session.getUser();
      NewsGroupRepository group_repository = session.getNewsGroups();

      if (group_names.length == 0) {
         throw new Failed("no news groups specified");
      }

      int num_groups = 0;
      for (int i = 0; i < group_names.length; i++) {
         String group_name = group_names[i];
         NewsGroup group = group_repository.findNewsGroup(group_name);
         if (group == null) {
            // ignored
         } else if (!canPostToNewsGroup(session, user, group)) {
            throw new Failed("permission to post to " + group_name + " denied");
         } else {
            groups[i] = group;
            num_groups++;
         }
      }

      if (num_groups == 0) {
         throw new Failed("no valid groups specified");
      }
   }

   private void handleControlMessage(NntpSession session,
                                     String control_header,
                                     Message control_message)
   throws CoffeeServiceException, Failed
   {
      String command;
      try {
         command = control_header.substring(0, control_header.indexOf(" "));
      } catch (StringIndexOutOfBoundsException ex) {
         throw new Failed("invalid control header: " + control_header);
      }

      ControlMessageStrategy strategy = (ControlMessageStrategy)m_controls.get(command.toLowerCase());
      if (strategy == null) {
         throw new Failed("unsupported command: " + command);
      }

      strategy.execute(session, control_header, control_message);
   }

   private void enqueuePost(NntpSession session,
                            Vector header,
                            Vector body)
   throws Failed
   {
      try {
         Tuple request = new Tuple(session, header, body);

         // hopefully we can enqueue on first try
         if (m_messagePool.put(request, MAX_POOL_SIZE)) {
            return;
         }

         // otherwise we sleep for a while to give the queue time to drain
         try {
            Thread.sleep(CLIENT_POST_DELAY);
         } catch (InterruptedException ex) {
         }

         // then we try again and fail if queue still full
         if (!m_messagePool.put(request, MAX_POOL_SIZE)) {
            throw new Failed("server too busy to post, try again later");
         }
      } catch (ThreadPoolClosedException ex) {
         // a shutdown is in progress so we shoo new posts away
         throw new Failed("server not accepting posts, try again later");
      }
   }

   protected void handlePost(int ready_code,
                             int ok_code,
                             int failed_code,
                             NntpSession session,
                             NntpReader in,
                             NntpWriter out)
   throws IOException
   {
      out.writeResponse(ready_code, "send article to be posted.  End with <CR-LF>.<CR-LF>");
      out.flush();

      try {
         // Read message without any locks to avoid slowing down other users.
         Vector header = new Vector();
         Vector body = new Vector();
         String sender = session.getUser().getSenderAddress();
         readMessage(in, header, body, session.getFactory().getMaxMessageSize(), sender);
         enqueuePost(session, header, body);
         out.writeResponse(ok_code, "article queued for posting");
      } catch (CoffeeServiceException ex) {
         if (Assert.DEBUG) ex.printStackTrace();
         out.writeResponse(failed_code, "post failed: " + ex.getMessage());
      } catch (Failed ex) {
         out.writeResponse(failed_code, "post failed: " + ex.message());
      }
   }


   public boolean executeCommand(NntpSession session,
                                 NntpReader in,
                                 NntpWriter out)
   throws IOException, CoffeeServiceException
   {
      handlePost(340, 240, 441, session, in, out);
      return true;
   }

   private void parseGroupNames(String names,
                                SuperHashtable group_names)
   {
      int offset = 0;
      int length = names.length();
      for (int i = 0; i < length; i++) {
         char ch = names.charAt(i);
         if (ch == ',') {
            String group_name = names.substring(offset, i);
            group_names.put(group_name, group_name);
            i++;
            while ((i < length) && (names.charAt(i) == ' ')) {
               i++;
            }
            offset = i;
         }
      }
      if (offset < length) {
         String group_name = names.substring(offset);
         group_names.put(group_name, group_name);
      }
   }

   private void skipRestOfMessage(NntpReader in)
   throws IOException
   {
      String s = in.readLine();
      while (!s.equals(END_OF_TEXT)) {
         s = in.readLine();
      }
   }

   private void readMessage(NntpReader in,
                            Vector header,
                            Vector body,
                            int max_message_size,
                            String sender)
   throws IOException, CoffeeServiceException, Failed
   {
      int total_bytes = 0;
      int sender_index = -1;
      boolean have_date = false;
      boolean have_message_id = false;
      String from = null;
      String s = in.readLine();
      while (s.length() > 0) {
         if (s.equals(END_OF_TEXT)) {
            throw new Failed(INVALID_HEADER);
         }
         header.addElement(s);
         if (from == null || sender_index < 0) {
            String as_lower = s.toLowerCase();
            if (from == null && as_lower.startsWith("from: ")) {
               from = s.substring(6).trim();
            }
            if (sender_index < 0 && as_lower.startsWith("sender: ")) {
               sender_index = header.size() - 1;
            }
            if (!have_date && as_lower.startsWith("date: ")) {
               have_date = true;
            }
            if (!have_message_id && as_lower.startsWith("message-id: ")) {
              have_message_id = true;
            }
         }
         total_bytes += s.length();
         if (total_bytes > max_message_size) {
            skipRestOfMessage(in);
            throw new Failed(MessageFormat.format(MESSAGE_TOO_LONG,
                                                  new Object[] { new Integer(max_message_size) }));
         }
         s = in.readLine();
      }

      if (header.size() == 0) {
         skipRestOfMessage(in);
         throw new Failed(INVALID_HEADER);
      }

      if (!have_date) {
         header.addElement("Date: " + DateUtils.dateToNntpString(new Date()));
      }

      if (!have_message_id) {
        header.addElement("Message-ID: " + Message.createMessageID());
      }

      s = in.readLine();
      while (!s.equals(END_OF_TEXT)) {
         body.addElement(s);
         total_bytes += s.length();
         if (total_bytes > max_message_size) {
            skipRestOfMessage(in);
            throw new Failed(MessageFormat.format(MESSAGE_TOO_LONG,
                                                  new Object[] { new Integer(max_message_size) }));
         }
         s = in.readLine();
      }

      if (from == null || !from.equals(sender)) {
         String sender_field = "Sender: " + sender;
         if (sender_index < 0) {
            header.addElement(sender_field);
         } else {
            header.setElementAt(sender_field, sender_index);
         }
      }
   }

   private Thread m_thread;
   private Hashtable m_controls;
   private ThreadPool m_messagePool;

   private static final long   POST_THREAD_DELAY = 25;
   private static final long   CLIENT_POST_DELAY = 1500;
   private static final int    MAX_POOL_SIZE     = 100;
   private static final int    START_ID_CHAR     = (int)'<';
   private static final int    END_ID_CHAR       = (int)'>';
   private static final String END_OF_TEXT       = ".";

   private static final String CANCEL_COMMAND    = "cancel";
   private static final String NEWGROUP_COMMAND  = "newgroup";
   private static final String RMGROUP_COMMAND   = "rmgroup";

   private static final String INVALID_HEADER    = "invalid header format";
   private static final String MESSAGE_TOO_LONG  = "article too long: max length is {0} characters";
}
