///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.coffee_server.services.nntp.protocol;

import java.io.*;
import java.net.*;
import java.util.Date;
import org.cooldevtools.utility.*;
import org.cooldevtools.coffee_server.services.nntp.pub.*;

/**
   This class implements the client side of the NNTP protocol including
   the administration extensions used by the CoffeeLink News Server
   configuration client.
 */
public class ServerConnection
{
   public ServerConnection()
   {
   }

   /**
      Closes the connection to the server if any.
    */
   public void reset()
   throws IOException
   {
      if (m_socket != null) {
         Socket sock = m_socket;
         m_socket = null;
         m_reader = null;
         m_writer = null;
         sock.close();
      }
   }

   public String getUserName()
   {
      return m_username;
   }

   /**
      Sets the host and port number used to connect to the server.
    */
   public void setHost(String host,
                       int port)
   {
      m_host = host;
      m_port = port;
   }

   /**
      Sets the user's username and password for authentication
      with the server.  This is used whenever AUTHINFO USER
      authentication is needed.
    */
   public void setIdentity(String username,
                           String password)
   {
      m_username = username;
      m_password = password;
   }

   /**
      Closes any current connection and then attempts to connect to
      the NNTP server.  If a newsgroup has been previously selected
      then an attempt is made to open that group with an NNTP GROUP
      request.
   */
   public void connect()
   throws IOException
   {
      reset();
      m_socket = new Socket(m_host, m_port);
      try {
         m_reader = new NntpReader(m_socket.getInputStream());
         m_writer = new NntpWriter(m_socket.getOutputStream());
         m_reader.confirmResponse(CODES_MODE_READER);
         login();

         if (m_group != null) {
            sendCommandImpl("GROUP " + m_group, GROUP_FOUND, null);
         }
      } catch (IOException ex) {
         reset();
         throw ex;
      }
   }

   /**
      Simple class to hold information about the server's address
      configuration.  Includes the address (hostname or IP address)
      and the primary and alternate port numbers used by the server.
    */
   public static class AddressInfo
   {
      public AddressInfo(String address, int port, int alt_port)
      {
	 m_address = address;
	 m_port    = port;
	 m_altPort = alt_port;
      }

      public boolean hasAddress()
      {
	 return m_address.length() > 0;
      }

      public String getAddress()
      {
	 return m_address;
      }

      public int getPrimaryPort()
      {
	 return m_port;
      }

      public int getAlternatePort()
      {
	 return m_altPort;
      }

      private String m_address;
      private int m_port;
      private int m_altPort;
   }

   private void login()
   throws IOException
   {
      modeReader();
      if (m_username != null) {
         m_writer.writeEscapedText("AUTHINFO USER " + m_username);
         m_writer.flush();
         if (m_reader.confirmResponse(AUTHINFO_USER_CODES) == CODE_AUTH_SEND_PASSWORD) {
            m_writer.writeEscapedText("AUTHINFO PASS " + m_password);
            m_writer.flush();
	    m_reader.confirmResponse(CODE_AUTH_ACCEPTED);
         }
      }
   }

   private void sendCommandImpl(String command)
   throws IOException
   {
      if (m_socket == null) {
         connect();
      }
      try {
         m_writer.writeEscapedText(command);
         m_writer.flush();
      } catch (IOException ex) {
         ex.printStackTrace();
         connect();
         m_writer.writeEscapedText(command);
         m_writer.flush();
      }
   }

   private void sendCommandImpl(String command,
                                int response,
                                String[] response_line)
   throws IOException
   {
      sendCommandImpl(command);
      if (response > 0) {
         m_reader.confirmResponse(response, response_line);
      }
   }

   private int sendCommandImpl(String command,
                               int[] responses,
                               String[] response_line)
   throws IOException
   {
      sendCommandImpl(command);
      return m_reader.confirmResponse(responses, response_line);
   }

   private String[] getListImpl(String command,
                                int response)
   throws IOException
   {
      sendCommandImpl(command, response, null);
      return m_reader.readMultipleLines();
   }

   private String[] getList(String command,
                            int response)
   throws IOException
   {
      try {
         return getListImpl(command, response);
      } catch (IOException ex) {
         connect();
         return getListImpl(command, response);
      }
   }

   private void sendListImpl(String command,
                             int pre_response,
                             String[] list,
                             int post_response)
   throws IOException
   {
      sendCommandImpl(command, pre_response, null);

      m_writer.writeText(list);
      m_writer.writeEndOfText();
      m_writer.flush();

      if (post_response > 0) {
         m_reader.confirmResponse(post_response);
      }
   }

   private void sendList(String command,
                         int pre_response,
                         String[] list,
                         int post_response)
   throws IOException
   {
      try {
         sendListImpl(command, pre_response, list, post_response);
      } catch (IOException ex) {
         connect();
         sendListImpl(command, pre_response, list, post_response);
      }
   }

   private void sendList(String command,
                         String[] list,
                         int response)
   throws IOException
   {
      try {
         sendListImpl(command, -1, list, response);
      } catch (IOException ex) {
         connect();
         sendListImpl(command, -1, list, response);
      }
   }

   private void sendCommand(String command,
                            int response,
                            String[] response_message)
   throws IOException
   {
      try {
         sendCommandImpl(command, response, response_message);
      } catch (IOException ex) {
         connect();
         sendCommandImpl(command, response, response_message);
      }
   }

   private int sendCommand(String command,
                           int[] responses,
                           String[] response_message)
   throws IOException
   {
      try {
         return sendCommandImpl(command, responses, response_message);
      } catch (IOException ex) {
         connect();
         return sendCommandImpl(command, responses, response_message);
      }
   }

   /**
      Breaks line into up to parts.length parts delimited by spaces.
      If the line contains more than parts.length parts then the
      last element in the array will contain the remainder of the line
      following the parts.length-1 part.
      If the line contains fewer than parts.length parts then the elements
      in parts after the last part will be set to empty string.
      Returns the first part found.
    */
   private String parseLine(String line,
                            String[] parts)
   {
      int part_index = 0;
      int offset = 0;
      while (true) {
	 if (part_index == parts.length - 1) {
	    parts[part_index++] = line.substring(offset);
	    break;
	 }

	 int space_index = line.indexOf(' ', offset);
	 if (space_index < 0) {
	    parts[part_index++] = line.substring(offset);
	    break;
	 }

	 parts[part_index++] = line.substring(offset, space_index);
	 offset = space_index + 1;
      }

      while (part_index < parts.length) {
	 parts[part_index++] = "";
      }

      return parts[0];
   }

   /**
      Returns the list of all newsgroups on the server.
    */
   public String[] getGroupNames()
   throws IOException
   {
      String[] groups = getList("LIST NEWSGROUPS", 215);

      String[] parts = new String[2];
      for (int i = 0; i < groups.length; i++) {
         groups[i] = parseLine(groups[i], parts);
      }

      return groups;
   }

   /**
      Returns the list of all user names known by the server.
    */
   public String[] getUserNames()
   throws IOException
   {
      return getList("BCC_LIST_USERS", CONFIG_MORE);
   }

   /**
      Returns the list of all roles known by the server.
    */
   public String[] getRoleNames()
   throws IOException
   {
      return getList("BCC_LIST_ROLES", CONFIG_MORE);
   }

   /**
      Returns the password, email address and roles for the specified user.
    */
   public void getUser(String username,
                       StringBuffer password,
                       StringBuffer email,
                       SuperVector roles)
   throws IOException
   {
      String[] data = getList("BCC_LIST_USER " + username, CONFIG_MORE);
      String[] parts = new String[2];
      for (int i = 0; i < data.length; i++) {
         parseLine(data[i], parts);
         String command = parts[0].toLowerCase();
         if (command.equals("email")) {
            email.append(parts[1]);
         } else if (command.equals("password")) {
            password.append(parts[1]);
         } else if (command.equals("role")) {
            roles.addElement(parts[1]);
         }
      }
   }

   /**
      Returns the email address and description for a group.
    */
   public void getNewsGroup(String name,
                            StringBuffer email,
                            StringBuffer description)
   throws IOException
   {
      String[] data = getList("BCC_LIST_GROUP " + name, CONFIG_MORE);
      String[] parts = new String[2];
      for (int i = 0; i < data.length; i++) {
         parseLine(data[i], parts);
         String command = parts[0].toLowerCase();
         if (command.equals("email")) {
            email.append(parts[1]);
         } else if (command.equals("description")) {
            description.append(parts[1]);
         }
      }
   }

   /**
      Creates a new user with the specified username, password, email address,
      and roles.
    */
   public void defineUser(String username,
                          String password,
                          String email,
                          String[] roles)
   throws IOException
   {
      String[] text = new String[2 + roles.length];
      text[0] = "PASSWORD " + password;
      text[1] = "EMAIL " + email;
      for (int i = 0; i < roles.length; i++) {
         text[i+2] = "ROLE " + roles[i];
      }
      sendList("BCC_DEFINE_USER " + username, text, CONFIG_OK);
   }

   /**
      Creates a new news group with the specified name, email address,
      and description.
    */
   public void defineNewsGroup(String name,
                               String email,
                               String description)
   throws IOException
   {
      String[] text = new String[2];
      text[0] = "DESCRIPTION " + description;
      text[1] = "EMAIL " + email;
      sendList("BCC_DEFINE_GROUP " + name, text, CONFIG_OK);
   }

   /**
      Simple class to hold the information about a single role.
    */
   public static class RoleInfo
   {
      public String name;
      public String admin_level;
      public String[] groups;
      public String[] group_perms;
      public String[] user_roles;
   }

   /**
      Returns the information about a role.
    */
   public RoleInfo getRole(String role_name)
      throws IOException
   {
      String[] data = getList("BCC_LIST_ROLE " + role_name, CONFIG_MORE);
      String[] parts = new String[3];

      SuperVector groups      = new SuperVector();
      SuperVector group_perms = new SuperVector();
      SuperVector user_roles  = new SuperVector();

      RoleInfo role = new RoleInfo();
      role.name = role_name;

      for (int i = 0; i < data.length; i++) {
         parseLine(data[i], parts);
         String command = parts[0].toLowerCase();
         if (command.equals("admin")) {
            role.admin_level = parts[1];
         } else if (command.equals("role")) {
            user_roles.addElement(parts[1]);
         } else if (command.equals("group")) {
            groups.addElement(parts[1]);
            group_perms.addElement(parts[2]);
         }
      }

      role.groups = groups.getStrings();
      role.group_perms = group_perms.getStrings();
      role.user_roles = user_roles.getStrings();

      return role;
   }

   /**
      Defines a new role.
    */
   public void defineRole(RoleInfo role)
      throws IOException
   {
      if (Assert.DEBUG) Assert.pre(role.groups.length == role.group_perms.length);

      String[] text = new String[1 + role.groups.length + role.user_roles.length];

      int index = 0;
      text[index++] = "ADMIN " + role.admin_level;

      for (int i = 0; i < role.groups.length; ++i) {
	 text[index++] = "GROUP " + role.groups[i] + " " + role.group_perms[i];
      }

      for (int i = 0; i < role.user_roles.length; ++i) {
	 text[index++] = "ROLE " + role.user_roles[i];
      }

      if (Assert.DEBUG) Assert.post(index == text.length);

      sendList("BCC_DEFINE_ROLE " + role.name, text, CONFIG_OK);
   }

   public void deleteRole(String username)
   throws IOException
   {
      sendCommand("BCC_DELETE_ROLE " + username, CONFIG_OK, null);
   }

   public void deleteUser(String username)
   throws IOException
   {
      sendCommand("BCC_DELETE_USER " + username, CONFIG_OK, null);
   }

   public void deleteNewsGroup(String name)
   throws IOException
   {
      sendCommand("BCC_DELETE_GROUP " + name, CONFIG_OK, null);
   }

   public void shutdown()
   throws IOException
   {
      sendCommand("bcc_shutdown", 111, null);
   }

   /**
      Instructs the server to save the current configuration to disk.
    */
   public void saveConfig()
   throws IOException
   {
      sendCommand("BCC_SAVE_CONFIG", CONFIG_OK, null);
   }

   public void postMessage(String[] message)
   throws IOException
   {
      sendList("POST", 340, message, 240);
   }

   public String[] readMessage(String message_id)
   throws IOException
   {
      return getList("ARTICLE " + message_id, 220);
   }

   public String[] newNews(String group_name,
                           String date_time)
   throws IOException
   {
      String command = "NEWNEWS " + group_name;
      if (date_time != null) {
         command += " " + date_time;
      }
      return getList(command, 230);
   }

   public String[] newNews(String group_name,
                           Date when)
   throws IOException
   {
      return newNews(group_name, (when == null) ? null : DateUtils.dateToString(when));
   }

   private String getMessageId(String command,
                               int[] ok_codes,
                               int found_code,
                               long[] message_number)
   throws IOException
   {
      String[] parts = new String[3];
      int response = sendCommand(command, ok_codes, parts);
      if (response != found_code) {
         return null;
      }

      String num_str = parts[0];
      String id_str  = parts[1];
      if (message_number != null) {
         message_number[0] = Long.parseLong(num_str);
      }
      return id_str;
   }

   public String stat(String message_id,
                      long[] message_number)
   throws IOException
   {
      return getMessageId("STAT " + message_id, CODES_STAT, CODE_STAT_FOUND, message_number);
   }

   public String stat()
   throws IOException
   {
      return getMessageId("STAT", CODES_STAT, CODE_STAT_FOUND, null);
   }

   public String stat(long message_number)
   throws IOException
   {
      return getMessageId("STAT " + message_number, CODES_STAT, CODE_STAT_FOUND, null);
   }

   public String nextMessageId()
   throws IOException
   {
      return getMessageId("NEXT", CODES_NEXT, CODE_NEXT_FOUND, null);
   }

   public boolean ihaveQuery(String message_id)
   throws IOException
   {
      int response = sendCommand("IHAVE " + message_id, CODES_IHAVE_QUERY, null);
      return response == CODE_IHAVE_SEND;
   }

   public void ihaveSend(String[] message)
   throws IOException
   {
      m_writer.writeText(message);
      m_writer.writeEndOfText();
      m_writer.flush();
      m_reader.confirmResponse(CODES_IHAVE_SEND);
   }

   public boolean setGroup(String group)
   throws IOException
   {
      if (group == null) {
         m_group = null;
         return true;
      }

      sendCommandImpl("GROUP " + group);
      if (m_reader.confirmResponse(GROUP_CODES) == GROUP_UNKNOWN) {
         return false;
      }

      m_group = group;
      return true;
   }

   public boolean setGroup(String group,
                           long[] counters)
   throws IOException
   {
      if (group == null) {
         m_group = null;
         return true;
      }

      int parts_len = (counters == null) ? 1 : counters.length;
      String[] parts = new String[parts_len + 1];
      sendCommandImpl("GROUP " + group);
      if (m_reader.confirmResponse(GROUP_CODES, parts) == GROUP_UNKNOWN) {
         return false;
      }

      m_group = group;

      if (counters != null) {
         for (int i = 0; i < counters.length; i++) {
            counters[i] = Long.parseLong(parts[i]);
         }
      }

      return true;
   }

   public int getMaxConnections()
   throws IOException
   {
      String[] answer = new String[1];
      sendCommand("BCC_GET_MAX_CONNECT", CONFIG_OK, answer);
      return Integer.parseInt(answer[0]);
   }

   public int getMaxMessageSize()
   throws IOException
   {
      String[] answer = new String[1];
      sendCommand("BCC_GET_MAX_MESSAGE_SIZE", CONFIG_OK, answer);
      return Integer.parseInt(answer[0]);
   }

   public int getMaxMessageDays()
   throws IOException
   {
      String[] answer = new String[1];
      sendCommand("BCC_GET_MAX_MESSAGE_DAYS", CONFIG_OK, answer);
      return Integer.parseInt(answer[0]);
   }

   public String[] getUnixIdentity()
   throws IOException
   {
      String[] answer = new String[2];
      sendCommand("BCC_GET_UNIX_IDENTITY", CONFIG_OK, answer);
      return answer;
   }

   public AddressInfo getAddressInfo()
   throws IOException
   {
      String[] answer = new String[3];
      sendCommand("BCC_GET_PORTS", CONFIG_OK, answer);
      return new AddressInfo(answer[2], Integer.parseInt(answer[0]), Integer.parseInt(answer[1]));
   }

   public void setMaxConnections(int value)
   throws IOException
   {
      sendCommand("BCC_SET_MAX_CONNECT " + value, CONFIG_OK, null);
   }

   public void setMaxMessageSize(int value)
   throws IOException
   {
      sendCommand("BCC_SET_MAX_MESSAGE_SIZE " + value, CONFIG_OK, null);
   }

   public void setMaxMessageDays(int value)
   throws IOException
   {
      sendCommand("BCC_SET_MAX_MESSAGE_DAYS " + value, CONFIG_OK, null);
   }

   public void setAddressInfo(AddressInfo addr)
   throws IOException
   {
      StringBuffer command = new StringBuffer();
      command.append("BCC_SET_PORTS ");
      command.append(addr.getPrimaryPort());
      command.append(" ");
      command.append(addr.getAlternatePort());
      if (addr.hasAddress()) {
	 command.append(" ");
	 command.append(addr.getAddress());
      }
      sendCommand(command.toString(), CONFIG_OK, null);
   }

   public void setUnixIdentity(String user,
                               String group)
   throws IOException
   {
      sendCommand("BCC_SET_UNIX_IDENTITY " + user + " " + group, CONFIG_OK, null);
   }

   public void setLogging(int log_code,
                          String log_param)
   throws IOException
   {
      sendCommand("BCC_SET_LOGGING " + LOGGING_TYPES[log_code] + " " + log_param, CONFIG_OK, null);
   }

   public void setLogLevel(String level_code)
   throws IOException
   {
      sendCommand("BCC_SET_LOG_LEVEL " + level_code, CONFIG_OK, null);
   }

   public int getLogging(String[] log_param)
   throws IOException
   {
      String[] answer = new String[2];
      sendCommand("BCC_GET_LOGGING", CONFIG_OK, answer);
      log_param[0] = answer[1];
      String log_type = answer[0].toLowerCase();
      for (int i = 0; i < LOGGING_TYPES.length; i++) {
         if (log_type.equals(LOGGING_TYPES[i])) {
            return i;
         }
      }
      log_param[0] = "-";
      return LOG_TO_FILE;
   }

   public String getLogLevel()
   throws IOException
   {
      String[] answer = new String[2];
      sendCommand("BCC_GET_LOG_LEVEL", CONFIG_OK, answer);
      return answer[0];
   }

   public void modeReader()
   throws IOException
   {
      sendCommand("MODE READER", CODES_MODE_READER, null);
   }

   public void quit()
   throws IOException
   {
      sendCommand("QUIT", 205, null);
      reset();
   }

   public String[] listExtensions()
   throws IOException
   {
      String[] answer = getList("LIST EXTENSIONS", CODE_LIST_EXTENSIONS);
      for (int i = 0; i < answer.length; i++) {
         // strip the leading space required by the standard
         if (answer[i].charAt(0) == ' ') {
            answer[i] = answer[i].substring(1);
         }
      }
      return answer;
   }

   public String[] listActive()
   throws IOException
   {
      return getList("LIST ACTIVE", CODE_LIST_ACTIVE);
   }

   public String[] listNewsGroups(String wild_mat)
   throws IOException
   {
      if (wild_mat == null) {
         return getList("LIST NEWSGROUPS", CODE_LIST_NEWSGROUPS);
      } else {
         return getList("LIST NEWSGROUPS " + wild_mat, CODE_LIST_NEWSGROUPS);
      }
   }

   public String[] listNewsGroupNames(String wild_mat)
   throws IOException
   {
      String[] groups = listNewsGroups(wild_mat);
      for (int i = 0; i < groups.length; i++) {
         int space_index = groups[i].indexOf("\t");
         if (space_index == -1) {
            space_index = groups[i].indexOf(" ");
         }
         groups[i] = groups[i].substring(0, space_index);
      }
      return groups;
   }

   public long[] listGroup()
   throws IOException
   {
      String[] numbers = getList("LISTGROUP", CODE_LISTGROUP);
      long[] answer = new long[numbers.length];
      for (int i = 0; i < answer.length; i++) {
         answer[i] = Long.parseLong(numbers[i]);
      }
      return answer;
   }

   public String[] next()
   throws IOException
   {
      int response = sendCommand("NEXT", CODES_NEXT, null);
      if (response != CODE_NEXT_SELECTED) {
         return null;
      }

      return article();
   }

   public String[] xover()
   throws IOException
   {
      int response = sendCommand("XOVER", CODES_XOVER, null);
      if (response != CODE_XOVER) {
         return null;
      }

      String[] lines = m_reader.readMultipleLines();
      if (lines.length == 0) {
	 return null;
      }
      return Misc.splitString(lines[0], '\t');
   }

   public String[] parseXOVER(String xover_string)
   {
      return Misc.splitString(xover_string, '\t');
   }

   public String[] article(long article_number)
   throws IOException
   {
      int response = sendCommand("ARTICLE " + article_number, CODES_ARTICLE, null);
      if (response != CODE_ARTICLE_FOLLOWS) {
         return null;
      }

      return m_reader.readMultipleLines();
   }

   public String[] article()
   throws IOException
   {
      int response = sendCommand("ARTICLE", CODES_ARTICLE, null);
      if (response != CODE_ARTICLE_FOLLOWS) {
         return null;
      }

      return m_reader.readMultipleLines();
   }

   public String[] head(long article_number)
   throws IOException
   {
      int response = sendCommand("HEAD " + article_number, CODES_HEAD, null);
      if (response != CODE_HEAD_FOLLOWS) {
         return null;
      }

      return m_reader.readMultipleLines();
   }

   public String[] body(long article_number)
   throws IOException
   {
      int response = sendCommand("BODY " + article_number, CODES_BODY, null);
      if (response != CODE_BODY_FOLLOWS) {
         return null;
      }

      return m_reader.readMultipleLines();
   }

   public void post(String[] message)
   throws IOException
   {
      sendList("POST", CODE_POST_MORE, message, CODE_POST);
   }

   public void cancel(String message_id,
                      String group_name,
                      String user_email)
   throws IOException
   {
      String[] message = {
         "From: " + user_email,
         "Subject: cancel " + message_id,
         "Newsgroups: " + group_name,
         "Control: cancel " + message_id,
         ""
      };
      post(message);
   }

   public void cancel(String message_id)
   throws IOException
   {
      String[] message = {
         "From: anonymous@nowhere.com",
         "Subject: cancel " + message_id,
         "Newsgroups: no.group.specified",
         "Control: cancel " + message_id,
         ""
      };
      post(message);
   }

   private String m_host;
   private int    m_port;
   private String m_username;
   private String m_password;
   private String m_group;

   private Socket m_socket;
   private NntpReader m_reader;
   private NntpWriter m_writer;

   static final int CODE_AUTH_CAN_POST      = 200;
   static final int CODE_AUTH_CAN_READ      = 201;
   static final int CODE_AUTH_ACCEPTED      = 281;
   static final int CODE_AUTH_SEND_PASSWORD = 381;
   static final int CODE_AUTH_REQUIRED      = 480;
   static final int CODE_AUTH_REJECTED      = 502;
   static final int CODE_PERM_DENIED        = 502;
   static final int CODE_AUTH_SYNTAX        = 482;
   private static final int[] AUTHINFO_USER_CODES = {
      CODE_AUTH_ACCEPTED, CODE_AUTH_SEND_PASSWORD
   };
   private static final int[] GROUP_CODES = {
      211, 411
   };
   private static final int GROUP_FOUND = 211;
   private static final int GROUP_UNKNOWN = 411;
   private static final int CONFIG_OK = NntpCommand.BCC_OK;
   private static final int CONFIG_MORE = NntpCommand.BCC_MORE_FOLLOWS;

   private static final String[] LOGGING_TYPES = {
      "file", "syslog"
   };
   public static final int LOG_TO_FILE    = 0;
   public static final int LOG_TO_SYSLOG  = 1;
   static final int[] CODES_MODE_READER = {
      ServerConnection.CODE_AUTH_CAN_POST,
      ServerConnection.CODE_AUTH_CAN_READ,
      ServerConnection.CODE_AUTH_REQUIRED
   };

   static final int[] CODES_ARTICLE = { 220, 423, 430 };
   static final int[] CODES_HEAD = { 221, 423, 430 };
   static final int[] CODES_BODY = { 222, 423, 430 };
   static final int[] CODES_NEXT = { 223, 421 };
   static final int[] CODES_STAT = { 223, 423, 430 };
   static final int[] CODES_IHAVE_QUERY = { 335, 435, 437 };
   static final int[] CODES_IHAVE_SEND = { 235, 436, 437 };
   static final int[] CODES_XOVER = { 224, 412, 420 };

   static final int CODE_IHAVE_SEND = 335;
   static final int CODE_STAT_FOUND = 223;
   static final int CODE_NEXT_FOUND = 223;
   static final int CODE_POST = 240;
   static final int CODE_POST_MORE = 340;
   static final int CODE_LIST_EXTENSIONS = 202;
   static final int CODE_LIST_ACTIVE = 215;
   static final int CODE_LISTGROUP = 211;
   static final int CODE_GROUP = 211;
   static final int CODE_ARTICLE_FOLLOWS = 220;
   static final int CODE_HEAD_FOLLOWS    = 221;
   static final int CODE_BODY_FOLLOWS    = 222;
   static final int CODE_NEXT_SELECTED   = 223;
   static final int CODE_LIST_NEWSGROUPS = 215;
   static final int CODE_XOVER           = 224;

   public static final int XOVER_ARTICLE_NUMBER      = 0;
   public static final int XOVER_ARTICLE_SUBJECT     = 1;
   public static final int XOVER_ARTICLE_AUTHOR      = 2;
   public static final int XOVER_ARTICLE_DATE        = 3;
   public static final int XOVER_ARTICLE_MESSAGE_ID  = 4;
   public static final int XOVER_ARTICLE_REFERENCES  = 5;
   public static final int XOVER_ARTICLE_BYTE_COUNT  = 6;
   public static final int XOVER_ARTICLE_LINE_COUNT  = 7;
}
