///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//


package org.cooldevtools.coffee_server.services.nntp.pub;

import java.io.*;
import java.util.Vector;
import org.cooldevtools.utility.Assert;

public class NntpReader
{
   public NntpReader(InputStream s)
   {
      m_input = s;
      m_eof = false;
      m_line = null;
      m_offset = -1;
      m_length = -1;
   }

   public void close()
   throws IOException
   {
      m_input.close();
   }

   public void readNewLine()
   throws IOException
   {
      try {
         m_line = readLine();
         m_offset = 0;
         m_length = m_line.length();
      } catch (EOFException ex) {
         m_line = null;
         m_offset = -1;
         m_length = -1;
         throw ex;
      }
   }

   public String getNextCommand()
   throws EOFException
   {
      return getNextCommand(true);
   }

   public String getNextArgument()
   throws EOFException
   {
      return getNextCommand(false);
   }

   public String getNextCommand(boolean as_lower)
   throws EOFException
   {
      if (m_line == null) {
         throw new EOFException();
      }

      if (m_offset == m_length) {
         return EMPTY_STRING;
      }

      int i = m_offset;
      while (i < m_length) {
         char ch = m_line.charAt(i);
         if (ch == ' ' || ch == '\t') {
            break;
         }
         i++;
      }

      String command = m_line.substring(m_offset, i);

      while (i < m_length) {
         char ch = m_line.charAt(i);
         if (ch != ' ' && ch != '\t') {
            break;
         }
         i++;
      }

      m_offset = i;

      return as_lower ? command.toLowerCase() : command;
   }

   public String getEntireLine()
   throws EOFException
   {
      if (m_line == null) {
         throw new EOFException();
      }

      return m_line;
   }

   public String getRemainderOfLine()
   throws EOFException
   {
      if (m_line == null) {
         throw new EOFException();
      }

      if (m_offset == m_length) {
         return EMPTY_STRING;
      }

      return m_line.substring(m_offset);
   }

   public String readLine()
   throws IOException
   {
      String s = readLineImpl();
      if (DEBUG) System.out.println("=" + s + "=");
      return s;
   }

   public String readLineImpl()
   throws IOException
   {
      if (m_eof) {
         throw new EOFException();
      }

      StringBuffer buffer = new StringBuffer();
      boolean have_cr = false;
      while (true) {
         int ch;

         try {
            ch = m_input.read();
         } catch (InterruptedIOException ex) {
            // Socket has timed out during read.
            // Simulate end of file so we'll drop the client.
            ch = -1;
         }

         if (ch == -1) {
            m_eof = true;
            if (buffer.length() > 0) {
               return buffer.toString();
            } else {
               throw new EOFException();
            }
         }

         // We shouldn't need this according to the protocol spec
         // but GNUS only sends <LF> instead of <CR><LF>!
         if (ch == LF) {
            return buffer.toString();
         }

         // Detect <CR><LF>
         if (have_cr) {
            if (ch == LF) {
               return buffer.toString();
            } else {
               buffer.append((char)CR);
               have_cr = (ch == CR);
            }
         } else {
            if (ch == CR) {
               have_cr = true;
            } else {
               buffer.append((char)ch);
            }
         }
      }
   }

   public String[] readMultipleLines()
   throws IOException
   {
      Vector v = new Vector();
      readNewLine();
      while (!m_line.equals(MULTI_LINE_TERM)) {
         v.addElement(m_line);
         readNewLine();
      }

      String[] answer = new String[v.size()];
      v.copyInto(answer);
      return answer;
   }

   private void fillResponseText(String[] response_text)
   throws IOException
   {
      if (response_text == null || response_text.length == 0) {
         return;
      }

      int last = response_text.length - 1;
      for (int i = 0; i < last; i++) {
         response_text[i] = getNextCommand();
      }
      response_text[last] = getRemainderOfLine();
   }

   public void confirmResponse(int expected,
                               String[] response_text)
   throws IOException
   {
      readNewLine();
      try {
         int code = Integer.parseInt(getNextCommand());
         fillResponseText(response_text);
         if (code != expected) {
            throw new UnexpectedNntpResponseException(code, getRemainderOfLine());
         }
      } catch (NumberFormatException ex) {
         throw new IOException(ex.getMessage());
      }
   }

   public int confirmResponse(int[] expected,
                              String[] response_text)
   throws IOException
   {
      readNewLine();
      try {
         int code = Integer.parseInt(getNextCommand());
         fillResponseText(response_text);
         for (int i = 0; i < expected.length; i++) {
            if (code == expected[i]) {
               return code;
            }
         }
         throw new UnexpectedNntpResponseException(code, getRemainderOfLine());
      } catch (NumberFormatException ex) {
         throw new UnexpectedNntpResponseException(getEntireLine());
      }
   }

   public void confirmResponse(int expected)
   throws IOException
   {
      confirmResponse(expected, null);
   }

   public int confirmResponse(int[] expected)
   throws IOException
   {
      return confirmResponse(expected, null);
   }

   private InputStream m_input;
   private String m_line;
   private int m_offset;
   private int m_length;
   private boolean m_eof;

   private static int CR = 13;
   private static int LF = 10;

   public static String MULTI_LINE_TERM = ".";

   private static final boolean DEBUG = false;
   private static final String EMPTY_STRING = "";
}
