///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//


package org.cooldevtools.coffee_server.services.thttp;

import java.io.*;
import java.net.*;
import java.awt.*;
import java.text.*;
import java.util.*;
import javax.swing.*;
import org.cooldevtools.hdl.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.coffee_server.config.*;
import org.cooldevtools.coffee_server.framework.*;

public class TrivialHttpService extends BasicCoffeeServiceManager implements CoffeeService
{
   public TrivialHttpService()
   throws IOException
   {
      this(DEFAULT_DIRECTORY);
   }

   public TrivialHttpService(String base_dir)
      throws IOException
   {
      m_root = new File(base_dir);
      if (!m_root.exists() || !m_root.isDirectory()) {
	 throw new FileNotFoundException(base_dir);
      }
      m_root = new File(m_root.getCanonicalPath());
   }

   public void configure(ServiceConfigInfo config)
      throws CoffeeServiceException
   {
      super.configure(config);
   }

   public void serviceRequest(ClientRequest req)
      throws IOException, CoffeeServiceException
   {
      try {
	 BufferedReader in = new BufferedReader(new InputStreamReader(req.getInputStream()));
	 try {
	    BufferedWriter out = new BufferedWriter(new OutputStreamWriter(req.getOutputStream()));
	    try {
	       //
	       // Read the request and make sure we can support sender's version.
	       //
	       String request_text = in.readLine();
	       String request_upper = request_text.toUpperCase();
	       if (TRACE_ON) trace("REQUEST: ", request_text);
	       if (!request_upper.startsWith("GET ")) {
		  error(out, "501 Not implemented.");
		  return;
	       }
	       if (!request_upper.endsWith(" HTTP/1." + request_upper.charAt(request_upper.length() - 1))) {
		  error(out, "400 Incompatible protocol version.");
		  return;
	       }

	       //
	       // Skip all of the headers.
	       //
	       for (String hdr = in.readLine(); hdr.length() > 0; hdr = in.readLine()) {
		  if (TRACE_ON) trace("HEADER: ", hdr);
	       }

	       //
	       // Extract the URI and verify it fits our supported subset and convert
	       // to valid platform path with no leading /.
	       //
	       String uri = request_text.substring(4, request_text.length() - 9).trim();
	       if (TRACE_ON) trace("URI: ", uri);
	       if (!isValidURI(uri)) {
		  error(out, "400 Requested URI not recognized.");
		  return;
	       }
	       if (uri.length() == 1) {
		  uri = "index.html";
	       } else {
		  uri = uri.substring(1).replace('/', File.separatorChar);
	       }
	       if (TRACE_ON) trace("URI: ", uri);

	       //
	       // Verify file exists and is readable.
	       //
	       File f = new File(m_root, uri);
	       if (f.isDirectory() && uri.endsWith(File.separator)) {
		  f = new File(f, "index.html");
	       }
	       if (TRACE_ON) trace("FILE: ", f.getPath());
	       if (TRACE_ON) trace("PATH: ", f.getCanonicalPath());
	       if (!f.isFile()) {
		  error(out, "404 File not found.");
		  return;
	       }
	       if (!f.canRead()) {
		  error(out, "403 File unreadable.");
		  return;
	       }

	       //
	       // Sanity checks to ensure we don't serve a file outside the root's tree.
	       //
	       f = new File(f.getCanonicalPath());
	       if (TRACE_ON) trace("DEPTH: ", String.valueOf(determineDepth(uri)));
	       if (TRACE_ON) trace("CHILD? ", isChildOfRoot(f) ? "yes" : "no");
	       if (Assert.DEBUG) Assert.inv((determineDepth(uri) >= 0) == isChildOfRoot(f));
	       if (determineDepth(uri) < 0 || !isChildOfRoot(f)) {
		  error(out, "403 Access not permitted.");
		  return;
	       }

	       //
	       // Generate response header including MIME type if we recognize file suffix
	       // then send file in binary format.
	       //
	       out.write("HTTP/1.0 200 OK\r\n");
	       String mime_type = getMIMEType(uri);
	       if (mime_type != null) {
		  out.write("Content-type: " + mime_type + "\r\n");
	       }
	       out.write("Content-length: " + f.length() + "\r\n");
	       out.write("\r\n");
	       sendFile(out, f);
	    } catch (IOException ex) {
	       if (Assert.DEBUG) ex.printStackTrace();
	    } finally {
	       out.flush();
	       out.close();
	    }
	 } catch (IOException ex) {
	    if (Assert.DEBUG) ex.printStackTrace();
	 } finally {
	    in.close();
	 }
      } finally {
	 req.close();
      }
   }

   private String getMIMEType(String uri)
   {
      int suffix_start = uri.lastIndexOf(".");
      String mime_type = null;
      if (suffix_start > 0) {
	 String suffix = uri.substring(suffix_start + 1).toLowerCase();
	 if (suffix.equals("html")) {
	    mime_type = "text/html";
	 } else if (suffix.equals("gif")) {
	    mime_type = "image/gif";
	 } else if (suffix.equals("jpg")) {
	    mime_type = "image/jpeg";
	 } else if (suffix.equals("txt")) {
	    mime_type = "text/plain";
	 }
      }
      return mime_type;
   }

   /**
      Send the file in binary form.
   */
   private void sendFile(Writer out,
			 File f)
      throws IOException
   {
      FileReader file_in = new FileReader(f);
      try {
	 if (TRACE_ON) trace("WRITING: ", String.valueOf(f.length()) + " bytes");
	 char[] buffer = new char[8192];
	 int len = file_in.read(buffer);
	 while (len > 0) {
	    out.write(buffer, 0, len);
	    len = file_in.read(buffer);
	 }
	 if (TRACE_ON) trace("WRITING: ", "complete");
      } finally {
	 file_in.close();
      }
   }

   private void error(Writer out,
		      String error_text)
      throws IOException
   {
      out.write("HTTP/1.0 " + error_text + "\r\n");
      out.write("Content-Type: text/html\r\n");
      out.write("\r\n");
      out.write("<HTML><HEAD>\r\n");
      out.write("<TITLE>Error Processing HTTP Request</TITLE>\r\n");
      out.write("</HEAD><BODY>\r\n");
      out.write("<H1>" + error_text + "</H1>\r\n");
      out.write("Your browser sent a request that\r\n");
      out.write("this server could not process.<P>\r\n");
      out.write("</BODY></HTML>\r\n");
      if (TRACE_ON) trace("ERROR: ", error_text);
   }

   private boolean isValidURI(String uri)
   {
      char[] chars = uri.toCharArray();
      if (chars[0] != '/') {
	 if (TRACE_ON) trace("REJECT: ", "first character wasn't /");
	 return false;
      }
      for (int i = 0; i < chars.length; i++) {
	 char ch = chars[i];
	 if (!Character.isLetterOrDigit(ch) && (ch != '/') && (ch != '.') && (ch != '_') && (ch != '-')) {
	    if (TRACE_ON) trace("REJECT: ", String.valueOf(ch));
	    return false;
	 }
      }
      return true;
   }

   private void trace(String prefix,
		      String message)
   {
      synchronized (System.out) {
	 System.out.print(Thread.currentThread().getName());
	 System.out.print(": ");
	 System.out.print(prefix);
	 System.out.print(message);
	 System.out.println();
      }
   }

   private int determineDepth(String relative_path)
   {
      String[] dirs = Misc.splitString(relative_path, File.separatorChar);
      int depth = 0;
      for (int i = 0; i < dirs.length - 1; i++) {
	 String d = dirs[i];
	 if (d.equals(".") || d.length() == 0) {
	    // ignored
	 } else if (d.equals("..")) {
	    depth--;
	 } else {
	    depth++;
	 }
      }
      return depth;
   }

   private boolean isChildOfRoot(File f)
   {
      String parent_name = f.getParent();
      while (parent_name != null) {
	 File parent = new File(parent_name);
	 if (parent.equals(m_root)) {
	    return true;
	 }
	 parent_name = parent.getParent();
      }
      return false;
   }

   public CoffeeService newServiceInstance()
   {
      return this;
   }

   public CoffeeServiceConfigBean newConfigBeanInstance()
   {
      return null;
   }

   public boolean hasConfigBean()
   {
      return false;
   }

   public static void main(String[] args)
   {
      try {
	 String root_dir = DEFAULT_DIRECTORY;
	 if (args.length >= 1) {
	    root_dir = args[0];
	 }

         int port = 5250;
         if (args.length >= 2) {
            port = Integer.parseInt(args[1]);
         }

         CoffeeServer server = new CoffeeServer();
         server.addMultiThreadedService(null, port, port, new TrivialHttpService(root_dir));
         server.startup();
      } catch (Exception ex) {
         ex.printStackTrace();
      }
   }

   private File m_root;

   private static final boolean TRACE_ON = true;
   private static final String DEFAULT_DIRECTORY = "http";
}
