///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.hdl;

import java.io.StringReader;
import java.io.Reader;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.StreamTokenizer;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;

public class Parser {
   public Parser()
   {
   }

   public synchronized Statement parse(BufferedReader input)
   throws HDLException, IOException
   {
      StreamTokenizer tokens = createTokenizer(input);

      int tok = tokens.nextToken();
      if (tok == StreamTokenizer.TT_EOF)
         return null;

      Token id = makeId(tokens);

      Statement stmt;

      if (id.equals(Token.TOK_BEGIN))
         stmt = parseBlock(tokens);
      else
         stmt = parseStatement(id, false, tokens);

      return stmt;
   }

   public synchronized Statement parseString(String s)
   throws HDLException, IOException
   {
      m_filename = "<string>";
      BufferedReader in = new BufferedReader(new StringReader(s));
      try {
         return parse(in);
      } finally {
         in.close();
      }
   }

   public synchronized Statement parse(String filename)
   throws HDLException, IOException
   {
      m_filename = filename;
      BufferedReader in = new BufferedReader(new FileReader(filename));
      try {
         return parse(in);
      } finally {
         in.close();
      }
   }

   private Statement parseStatement(Token id,
                                    boolean is_block,
                                    StreamTokenizer tokens)
   throws HDLException, IOException
   {
      if (id.equals(Token.TOK_BEGIN) || id.equals(Token.TOK_END))
         throw new HDLException(id, "invalid statement name");

      Statement stmt = new Statement(id, is_block);
      parseArgumentsForStatement(stmt, tokens);
      return stmt;
   }

   private Statement parseBlock(StreamTokenizer tokens)
   throws HDLException, IOException
   {
      int tok = tokens.nextToken();
      Token id = makeId(tokens);

      Statement stmt = parseStatement(id, true, tokens);

      while (true) {
         tok = tokens.nextToken();
         id = makeId(tokens);

         if (id.equals(Token.TOK_BEGIN)) {
            Statement child = parseBlock(tokens);
            stmt.addChild(child);
         } else if (id.equals(Token.TOK_END)) {
            tok = tokens.nextToken();
            if (tok != Constants.STATEMENT_END_CHAR)
               throwUnexpectedToken(tokens);
            return stmt;
         } else {
            Statement child = parseStatement(id, false, tokens);
            stmt.addChild(child);
         }
      }
   }

   private void parseArgumentsForStatement(Statement stmt,
                                           StreamTokenizer tokens)
   throws HDLException, IOException
   {
      while (true) {
         int tok = tokens.nextToken();

         switch (tok) {
         case Constants.STATEMENT_END_CHAR:
            return;

         case StreamTokenizer.TT_NUMBER:
            stmt.addArgument(Token.intToken((int)tokens.nval, m_filename, tokens.lineno()));
            break;

         case StreamTokenizer.TT_WORD:
            stmt.addArgument(makeId(tokens));
            break;

         case Constants.QUOTE_CHAR:
            stmt.addArgument(Token.strToken(tokens.sval, m_filename, tokens.lineno()));
            break;

         default:
            throwUnexpectedToken(tokens);
         }
      }
   }

   private Token makeId(StreamTokenizer tokens)
   throws HDLException
   {
      if (tokens.ttype != StreamTokenizer.TT_WORD)
         throwUnexpectedToken(tokens);

      return Token.idToken(tokens.sval, m_filename, tokens.lineno());
   }

   private StreamTokenizer createTokenizer(Reader input)
   {
      StreamTokenizer tokens = new StreamTokenizer(input);

      tokens.eolIsSignificant(false);
      tokens.commentChar(Constants.COMMENT_CHAR);
      tokens.quoteChar(Constants.QUOTE_CHAR);
      tokens.slashSlashComments(false);
      tokens.slashStarComments(false);
      tokens.lowerCaseMode(false);
      tokens.parseNumbers();
      tokens.wordChars('_', '_');

      return tokens;
   }

   private void throwUnexpectedToken(StreamTokenizer tokens)
   throws HDLException
   {
      throw new HDLException(m_filename, tokens.lineno(),
                             "unexpected token type (" + tokens.ttype + ")");
   }

   public static void main(String[] args)
   throws Exception
   {
      if (args.length == 0) {
         System.out.println("no filenames on command line");
         System.exit(1);
      }
      Parser p = new Parser();
      PrintWriter w = new PrintWriter(new OutputStreamWriter(System.out));
      for (int i = 0; i < args.length; i++) {
         w.println("# File: " + args[i]);
         Statement stmt = p.parse(args[i]);
         stmt.write(w, "", "    ");
      }
      w.flush();
   }

   private String m_filename;
}
