///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.hdl;

import java.util.Vector;
import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;

public class SyntaxChecker {
   public SyntaxChecker(String syntax_string)
   throws HDLException, IOException
   {
      Parser parser = new Parser();
      Statement syntax = parser.parseString(syntax_string);
      initFromSyntaxTree(syntax);
   }

   public SyntaxChecker(File f)
   throws HDLException, IOException
   {
      Parser parser = new Parser();
      Statement syntax = parser.parse(f.getPath());
      initFromSyntaxTree(syntax);
   }

   public SyntaxChecker(BufferedReader source)
   throws HDLException, IOException
   {
      Parser parser = new Parser();
      Statement syntax = parser.parse(source);
      initFromSyntaxTree(syntax);
   }

   public SyntaxChecker(Statement syntax)
   throws HDLException
   {
      initFromSyntaxTree(syntax);
   }

   public void checkSyntax(Statement stmt)
   throws HDLException
   {
      m_syntax.checkStatement(stmt);
   }

   private int parseArgs(CheckStatement stmt_check,
                         Statement stmt,
                         int index)
   throws HDLException
   {
      if (stmt.numChildren() < 1)
         throw new HDLException(stmt.name(), "expected args child statement");

      Statement args = stmt.child(0);
      if (!args.name().equals(TOK_ARGS))
         throw new HDLException(args.name(), "expected args");

      if (args.numArguments() < 2)
         throw new HDLException(args.name(), "expected at least 2 arguments");

      verifyArgType(args, 0, Token.INTEGER);
      verifyArgType(args, 1, Token.INTEGER);

      int min = args.argument(0).intValue();
      int max = args.argument(1).intValue();
      stmt_check.setArgCounts(min, max);

      int expected = min;
      if (max > min)
         expected++;

      if (args.numArguments() < (expected + 2))
         throw new HDLException(args.name(), "not enough arguments to args statement");

      Vector tables = new Vector();
      for (int i = 2; expected > 0; i++) {
         Token arg = args.argument(i);
         CheckArgument check = new CheckArgument();
         if (arg.equals(TOK_STRING)) {
            check.acceptString();
         } else if (arg.equals(TOK_INT)) {
            check.acceptInteger();
         } else if (arg.equals(TOK_ANYID)) {
            check.acceptId();
         } else if (arg.equals(TOK_ID)) {
            if (args.numArguments() < i)
               throw new HDLException(args.name(), "identifier expected after id");
            i++;
            verifyArgType(args, i, Token.IDENTIFIER);
            check.acceptId(args.argument(i), 0);
         } else if (arg.equals(TOK_TABLE)) {
            tables.addElement(check);
         } else {
            throw new HDLException(arg, "expected string, int, id, or table");
         }
         stmt_check.addArgument(check);
         expected--;
      }

      int expected_children = 1 + tables.size();
      if (stmt.numChildren() < expected_children)
         throw new HDLException(stmt.name(),
                                "insufficient children for specified number of tables");

      for (int i = 1; i < expected_children; i++) {
         CheckArgument table = (CheckArgument)tables.elementAt(i - 1);
         Statement child = stmt.child(i);
         if (!child.name().equals(TOK_TABLE))
            throw new HDLException(child.name(), "expected table");
         parseTable(table, child);
      }

      return expected_children;
   }

   private void parseTable(CheckArgument table,
                           Statement stmt)
   throws HDLException
   {
      int id_index = 0;
      for (int i = 0; i < stmt.numArguments(); i++) {
         Token token = stmt.argument(i);
         if (token.equals(TOK_STRING)) {
            table.acceptString();
         } else if (token.equals(TOK_INT)) {
            table.acceptInteger();
         } else if (token.equals(TOK_ANYID)) {
            table.acceptId();
         } else if (token.equals(TOK_ID)) {
            if (stmt.numArguments() < i)
               throw new HDLException(stmt.name(), "expected identifier after id");
            i++;
            verifyArgType(stmt, i, Token.IDENTIFIER);
            table.acceptId(stmt.argument(i), id_index++);
         } else {
            throw new HDLException(token, "expected string, int, or id");
         }
      }
   }

   private void parseBlock(CheckStatement stmt_check,
                           Statement stmt)
   throws HDLException
   {
      verifyIsBlock(stmt, true);
      if (stmt.numArguments() == 0)
         throw new HDLException(stmt.name(), "at least one argument required");
      verifyArgType(stmt, 0, Token.IDENTIFIER);

      CheckStatement block = new CheckStatement(stmt.argument(0).strValue(), true);
      stmt_check.addChild(block);

      for (int i = 1; i < stmt.numArguments(); i++) {
         Token arg = stmt.argument(i);
         if (arg.equals(TOK_RECURSIVE)) {
            block.addChild(block);
         } else {
            throw new HDLException(arg, "unknown flag for block declaration");
         }
      }

      int next_child = parseArgs(block, stmt, 0);
      while (next_child < stmt.numChildren()) {
         Statement child = stmt.child(next_child);
         Token name = child.name();
         if (name.equals(TOK_STATEMENT)) {
            parseStatement(block, child);
         } else if (name.equals(TOK_BLOCK)) {
            parseBlock(block, child);
         } else if (name.equals(TOK_UNCHECKED)) {
            parseUnchecked(block, child);
         } else {
            throw new HDLException(name, "expected unchecked, block or statement");
         }
         next_child++;
      }

   }

   private void parseStatement(CheckStatement stmt_check,
                               Statement stmt)
   throws HDLException
   {
      verifyIsBlock(stmt, true);
      verifyArgCount(stmt, 1);
      verifyArgType(stmt, 0, Token.IDENTIFIER);

      CheckStatement block = new CheckStatement(stmt.argument(0).strValue(), false);
      stmt_check.addChild(block);

      int next_child = parseArgs(block, stmt, 0);
      if (next_child != stmt.numChildren()) {
         throw new HDLException(stmt.name(),
                                "statement declarations can only contain arg and table statements");
      }
   }

   private void parseUnchecked(CheckStatement stmt_check,
                               Statement stmt)
   throws HDLException
   {
      verifyIsBlock(stmt, false);
      verifyArgCount(stmt, 1);
      verifyArgType(stmt, 0, Token.IDENTIFIER);
      stmt_check.addChild(new CheckUnchecked(stmt.argument(0).strValue()));
   }

   private void initFromSyntaxTree(Statement stmt)
   throws HDLException
   {
      try {
         if (!stmt.name().equals(TOK_LANGUAGE))
            throw new HDLException(stmt.name(), "expected 'language'");

         verifyArgCount(stmt, 1);
         verifyArgType(stmt, 0, Token.IDENTIFIER);

         m_syntax = new CheckStatement(stmt.argument(0).strValue(), true);
         for (int i = 0; i < stmt.numChildren(); i++) {
            Statement child = stmt.child(i);
            Token name = child.name();
            if (name.equals(TOK_STATEMENT)) {
               parseStatement(m_syntax, child);
            } else if (name.equals(TOK_BLOCK)) {
               parseBlock(m_syntax, child);
            } else if (name.equals(TOK_UNCHECKED)) {
               parseUnchecked(m_syntax, child);
            } else {
               throw new HDLException(name, "expected unchecked, block or statement");
            }
         }
      } catch (HDLException e) {
         m_syntax = null;
         throw e;
      }
   }

   private void verifyArgCount(Statement stmt,
                               int count)
   throws HDLException
   {
      if (stmt.numArguments() != count) {
         throw new HDLException(stmt.name(),
                                "statement requires " + count +
                                " arguments but has " + stmt.numArguments());
      }
   }

   private void verifyArgEquals(Statement stmt,
                                int index,
                                Token token)
   throws HDLException
   {
      if (!stmt.argument(index).equals(token)) {
         throw new HDLException(stmt.argument(index),
                                "arg " + (index + 1) + ": expected '" + token.strValue() + "'");
      }
   }

   private void verifyArgType(Statement stmt,
                              int index,
                              int type)
   throws HDLException
   {
      if (stmt.argument(index).tokenType() != type) {
         throw new HDLException(stmt.argument(index),
                                "arg " + (index + 1) + ": expected " +
                                Token.typeName(type) + " found " +
                                Token.typeName(stmt.argument(index).tokenType()));
      }
   }

   private void verifyIsBlock(Statement stmt,
                              boolean is_block)
   throws HDLException
   {
      if (is_block) {
         if (!stmt.isBlock())
            throw new HDLException(stmt.name(), "statement must be a block");
      } else {
         if (stmt.isBlock())
            throw new HDLException(stmt.name(), "statement must not be a block");
      }
   }

   public static void main(String[] args)
   {
      try {
         SyntaxChecker checker = new SyntaxChecker(args[0]);
         Parser p = new Parser();
         for (int i = 1; i < args.length; i++) {
            System.out.println("Parsing file '" + args[i] + "'");
            Statement s = p.parse(args[i]);
            checker.checkSyntax(s);
         }
      } catch (Exception e) {
         e.printStackTrace();
      }
   }

   private CheckStatement m_syntax;

   private static final Token TOK_LANGUAGE  = Token.idToken("language");
   private static final Token TOK_STATEMENT = Token.idToken("statement");
   private static final Token TOK_UNCHECKED = Token.idToken("unchecked");
   private static final Token TOK_BLOCK     = Token.idToken("block");
   private static final Token TOK_ARGS      = Token.idToken("args");
   private static final Token TOK_TABLE     = Token.idToken("table");
   private static final Token TOK_ID        = Token.idToken("id");
   private static final Token TOK_ANYID     = Token.idToken("anyid");
   private static final Token TOK_INT       = Token.idToken("int");
   private static final Token TOK_STRING    = Token.idToken("string");
   private static final Token TOK_RECURSIVE = Token.idToken("recursive");
}
