///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility;

import java.io.*;

/**
   Utility class for computing CRC values.
 */
public class CrcCalc
{
   /*
     Compute the CRC tables used by the algorithms.
     Using a static initializer assures that the tables are
     computed exactly once (when the CrcCalc class is loaded).
    */
   static
   {
      precalc16();
      precalc32();
   }

   /**
      Compute and return 16 bit CRC for this block of data.
    */
   public static short crc16(byte[] data)
   {
      return crc16(data, (short)0);
   }

   /**
      Compute and return 16 bit CRC for this block of data plus
      padding using pad_data to make the total length of the data
      at least equal to pad_to_length.
    */
   public static short crc16(byte[] data,
                             int pad_to_length,
                             byte pad_data)
   {
      short crc = crc16(data);
      if (data.length < pad_to_length) {
         crc = crc16(pad_data, pad_to_length - data.length, crc);
      }
      return crc;
   }

   /**
      Compute and return 16 bit CRC for this block of data using the
      specified starting value.
    */
   public static short crc16(byte[] data,
                             short crc)
   {
      for (int i = 0; i < data.length; i++) {
         crc = (short)(CRC_16[((crc >>> 8) ^ data[i]) & 0xff] ^ (crc << 8));
      }
      return crc;
   }

   /**
      Compute and return 16 bit CRC for an array of "data" bytes of the
      specified length using the specified starting value.
    */
   public static short crc16(byte data,
                             int length,
                             short crc)
   {
      for (int i = 0; i < length; i++) {
         crc = (short)(CRC_16[((crc >>> 8) ^ data) & 0xff] ^ (crc << 8));
      }
      return crc;
   }

   /**
      Compute and return 32 bit CRC for this block of data.
    */
   public static int crc32(byte[] data)
   {
      return crc32(data, 0);
   }

   /**
      Compute and return 32 bit CRC for this block of data plus
      padding using pad_data to make the total length of the data
      at least equal to pad_to_length.
    */
   public static int crc32(byte[] data,
                           int pad_to_length,
                           byte pad_data)
   {
      int crc = crc32(data);
      if (data.length < pad_to_length) {
         crc = crc32(pad_data, pad_to_length - data.length, crc);
      }
      return crc;
   }

   /**
      Compute and return 32 bit CRC for this block of data using the
      specified starting value.
    */
   public static int crc32(byte[] data,
                           int start,
                           int length,
                           int crc)
   {
      crc = ~crc;
      int last = start + length;
      for (int i = start; i < last; i++) {
         crc = CRC_32[(crc ^ data[i]) & 0xff] ^ ((crc >>> 8) & 0xffffff);
      }
      return ~crc;
   }

   /**
      Compute and return 32 bit CRC for this block of data using the
      specified starting value.
    */
   public static int crc32(byte[] data,
                           int crc)
   {
      crc = ~crc;
      for (int i = 0; i < data.length; i++) {
         crc = CRC_32[(crc ^ data[i]) & 0xff] ^ ((crc >>> 8) & 0xffffff);
      }
      return ~crc;
   }

   /**
      Compute and return 32 bit CRC for an array of "data" bytes of the
      specified length using the specified starting value.
    */
   public static int crc32(byte data,
                           int length,
                           int crc)
   {
      crc = ~crc;
      for (int i = 0; i < length; i++) {
         crc = CRC_32[(crc ^ data) & 0xff] ^ ((crc >>> 8) & 0xffffff);
      }
      return ~crc;
   }

   private static void precalc16()
   {
      CRC_16 = new short[256];
      for (int i = 0; i < 256; i++) {
         short crc = (short)(i << 8);
         for (int k = 0; k < 8; k++) {
            crc = (short)((crc << 1) ^ (((crc & 0x8000) != 0) ? 0x1021 : 0));
         }
         CRC_16[i] = (short)(crc & 0xffff);
      }
   }

   private static void precalc32()
   {
      CRC_32 = new int[256];
      for (int i = 0; i < 256; i++) {
         int crc = i;
         for (int k = 0; k < 8; k++) {
            crc = (crc >>> 1) ^ (((crc & 0x00000001) != 0) ? 0xedb88320 : 0);
         }
         CRC_32[i] = crc & 0xffffffff;
      }
   }

   public static int checkFile(String filename)
      throws IOException
   {
      int sum = 0;
      InputStream in = new FileInputStream(filename);
      try {
	 byte[] buffer = new byte[1024];
	 for (int len = in.read(buffer); len > 0; len = in.read(buffer)) {
	    sum = crc32(buffer, 0, len, sum);
	 }
      } finally {
	 in.close();
      }
      return sum;
   }

   public static void main(String[] args)
   {
      for (int i = 0; i < args.length; i++) {
	 try {
	    int sum = checkFile(args[i]);
	    System.out.println(args[i] + "\t0x" + Integer.toHexString(sum));
	 } catch (IOException ex) {
	    System.err.println("error for " + args[i] + ": " + ex.getMessage());
	 }
      }
   }

   private static short[] CRC_16;
   private static int[]   CRC_32;
}
