///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility;

import java.util.Hashtable;

/**
   A hash table based LRU cache.
   The number of items to keep in the cache is set by arg to constructor.
   Objects are kept in a hash table and a linked list.
   The linked list is sorted so that most recently accessed object
   is always at the front of the list and least recently used is
   always at back of the list.
   The hash table points to the node in the linked list corresponding
   to the given key.
 */
public class LRUCache
{
   public LRUCache(int max_elements)
   {
      m_cache   = new Hashtable();
      m_lruList = new LinkedList();
      m_maxSize = max_elements;
   }

   // Caller has to assure syncronization.
   private void checkSize()
   {
      if (m_lruList.size() == m_maxSize) {
	 LinkedList.Node lru = m_lruList.tail();
	 lru.unlink();

	 Pair p = (Pair)lru.value();
	 m_cache.remove(p.key());
      }
   }

   public Object get(Object key)
   {
      synchronized (m_lruList) {
	 LinkedList.Node n = (LinkedList.Node)m_cache.get(key);
	 if (n == null) {
	    return null;
	 }

	 n.unlink();
	 m_lruList.push(n);

	 Pair p = (Pair)n.value();
	 return p.value();
      }
   }

   public int maxSize()
   {
      return m_maxSize;
   }

   public void put(Object key,
		   Object value)
   {
      synchronized (m_lruList) {
	 LinkedList.Node n = (LinkedList.Node)m_cache.get(key);
	 if (n == null) {
	    checkSize();
	    n = m_lruList.push(new Pair(key, value));
	    m_cache.put(key, n);
	 } else {
	    Pair p = (Pair)n.value();
	    p.setValue(value);
	    n.unlink();
	    m_lruList.push(n);
	 }
      }
   }

   public void remove(Object key)
   {
      synchronized (m_lruList) {
	 LinkedList.Node n = (LinkedList.Node)m_cache.get(key);
	 if (n != null) {
	    m_cache.remove(key);
	    n.unlink();
	 }
      }
   }

   public int size()
   {
      return m_lruList.size();
   }

   public void clear()
   {
      synchronized (m_lruList) {
	 m_cache.clear();
	 m_lruList.clear();
      }
   }

   private Hashtable  m_cache;
   private LinkedList m_lruList;
   private int        m_maxSize;
}
