///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility;

import java.util.Hashtable;

public class LRUPool
{
   public LRUPool(LRUPoolFactory factory,
                  int min_size)
   {
      this(factory, min_size, -1);
   }

   public LRUPool(LRUPoolFactory factory,
                  int min_size,
                  int max_size)
   {
      m_factory = factory;
      m_list = new LinkedList();
      m_hash = new Hashtable();
      m_minSize = min_size;
      m_maxSize = max_size;
   }

   public synchronized void setMinSize(int new_min)
   {
      m_minSize = new_min;
   }

   public static class ClassFactory
      implements LRUPoolFactory
   {
      public ClassFactory(Class item_class)
      {
         m_itemClass = item_class;
      }

      public LRUsable createPoolItem()
      {
         try {
            Object item = m_itemClass.newInstance();
            return (LRUsable)item;
         } catch (Exception ex) {
            throw new RuntimeException(ex.toString());
         }
      }
      
      private Class m_itemClass;
   }

   public synchronized void use(Object obj)
   {
      //System.out.println("pool.use(" + obj + ")");
      LinkedList.Node node = (LinkedList.Node)m_hash.get(obj);
      node.unlink();
      m_list.append(node);
   }

   public synchronized Object getNextAvailable()
   {
      if (m_list.size() < m_minSize) {
         Object item = newItem();
         //System.out.println("pool.getNextAvailable() below min, created " + item);
         return item;
      }

      LinkedList.Node node = m_list.head();
      while (node != null) {
         LRUsable item = (LRUsable)node.value();
         if (item.isAvailableForUse()) {
            node.unlink();
            m_list.append(node);
            //System.out.println("pool.getNextAvailable() returned " + item);
            return item;
         }
         node = node.next();
      }

      if ((m_maxSize <= 0) || (m_list.size() < m_maxSize)) {
         Object item = newItem();
         //System.out.println("pool.getNextAvailable() created " + item);
         return item;
      }

      //System.out.println("pool.getNextAvailable() none available!");
      return null;
   }

   private Object newItem()
   {
      Object item = m_factory.createPoolItem();
      if (Assert.DEBUG) Assert.pre(item != null);
      LinkedList.Node node = m_list.append(item);
      m_hash.put(item, node);
      return item;
   }

   public int size()
   {
      if (Assert.DEBUG) Assert.pre(m_list.size() == m_hash.size());
      return m_list.size();
   }

   private LinkedList m_list;
   private Hashtable m_hash;
   private LRUPoolFactory m_factory;
   private int m_minSize;
   private int m_maxSize;
}

