///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility;

import java.util.Enumeration;
import java.lang.reflect.Array;

/**
   A doubly linked list.
 */
public class LinkedList implements Cloneable
{
   public LinkedList()
   {
      m_head = new Node();
      m_size = 0;
   }

   public LinkedList(Enumeration en)
   {
      this();
      while (en.hasMoreElements()) {
	 append(en.nextElement());
      }
   }

   public LinkedList(LinkedList ll)
   {
      this();
      for (Node n = ll.head(); n != null; n = n.next()) {
	 append(n.value());
      }
   }

   public LinkedList(Object[] objects)
   {
      this();
      for (int i = 0; i < objects.length; i++) {
	 append(objects[i]);
      }
   }

   public static class AlreadyLinked extends RuntimeException {
      public AlreadyLinked()
      {
         super("node already linked, call unlink() before reinserting");
      }
   }

   public static class NotLinked extends RuntimeException {
      public NotLinked()
      {
         super("node not linked");
      }
   }

   public final class Node {
      Node()
      {
         m_next  = this;
         m_prev  = this;
         m_value = null;
      }

      Node(Object value)
      {
         m_next = null;
         m_prev = null;
         m_value = value;
      }

      public Object value()
      {
         synchronized (LinkedList.this)
         {
            return m_value;
         }
      }

      public void setValue(Object obj)
      {
         synchronized (LinkedList.this)
         {
            m_value = obj;
         }
      }

      public Node next()
      {
         synchronized (LinkedList.this)
         {
            return (m_next == m_head) ? null : m_next;
         }
      }

      public Node prev()
      {
         synchronized (LinkedList.this)
         {
            return (m_prev == m_head) ? null : m_prev;
         }
      }

      public void unlink()
      {
         synchronized (LinkedList.this)
         {
            if (!isLinked()) {
               throw new NotLinked();
            }

            m_size--;

            m_next.m_prev = m_prev;
            m_prev.m_next = m_next;

            m_next = null;
            m_prev = null;
         }
      }

      public void insertBefore(Node other)
      {
         synchronized (LinkedList.this)
         {
            if (isLinked()) {
               throw new AlreadyLinked();
            }

            if (!other.isLinked()) {
               throw new NotLinked();
            }

            m_next = other;
            m_prev = other.m_prev;

            m_next.m_prev = this;
            m_prev.m_next = this;

            m_size++;
         }
      }

      public void insertAfter(Node other)
      {
         synchronized (LinkedList.this)
         {
            if (isLinked()) {
               throw new AlreadyLinked();
            }

            if (!other.isLinked()) {
               throw new NotLinked();
            }

            m_prev = other;
            m_next = other.m_next;

            m_next.m_prev = this;
            m_prev.m_next = this;

            m_size++;
         }
      }

      private boolean isLinked()
      {
         return (m_prev != null) || (m_next != null);
      }

      Node m_next;
      Node m_prev;
      Object m_value;
   }

   public synchronized void clear()
   {
      m_head = new Node();
      m_size = 0;
   }

   public synchronized Node insertAfter(Node node,
                                        Object value)
   {
      Node n = new Node(value);
      n.insertAfter(node);
      return n;
   }

   public synchronized Node insertBefore(Node node,
                                         Object value)
   {
      Node n = new Node(value);
      n.insertBefore(node);
      return n;
   }

   public synchronized Node append(Object value)
   {
      Node n = new Node(value);
      n.insertBefore(m_head);
      return n;
   }

   public synchronized Node append(Node n)
   {
      n.insertBefore(m_head);
      return n;
   }

   public synchronized void copyInto(Object[] items,
                                     int offset)
   {
      int index = offset;
      for (Node n = m_head.next(); n != null; n = n.next()) {
         items[index++] = n.m_value;
      }
   }

   public synchronized void copyInto(Object[] items)
   {
      copyInto(items, 0);
   }

   public synchronized Node nodeOf(Object obj)
   {
      for (Node n = m_head.next(); n != null; n = n.next()) {
         if (obj.equals(n.m_value)) {
            return n;
         }
      }
      return null;
   }

   public synchronized Node nodeAt(int index)
   {
      if (index < 0 || index >= m_size) {
	 throw new IndexOutOfBoundsException(String.valueOf(index));
      }

      Node n = m_head.next();
      for (int i = 1; i < index; i++) {
	 n = n.next();
      }

      return n;
   }

   public int indexOf(Object o)
   {
      int index = 0;
      for (Node n = m_head.next(); n != null; n = n.next()) {
         if (o.equals(n.m_value)) {
            return index;
         }
	 index++;
      }
      return -1;
   }

   public int lastIndexOf(Object o)
   {
      int index = 0;
      int last_found = -1;
      for (Node n = m_head.next(); n != null; n = n.next()) {
         if (o.equals(n.m_value)) {
            last_found = index;
         }
	 index++;
      }
      return last_found;
   }

   public synchronized boolean contains(Object obj)
   {
      return indexOf(obj) >= 0;
   }

   public synchronized Node head()
   {
      return m_head.next();
   }

   /**
      Returns an array of references to the list's contents.
      The type of the array is _class[].  If any element in the
      list is not of the specified class (or an extension of the
      class) an ArrayStoreException will be thrown.
    */
   public synchronized Object[] getValues(Class _class)
   {
      Object[] array = (Object[])Array.newInstance(_class, m_size);
      copyInto(array, 0);
      return array;
   }

   public synchronized Object[] getValues()
   {
      Object[] array = new Object[m_size];
      copyInto(array, 0);
      return array;
   }

   public synchronized Enumeration elements()
   {
      return new ObjectArrayEnumerator(getValues());
   }

   public synchronized Node[] nodes()
   {
      Node[] array = new Node[m_size];

      int index = 0;
      for (Node n = m_head.next(); n != null; n = n.next()) {
         array[index++] = n;
      }
      return array;
   }

   public synchronized Object clone()
   {
      return new LinkedList(this);
   }

   public synchronized Object pop()
   {
      Node n = m_head.next();
      n.unlink();
      return n.m_value;
   }

   public synchronized Object popLast()
   {
      Node n = m_head.prev();
      n.unlink();
      return n.m_value;
   }

   public synchronized Node push(Object value)
   {
      Node n = new Node(value);
      n.insertAfter(m_head);
      return n;
   }

   public synchronized Node push(Node n)
   {
      n.insertAfter(m_head);
      return n;
   }

   public synchronized int size()
   {
      return m_size;
   }

   public synchronized boolean isEmpty()
   {
      return m_size == 0;
   }

   public synchronized Node tail()
   {
      return m_head.prev();
   }

   public boolean equals(Object o)
   {
      try {
	 LinkedList ll = (LinkedList)o;
	 Object[] mine = getValues();
	 Object[] his = ll.getValues();

	 if (mine.length != his.length) {
	    return false;
	 }

	 for (int i = 0; i < mine.length; i++) {
	    if (!mine[i].equals(his[i])) {
	       return false;
	    }
	 }

	 return true;
      } catch (ClassCastException ex) {
	 return false;
      }
   }

   public synchronized int hashCode()
   {
      int hash_code = super.hashCode();
      for (Node n = head(); n != null; n = n.next()) {
	 hash_code = hash_code ^ n.value().hashCode();
      }
      return hash_code;
   }

   private Node m_head;
   private int  m_size;
}
