///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility;

import java.util.Hashtable;

/**
  Class to provide single writer/multipel readers synchronization.
  This class is generally used to provide synchronization in a more controlled
  fashion than the basic Java synchronized construct.

  A thread with a read lock can acquire a nested read lock.
  A thread with a write lock can acquire a nested read or write lock.
  An attempt to get a write lock when the thread already owns a
  read lock throws a LockChangedException.
 */
public final class RWLock
{
   public RWLock()
   {
      m_numReaders = 0;
      m_numWriters = 0;
      m_writerActive = false;
      m_lockers = new Hashtable();
   }

   private static class LockChangeException
   extends RuntimeException {
      LockChangeException()
      {
         super("current thread already has read lock, cannot promote read lock to write lock");
      }
   }

   /**
     The RWLOck maintains a hash table mapping threads to LockerNodes.
     The LockerNode maintains the original type of lock (note that nested
     locks for the same thread could be of a different type than the
     original lock) and a reference count of the number of nested locks the
     thread owns.  The reference count is used to remove the node when the
     top most lock is released.
    */
   private static final class LockerNode {
      LockerNode(int type)
      {
         m_type = type;
         m_refCount = 1;
      }

      void addRef()
      {
         m_refCount++;
      }

      boolean removeRef()
      {
         m_refCount--;
         return m_refCount == 0;
      }

      int getRefCount()
      {
         return m_refCount;
      }

      int getType()
      {
         return m_type;
      }

      private int m_type;
      private int m_refCount;
   }

   /**
     Acquire a reader lock.
     Blocks until the lock is acquired.
     If an InterruptedException is thrown at us, we pass it through
     to the caller.
    */
   public synchronized void acquireReadLockInt()
   throws InterruptedException
   {
      LockerNode node = getLocker();
      if (node != null) {
         // we can add a read lock to an existing read or write lock
         node.addRef();
         return;
      }

      while (m_numWriters > 0) {
         wait();
      }

      addLocker(READ_LOCK);
      m_numReaders++;
   }

   /**
     Acquire a reader lock.
     Blocks until the lock is acquired.
     InterruptedExceptions are silently ignored.
    */
   public synchronized void acquireReadLock()
   {
      while (true) {
         try {
            acquireReadLockInt();
            return;
         } catch (InterruptedException ex) {
            
         }
      }
   }

   /**
     Acquire a writer lock.
     Blocks until the lock is acquired.
     If an InterruptedException is thrown at us, we pass it through
     to the caller.
    */
   public synchronized void acquireWriteLockInt()
   throws InterruptedException
   {
      LockerNode node = getLocker();
      if (node != null) {
         // cannot convert read lock into write lock safely
         if (node.getType() != WRITE_LOCK) {
            throw new LockChangeException();
         }

         node.addRef();
         return;
      }

      m_numWriters++;

      try {
         // wait for all of the readers to release their locks
         while (m_numReaders > 0) {
            wait();
         }

         // wait for our turn at the write lock
         // (other would-be writers could get there before us)
         while (m_writerActive) {
            wait();
         }
      } catch (InterruptedException ex) {
         m_numWriters--;
         throw ex;
      }

      // our turn!
      addLocker(WRITE_LOCK);
      m_writerActive = true;
   }

   /**
     Acquire a writer lock.
     Blocks until the lock is acquired.
     InterruptedExceptions are silently ignored.
    */
   public synchronized void acquireWriteLock()
   {
      while (true) {
         try {
            acquireWriteLockInt();
            return;
         } catch (InterruptedException ex) {
            // ignored
         }
      }
   }

   public synchronized void releaseReadLock()
   {
      LockerNode node = getLocker();
      if (node.removeRef()) {
         m_numReaders--;
         removeLocker(READ_LOCK);
         notifyAll();
      }
   }

   public synchronized void releaseWriteLock()
   {
      LockerNode node = getLocker();
      if (node.removeRef()) {
         m_numWriters--;
         m_writerActive = false;
         removeLocker(WRITE_LOCK);
         notifyAll();
      }
   }

   private void addLocker(int type)
   {
      Thread t = Thread.currentThread();
      
      if (Assert.DEBUG) Assert.pre(!m_lockers.containsKey(t));
      
      m_lockers.put(t, new LockerNode(type));
   }

   private void removeLocker(int type)
   {
      LockerNode node = getLocker();
      
      if (Assert.DEBUG) Assert.pre(node.getType() == type);
      if (Assert.DEBUG) Assert.pre(node.getRefCount() == 0);

      m_lockers.remove(Thread.currentThread());
   }

   private LockerNode getLocker()
   {
      return (LockerNode)m_lockers.get(Thread.currentThread());
   }

   public boolean isOuterWriteLock()
   {
      LockerNode node = getLocker();
      return (node.getType() == WRITE_LOCK) && (node.getRefCount() == 1);
   }

   public boolean isOuterReadLock()
   {
      LockerNode node = getLocker();
      return (node.getType() == READ_LOCK) && (node.getRefCount() == 1);
   }

   private int m_numReaders;
   private int m_numWriters;
   private boolean m_writerActive;
   private Hashtable m_lockers;

   private static final int READ_LOCK = 1;
   private static final int WRITE_LOCK = 2;
}
