///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility;

import java.util.Random;

public class Sorter
{
   public Sorter(BinaryComparator compare)
   {
      m_compare = compare;
   }

   public void setComparator(BinaryComparator compare)
   {
      m_compare = compare;
   }

   protected final void swap(Object[] array,
			     int first,
			     int second)
   {
      if (Assert.DEBUG) Assert.pre(m_compare.compare(array[first], array[second]) >= 0);
      Object tmp    = array[first];
      array[first]  = array[second];
      array[second] = tmp;
   }

   private boolean checkPivot(Object[] array,
			      int first,
			      int limit,
			      int pivot_index,
			      Object pivot_obj)
   {
      if (Assert.DEBUG) Assert.inv(array[pivot_index] == pivot_obj);

      for (int i = first; i < pivot_index; i++) {
	 if (m_compare.compare(array[i], pivot_obj) > 0) {
	    return false;
	 }
      }

      for (int i = pivot_index; i < limit; i++) {
	 if (m_compare.compare(array[i], pivot_obj) < 0) {
	    return false;
	 }
      }

      return true;
   }

   private int pivot(Object[] array,
		     int first,
		     int limit)
   {
      if (Assert.DEBUG) Assert.pre(first < limit);
      if (Assert.DEBUG) Assert.pre(limit - first >= 3);
      Object pivot_obj = array[first];
      int left  = first + 1;
      int right = limit - 1;
      while (left < limit && m_compare.compare(array[left], pivot_obj) <= 0) {
	 left++;
      }
      while (m_compare.compare(array[right], pivot_obj) > 0) {
	 if (Assert.DEBUG) Assert.inv(right > first);
	 right--;
      }
      while (left < right) {
	 if (Assert.DEBUG) Assert.inv(left > first);
	 if (Assert.DEBUG) Assert.inv(left < limit);
	 if (Assert.DEBUG) Assert.inv(right > first);
	 if (Assert.DEBUG) Assert.inv(right < limit);
	 swap(array, left, right);
	 left++;
	 while (m_compare.compare(array[left], pivot_obj) <= 0) {
	    left++;
	 }
	 right--;
	 while (m_compare.compare(array[right], pivot_obj) > 0) {
	    right--;
	 }
      }
      if (Assert.DEBUG) Assert.inv(right >= first);
      if (Assert.DEBUG) Assert.inv(right < limit);
      swap(array, first, right);

      if (Assert.DEBUG) Assert.post(checkPivot(array, first, limit, right, pivot_obj));

      return right;
   }

   protected boolean isSorted(Object[] array, int first, int limit)
   {
      for (int i = first; i < limit - 1; i++) {
	 if (m_compare.compare(array[i], array[i+1]) >= 0) {
	    System.out.println("array[" + i + "] = " + array[i]);
	    System.out.println("array[" + (i+1) + "] = " + array[i+1]);
	    return false;
	 }
      }
      return true;
   }

   protected void quicksort(Object[] array,
			    int first,
			    int limit)
   {
      int num_values = limit - first;
      switch (num_values) {
      case 0:
      case 1:
	 break;

      case 2:
	 if (m_compare.compare(array[first], array[first + 1]) > 0) {
	    swap(array, first, first + 1);
	 }
	 break;

      case 3:
	 if (m_compare.compare(array[first], array[first + 1]) > 0) {
	    swap(array, first, first + 1);
	 }
	 if (m_compare.compare(array[first + 1], array[first + 2]) > 0) {
	    swap(array, first + 1, first + 2);
	 }
	 if (m_compare.compare(array[first], array[first + 1]) > 0) {
	    swap(array, first, first + 1);
	 }
	 break;

      default:
	 {
	    int pivot_index = pivot(array, first, limit);
	    if (pivot_index > first) {
	       quicksort(array, first, pivot_index);
	    }
	    if ((pivot_index + 1) < limit) {
	       quicksort(array, pivot_index + 1, limit);
	    }
	    break;
	 }
      }
      if (Assert.DEBUG) Assert.post(isSorted(array, first, limit));
   }

   public void sort(Object[] array)
   {
      quicksort(array, 0, array.length);
      if (Assert.DEBUG) Assert.post(isSorted(array, 0, array.length));
   }

   public static void main(String[] args)
   {
      Sorter sort = new Sorter(new IntegerComparator());
      long seed = System.currentTimeMillis();
      Random r = new Random(seed);
      for (int i = 0; i < 100; i++) {
	 int array_size = Math.abs(r.nextInt()) % 500;
	 Object[] array = new Integer[array_size];
	 for (int k = 0; k < array_size; k++) {
	    array[k] = new Integer(r.nextInt());
	 }
	 sort.sort(array);
	 if (!sort.isSorted(array, 0, array.length)) {
	    System.out.println("failed to sort array " + i + " for seed " + seed);
	 }
      }
   }

   private BinaryComparator m_compare;
}
