///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility;

import java.util.Enumeration;
import java.util.Random;
import java.util.Vector;

/**
   Extends java.util.Vector to provide some extra functionality.
   Makes few assumptions about java.util.Vector to minimize risk of
   changes to Vector breaking this class.  Assumes read access to
   elementData and elementCount are safe.
 */
public class SuperVector extends Vector implements Cloneable
{
   public SuperVector()
   {
   }

   public SuperVector(int initialCapacity)
   {
      super(initialCapacity);
   }

   public SuperVector(int initialCapacity,
                      int capacityIncrement)
   {
      super(initialCapacity, capacityIncrement);
   }

   public SuperVector(Object[] initialElements)
   {
      super(initialElements.length);
      setElementsFromArray(initialElements);
   }

   public SuperVector(Object[] initialElements,
                      int capacityIncrement)
   {
      super(initialElements.length, capacityIncrement);
      setElementsFromArray(initialElements);
   }

   public SuperVector(Enumeration initialElements)
   {
      addElementsFromEnum(initialElements);
   }

   public SuperVector(Enumeration initialElements,
                      int initialCapacity)
   {
      super(initialCapacity);
      addElementsFromEnum(initialElements);
   }

   public SuperVector(Enumeration initialElements,
                      int initialCapacity,
                      int capacityIncrement)
   {
      super(initialCapacity, capacityIncrement);
      addElementsFromEnum(initialElements);
   }

   public SuperVector(Vector vector)
   {
      super(vector.size());
      synchronized (vector) {
	 addElementsFromEnum(vector.elements());
      }
   }

   public Object clone()
   {
      return new SuperVector(this);
   }

   public void push(Object o)
   {
      addElement(o);
   }

   public synchronized Object pop()
   {
      int index = size() - 1;
      Object result = elementAt(index);
      removeElementAt(index);
      return result;
   }

   public Object peek()
   {
      return lastElement();
   }

   /**
      Set Vector to contain exactly the elements in specified array.
   */
   public synchronized void setElementsFromArray(Object[] objects)
   {
      removeAllElements();
      for (int i = 0; i < objects.length; i++) {
         addElement(objects[i]);
      }
   }

   /**
      Add array of objects to Vector.
   */
   public synchronized void addElementsFromArray(Object[] objects)
   {
      ensureCapacity(capacity() + objects.length);
      for (int i = 0; i < objects.length; i++) {
         addElement(objects[i]);
      }
   }

   /**
      Set Vector to only the objects in specified enumeration.
   */
   public synchronized void setElementsFromEnum(Enumeration enum)
   {
      removeAllElements();
      while (enum.hasMoreElements()) {
         addElement(enum.nextElement());
      }
   }

   /**
      Set Vector to only the objects in specified Vector.
   */
   public void setElementsFromVector(Vector v)
   {
      setElementsFromEnum(v.elements());
   }

   /**
      Add enumeration of objects to Vector.
   */
   public synchronized void addElementsFromEnum(Enumeration enum)
   {
      while (enum.hasMoreElements()) {
         addElement(enum.nextElement());
      }
   }

   /**
      Similar to elements() except it returns an array of references
      to the elements of the Vector.  The array is a copy of the Vector's
      underlying array and has length equal to the vector's size().
      Changing the array has no effect on the underlying Vector (and 
      vice versa)so this is just as safe as using an Enumeration but
      has the benefit of allowing random access and multiple passes
      over the elements.
   */
   public synchronized Object[] getValues()
   {
      Object[] result = new Object[ elementCount ];
      copyInto(result);
      return result;
   }

   /**
      Similar to elements() except it returns an array of references
      to the elements of the Vector.  The array is a copy of the Vector's
      underlying array and has length equal to the vector's size().
      Changing the array has no effect on the underlying Vector (and 
      vice versa)so this is just as safe as using an Enumeration but
      has the benefit of allowing random access and multiple passes
      over the elements.
   */
   public synchronized Object[] getValues(Class _class)
   {
      Object[] result = (Object[])java.lang.reflect.Array.newInstance(_class, elementCount);
      copyInto(result);
      return result;
   }

   public synchronized String[] getStrings()
   {
      String[] result = new String[ elementCount ];
      copyInto(result);
      return result;
   }

   public synchronized Integer[] getIntegers()
   {
      Integer[] result = new Integer[ elementCount ];
      copyInto(result);
      return result;
   }

   public synchronized Long[] getLongs()
   {
      Long[] result = new Long[ elementCount ];
      copyInto(result);
      return result;
   }

   /**
      Convert an enumeration to an array.
   */
   public static Object[] enumToArray(Enumeration enum)
   {
      return new SuperVector(enum).getValues();
   }

   /**
      Convert an enumeration to an array.
   */
   public static Object[] enumToArray(Class _class,
                                      Enumeration enum)
   {
      return (Object[])new SuperVector(enum).getValues(_class);
   }

   /**
      Store an enumeration in a specified array.
      Array must be large enough to hold the objects in the enumeration.
   */
   public static void copyEnumToArray(Enumeration enum,
                                      Object[] array,
                                      int offset)
   {
      while (enum.hasMoreElements()) {
         array[offset++] = enum.nextElement();
      }
   }

   /**
      Store an enumeration in a specified array.
      Array must be large enough to hold the objects in the enumeration.
   */
   public static void copyEnumToArray(Enumeration enum,
                                      Object[] array)
   {
      copyEnumToArray(enum, array, 0);
   }

   /**
      Randomly shuffle an array of objects using random number generator
      provided by caller..
   */
   public static void shuffle(Object[] array,
                              Random r)
   {
      for (int pass = 0; pass < 3; pass++) {
         for (int i = 0; i < array.length; i++) {
            int k = Math.abs(r.nextInt() >> 3) % array.length;
            Object tmp = array[i];
            array[i] = array[k];
            array[k] = tmp;
         }
      }
   }

   /**
      Randomly shuffle an array of objects using a temporary random number
      generator.
    */
   public static void shuffle(Object[] array)
   {
      shuffle(array, new Random());
   }
}
