///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility.storage;

import java.io.IOException;
import java.util.Stack;
import org.cooldevtools.utility.Assert;
import org.cooldevtools.utility.LRUCache;

public class BTreeIndex
implements StorageIndex
{
   public BTreeIndex(Storage storage,
                     int key_length,
                     int keys_per_node)
   throws IOException, StorageException
   {
      m_storage      = storage;
      m_keyLength    = key_length;
      m_cache        = new LRUCache(128);
      m_keysPerNode  = keys_per_node;
      m_recordLength = calcRecordLength(key_length, keys_per_node);
      if (storage.getRecordLength() != m_recordLength) {
         throw new InvalidRecordLengthException();
      }
   }

   public static int calcRecordLength(int key_length,
                                      int keys_per_node)
   {
      return (4 + 2 * key_length + 4) * keys_per_node + 4;
   }

   private final class Node
   {
      public Node()
      {
         this(Storage.NULL);
      }

      public Node(int record_id)
      {
         m_recordId = record_id;
         m_keys     = new String[m_keysPerNode + 1];
         m_links    = new int[m_keysPerNode + 1];
         m_isLeaf   = false;
         m_numKeys  = 0;
      }

      public Node(int record_id,
                  byte[] bytes)
      throws IOException
      {
         m_recordId = record_id;
         m_keys     = new String[m_keysPerNode + 1];
         m_links    = new int[m_keysPerNode + 1];
         m_isLeaf   = (Packer.getShort(bytes, 0) == 1);
         m_numKeys  = Packer.getShort(bytes, 2);
         int offset = 4;
         for (int i = 0; i < m_numKeys; i++) {
            m_keys[i]  = Packer.getUString(bytes, offset);
            offset += 4 + 2 * m_keyLength;
            m_links[i] = Packer.getInt(bytes, offset);
            offset += 4;
         }
      }

      public byte[] toByteArray()
      throws IOException
      {
         byte[] bytes = new byte[m_recordLength];
         Packer.putShort((short)(m_isLeaf ? 1 : 0), bytes, 0);
         Packer.putShort((short)m_numKeys, bytes, 2);
         int offset = 4;
         for (int i = 0; i < m_numKeys; i++) {
            Packer.putUString(m_keys[i], m_keyLength, bytes, offset);
            offset += 4 + 2 * m_keyLength;
            Packer.putInt(m_links[i], bytes, offset);
            offset += 4;
         }
         return bytes;
      }

      public void checkSanity()
      {
         for (int i = 1; i < m_numKeys; i++) {
            if (m_keys[i-1].compareTo(m_keys[i]) >= 0) {
               Assert.impossible(this + " key ordering invariant violation: "
                                 + m_keys[i-1] + " GT " + m_keys[i]);
            }
         }

         if (!isRoot()) {
            if (m_numKeys > m_keysPerNode) {
               Assert.impossible("overflow: " + this);
            }
            if (m_numKeys < (m_keysPerNode / 2)) {
               Assert.impossible("underflow: " + this);
            }
         }
         
         if (!isLeaf()) {
            try {
               for (int i = 0; i < m_numKeys; i++) {
                  Node child = getChild(i);
                  if (!m_keys[i].equals(child.lastKey())) {
                     Assert.impossible("invalid key in parent " + this + " for child " + i + ": " + child);
                  }
               }
            } catch (IOException ex) {
               ex.printStackTrace();
               Assert.impossible(ex);
            }
         }
      }

      public boolean greaterThan(Node other)
      {
         return m_keys[0].compareTo(other.m_keys[0]) > 0;
      }

      public int getSibling(int index)
      throws IOException, StorageException
      {
         if (TRACE_FLOW) debug("getSibling:parent " + this);
         if (TRACE_FLOW) debug("getSibling:index " + index + " link " + m_links[index]);

         if (index == 0) {
            return index + 1;
         }

         if (index == (m_numKeys - 1)) {
            return index - 1;
         }

         Node node = getChild(index);

         int left_index = index - 1;
         Node left = getChild(left_index);
         if (left.canMerge(node)) {
            return left_index;
         }

         int right_index = index + 1;
         Node right = getChild(right_index);
         if (right.canMerge(node)) {
            return right_index;
         }

         return left.canAbsorb(node) ? left_index : right_index;
      }

      public boolean isRoot()
      {
         return m_recordId == ROOT_ID;
      }

      public String toString()
      {
         String value = ("Node("
                         + m_recordId
                         + " "
                         + (m_isLeaf ? "leaf" : "branch")
                         + " with "
                         + m_numKeys
                         + " keys");
         for (int i = 0; i < m_numKeys; i++) {
            value = (value
                     + "|"
                     + m_keys[i]
                     + ","
                     + m_links[i]);
         }
         return value + ")";
      }

      public void setIsLeaf(boolean value)
      {
         m_isLeaf = value;
      }

      public void setLink(int index,
                          int link)
      {
         m_links[index] = link;
      }

      public String key(int index)
      {
         return m_keys[index];
      }

      public void setKey(int index,
                         String key)
      {
         m_keys[index] = key;
      }

      public int link(int index)
      {
         return m_links[index];
      }

      public String lastKey()
      {
         return m_keys[m_numKeys - 1];
      }

      public int lastLink()
      {
         return m_links[m_numKeys - 1];
      }

      public boolean isLeaf()
      {
         return m_isLeaf;
      }

      public int numKeys()
      {
         return m_numKeys;
      }

      public int recordId()
      {
         return m_recordId;
      }

      public boolean isOverflow()
      {
         return (m_numKeys == m_keys.length);
      }

      public boolean isUnderflow()
      {
         return (m_numKeys < (m_keys.length / 2));
      }

      public boolean canAbsorb(Node other)
      {
         return (m_numKeys + other.m_numKeys) < m_keys.length;
      }

      public boolean canMerge(Node other)
      {
         int total = (m_numKeys + other.m_numKeys);
         return (total >= m_keys.length) && (total <= 2 * m_keysPerNode);
      }

      public void release()
      throws IOException, StorageException
      {
         if (TRACE_FLOW) debug("RELEASE: " + this);
         forgetNode(this);
         m_storage.release(m_recordId);
         m_recordId = Storage.NULL;
      }

      public void update()
      throws IOException, StorageException
      {
         if (TRACE_FLOW) debug("UPDATE: " + this);
         if (Assert.DEBUG) Assert.pre(m_recordId != Storage.NULL);
         m_storage.set(m_recordId, toByteArray());
         if (CHECK_SANITY) checkSanity();
      }

      public void append()
      throws IOException, StorageException
      {
         if (TRACE_FLOW) debug("APPEND: " + this);
         if (m_recordId != Storage.NULL) {
            forgetNode(this);
         }

         m_recordId = m_storage.put(toByteArray());
         rememberNode(this);
         if (CHECK_SANITY) checkSanity();
      }

      public void addChild(String key,
                           int link)
      {
         m_links[m_numKeys] = link;
         m_keys[m_numKeys] = key;
         m_numKeys++;
      }

      public Node getChild(int index)
      throws IOException, StorageException, BadRecordIdException
      {
         if (m_links[index] == Storage.NULL) {
            return null;
         }
         return getNode(m_links[index]);
      }

      private void shiftRight(int from,
                              int how_many)
      {
         int num_bytes = m_numKeys - from;
         int dst = m_numKeys + how_many - 1;
         int src = from + num_bytes - 1;
         for (int i = 0; i < num_bytes; i++) {
            m_keys[dst] = m_keys[src];
            m_links[dst] = m_links[src];
            dst--;
            src--;
         }
      }

      public void absorb(Node other)
      {
         if (Assert.DEBUG) Assert.pre(m_keys[0].compareTo(other.lastKey()) > 0);

         shiftRight(0, other.m_numKeys);
         System.arraycopy(other.m_keys, 0, m_keys, 0, other.m_numKeys);
         System.arraycopy(other.m_links, 0, m_links, 0, other.m_numKeys);

         m_numKeys += other.m_numKeys;

         if (Assert.DEBUG) Assert.post(!isOverflow());
      }

      public void merge(Node other)
      {
         if (Assert.DEBUG) Assert.pre(lastKey().compareTo(other.m_keys[0]) < 0);

         int total = m_numKeys + other.m_numKeys;
         int mine  = (total / 2) + (total % 2);
         int his = total - mine;

         if (mine > m_numKeys) {
            int copied = mine - m_numKeys;
            System.arraycopy(other.m_keys, 0, m_keys, m_numKeys, copied);
            System.arraycopy(other.m_links, 0, m_links, m_numKeys, copied);
            System.arraycopy(other.m_keys, copied, other.m_keys, 0, his);
            System.arraycopy(other.m_links, copied, other.m_links, 0, his);
         } else if (mine < m_numKeys) {
            int copied = m_numKeys - mine;
            other.shiftRight(0, copied);
            System.arraycopy(m_keys, mine, other.m_keys, 0, copied);
            System.arraycopy(m_links, mine, other.m_links, 0, copied);
         } else {
            throw new RuntimeException();
         }

         m_numKeys = mine;
         other.m_numKeys = his;
      }

      public void removeChild(int index)
      {
         int copied = m_numKeys - 1 - index;
         System.arraycopy(m_keys, index + 1, m_keys, index, copied);
         System.arraycopy(m_links, index + 1, m_links, index, copied);
         m_numKeys--;
      }

      public boolean isLastIndex(int index)
      {
         return index == (m_numKeys - 1);
      }

      public int insertChild(String key,
                             int record_id)
      throws IOException, StorageException
      {
         if (Assert.DEBUG) Assert.pre(m_numKeys < m_links.length);

         int index = 0;
         if (m_numKeys > 0) {
            index = findFirstGreaterOrEqual(key, 0, m_numKeys - 1);
            int diff = m_keys[index].compareTo(key);
            if (diff == 0) {
               throw new DuplicateKeyException(key);
            }

            if (diff < 0) {
               index++;
               if (Assert.DEBUG) Assert.post(index == m_numKeys);
            }
         }

         shiftRight(index, 1);
         m_keys[index]  = key;
         m_links[index] = record_id;
         m_numKeys++;

         return index;
      }

      private int findExactMatch(String key,
                                 int left,
                                 int right)
      {
         int mid = (left + right) / 2;
         if (mid == left) {
            if (left == right) {
               return m_keys[left].equals(key) ? left : NOT_FOUND;
            }
            if (m_keys[left].equals(key)) {
               return left;
            }
            if (m_keys[right].equals(key)) {
               return right;
            }
            return NOT_FOUND;
         }

         if (Assert.DEBUG) Assert.pre(mid > left);
         if (Assert.DEBUG) Assert.pre(mid < right);

         int diff = key.compareTo(m_keys[mid]);
         if (diff == 0) {
            return mid;
         } else if (diff < 0) {
            return findExactMatch(key, left, mid - 1);
         } else {
            return findExactMatch(key, mid + 1, right);
         }
      }

      private int findFirstGreaterOrEqual(String key,
                                          int left,
                                          int right)
      {
         int mid = (left + right) / 2;
         if (mid == left) {
            return (m_keys[left].compareTo(key) >= 0) ? left : right;
         }

         if (Assert.DEBUG) Assert.pre(mid > left);
         if (Assert.DEBUG) Assert.pre(mid < right);

         int diff = key.compareTo(m_keys[mid]);
         if (diff == 0) {
            return mid;
         } else if (diff < 0) {
            return findFirstGreaterOrEqual(key, left, mid);
         } else {
            return findFirstGreaterOrEqual(key, mid + 1, right);
         }
      }

      public int findMatchingChild(String key,
                                   boolean require_exact)
      {
         if (m_numKeys == 0) {
            return NOT_FOUND;
         }

         if (require_exact) {
            return findExactMatch(key, 0, m_numKeys - 1);
         } else {
            int index = findFirstGreaterOrEqual(key, 0, m_numKeys - 1);
            return (m_keys[index].compareTo(key) >= 0) ? index : NOT_FOUND;
         }
      }

      public int findContainingChild(String key)
      {
         if (Assert.DEBUG) Assert.pre(m_numKeys > 0);
         return findFirstGreaterOrEqual(key, 0, m_numKeys - 1);
      }

      public Node split()
      {
         if (Assert.DEBUG) Assert.pre(isOverflow());

         int middle = (m_numKeys / 2) + (m_numKeys % 2);

         Node right = new Node();

         right.m_isLeaf  = m_isLeaf;
         right.m_numKeys = m_numKeys - middle;

         for (int i = middle; i < m_numKeys; i++) {
            right.m_links[i - middle] = m_links[i];
            right.m_keys[i - middle]  = m_keys[i];
         }

         m_numKeys = middle;

         return right;
      }

      private String[] m_keys;
      private int[]    m_links;
      private int      m_numKeys;
      private int      m_recordId;
      private boolean  m_isLeaf;
   }

   private final class Match {
      public Match(Node parent,
                   int index)
      {
         m_parent = parent.recordId();
         m_index  = index;
      }

      public int index()
      {
         return m_index;
      }

      public String key()
      throws IOException, StorageException
      {
         return node().key(m_index);
      }

      public int link()
      throws IOException, StorageException
      {
         return node().link(m_index);
      }

      public int recordId()
      throws IOException, StorageException
      {
         return node().recordId();
      }

      public boolean isLastIndex()
      throws IOException, StorageException
      {
         return node().isLastIndex(m_index);
      }

      public String toString()
      {
         try {
            return node().toString() + " AT INDEX " + m_index;
         } catch (IOException ex) {
            ex.printStackTrace();
            return ex.getMessage();
         }
      }

      public Node node()
      throws IOException, StorageException
      {
         return getNode(m_parent);
      }

      public void setLink(int link)
      throws IOException, StorageException
      {
         node().setLink(m_index, link);
      }

      public void setKey(String key)
      throws IOException, StorageException
      {
         node().setKey(m_index, key);
      }

      public void next()
      {
         m_index++;
      }

      private int m_parent;
      private int m_index;
   }

   private Node rememberNode(Node n)
   {
      m_cache.put(new Integer(n.recordId()), n);
      return n;
   }

   private void forgetNode(Node n)
   {
      m_cache.remove(new Integer(n.recordId()));
   }

   private Node getNode(int record_id)
   throws IOException, StorageException
   {
      if (Assert.DEBUG) Assert.pre(record_id != Storage.NULL);
      Integer key = new Integer(record_id);
      Object o = m_cache.get(key);
      if (o == null) {
         return rememberNode(new Node(record_id, m_storage.get(record_id)));
      }

      return (Node)o;
   }

   private Node getRoot()
   throws IOException, StorageException
   {
      try {
         return getNode(ROOT_ID);
      } catch (BadRecordIdException ex) {
         Node root = new Node();
         root.setIsLeaf(true);
         root.append();
         if (Assert.DEBUG) Assert.pre(root.recordId() == ROOT_ID);
         return root;
      }
   }

   public synchronized int getKeyLength()
   {
      return m_keyLength;
   }

   public synchronized void close()
   {
      m_storage.close();
   }

   public synchronized void sync()
   throws IOException, StorageException
   {
      m_storage.sync();
   }

   public void abort()
   {
      m_storage.abort();
      m_cache.clear();
   }

   public void commitTo(CommitManager cmgr)
   throws IOException
   {
      m_storage.commitTo(cmgr);
   }

   public void commitWrite()
   throws IOException
   {
      m_storage.commitWrite();
   }

   public void commitFailed()
   {
      m_storage.commitFailed();
      m_cache.clear();
   }

   public void commitSuccessful()
   {
      m_storage.commitSuccessful();
   }

   public void commitRecovered()
   {
      m_storage.commitRecovered();
      m_cache.clear();
   }

   private Match find(String key,
                      Stack stack,
                      boolean require_exact)
   throws IOException, StorageException, UnknownKeyException
   {
      Node node = getRoot();

      if (TRACE_FLOW) debug("find: " + key + " ROOT   " + node);
      while (!node.isLeaf()) {
         if (TRACE_FLOW) debug("find: " + key + " BRANCH " + node);
         int index = node.findContainingChild(key);
         if (TRACE_FLOW) debug("next index=" + index);
         stack.push(new Match(node, index));
         node = node.getChild(index);
      }

      if (TRACE_FLOW) debug("find: " + key + " LEAF   " + node);
      int index = node.findMatchingChild(key, require_exact);
      Match m = new Match(node, index);
      stack.push(m);
      return (index == NOT_FOUND) ? null : m;
   }

   public synchronized boolean exists(String key)
   throws IOException, StorageException
   {
      Match match = find(key, new Stack(), true);
      return (match != null);
   }

   public synchronized int get(String key)
   throws IOException, StorageException, UnknownKeyException
   {
      Match match = find(key, new Stack(), true);
      if (match == null) {
         throw new UnknownKeyException(key);
      }

      return match.link();
   }

   private Match[] stackToArray(Stack stack)
   {
      Match[] nodes = new Match[stack.size()];
      for (int i = 0; i < nodes.length; i++) {
         nodes[i] = (Match)stack.pop();
      }
      return nodes;
   }

   private void splitRoot(Node old_root)
   throws IOException, StorageException
   {
      Node right = old_root.split();

      old_root.append();
      right.append();

      Node new_root = new Node(ROOT_ID);
      new_root.addChild(old_root.lastKey(), old_root.recordId());
      new_root.addChild(right.lastKey(), right.recordId());
      rememberNode(new_root);
      new_root.update();
   }

   private void split(Node node,
                      Match[] parents,
                      int parent_index)
   throws IOException, StorageException
   {
      if (node.isRoot()) {
         if (Assert.DEBUG) Assert.pre(parent_index == parents.length);
         splitRoot(node);
         return;
      }

      int node_index = parents[parent_index].index();

      Node parent = parents[parent_index].node();
      if (Assert.DEBUG) Assert.pre(parent.key(node_index).equals(node.lastKey()));

      Node right = node.split();
      right.append();
      node.update();

      if (Assert.DEBUG) Assert.pre(parent.key(node_index).equals(right.lastKey()));
      parent.setLink(node_index, right.recordId());
      parent.insertChild(node.lastKey(), node.recordId());

      if (parent.isOverflow()) {
         split(parent, parents, parent_index + 1);
      } else {
         parent.update();
      }
   }

   private void updateParentKeys(String key,
                                 Match[] parents,
                                 int index)
   throws IOException, StorageException
   {
      for (int i = index; i < parents.length; i++) {
         parents[i].setKey(key);
         parents[i].node().update();
         if (!parents[i].isLastIndex()) {
            break;
         }
      }
   }

   /**
      Can be used during both insert() and remove().
    */
   private void merge(Node parent,
                      Node sibling,
                      int sibling_index,
                      Node node,
                      int node_index)
   throws IOException, StorageException
   {
      if (TRACE_FLOW) debug("before: " + sibling + "/" + sibling_index + " *** " + node + "/" + node_index);
      if (TRACE_FLOW) debug("parent: " + parent);
      if (TRACE_FLOW) debug("sibling_index: " + sibling_index);
      if (TRACE_FLOW) debug("sibling: " + sibling);
      if (TRACE_FLOW) debug("node_index: " + node_index);
      if (TRACE_FLOW) debug("node: " + node);

      if (Assert.DEBUG) Assert.pre(node.canMerge(sibling));
      if (Assert.DEBUG) Assert.pre(node != sibling);
      if (Assert.DEBUG) Assert.pre(parent.key(sibling_index).equals(sibling.lastKey()));

      if (sibling.greaterThan(node)) {
         node.merge(sibling);
         parent.setKey(node_index, node.lastKey());
      } else {
         sibling.merge(node);
         parent.setKey(sibling_index, sibling.lastKey());
      }

      if (TRACE_FLOW) debug("after : " + sibling + "/" + sibling_index + " *** " + node + "/" + node_index);
      if (TRACE_FLOW) debug("parent: " + parent);
      if (TRACE_FLOW) debug("sibling_index: " + sibling_index);
      if (TRACE_FLOW) debug("sibling: " + sibling);
      if (TRACE_FLOW) debug("node_index: " + node_index);
      if (TRACE_FLOW) debug("node: " + node);


      node.update();
      sibling.update();
      parent.update();
   }

   /**
      Can be used only during remove().
    */
   private void absorb(Node parent,
                       Node sibling,
                       int sibling_index,
                       Node node,
                       int node_index,
                       Match[] parents,
                       int parent_index)
   throws IOException, StorageException
   {
      if (Assert.DEBUG) Assert.pre(node != sibling);
      if (Assert.DEBUG) Assert.pre(node.canAbsorb(sibling));

      if (sibling.greaterThan(node)) {
         sibling.absorb(node);
         node.release();
         sibling.update();

         if (TRACE_FLOW) debug("recursively removing absorbed sibling");
         remove(new Match(parent, node_index), parents, parent_index + 1);
      } else {
         node.absorb(sibling);
         sibling.release();
         node.update();

         if (TRACE_FLOW) debug("recursively removing absorbed sibling");
         remove(new Match(parent, sibling_index), parents, parent_index + 1);
      }
   }

   public synchronized void put(String key,
                                int record_id)
   throws IOException, StorageException, DuplicateKeyException
   {
      Stack stack = new Stack();
      Match match = find(key, stack, true);
      if (match != null) {
         throw new DuplicateKeyException(key);
      }

      Match[] parents = stackToArray(stack);
      Node node = parents[0].node();
      if (Assert.DEBUG) Assert.pre(node.isLeaf());

      // Insert the new key into the node and update parents that are
      // affected if the new key is the last in the node.
      int index = node.insertChild(key, record_id);
      if (node.isLastIndex(index)) {
         updateParentKeys(key, parents, 1);
      }

      // Handle overflow.
      if (node.isOverflow()) {
         if (parents.length == 1) {
            if (TRACE_FLOW) debug("handling overflow using split");
            split(node, parents, 1);
         } else {
            Node parent   = parents[1].node();
            int node_index = parents[1].index();
            int sibling_index = parent.getSibling(node_index);
            Node sibling = parent.getChild(sibling_index);
            if (Assert.DEBUG) Assert.pre(node != sibling);

            if (sibling.canMerge(node)) {
               if (TRACE_FLOW) debug("handling overflow using merge");
               merge(parent, sibling, sibling_index, node, node_index);
            } else {
               if (TRACE_FLOW) debug("handling overflow using split");
               split(node, parents, 1);
            }
         }
      } else {
         node.update();
      }

      if (CHECK_SANITY) checkTreeSanity(getRoot());
   }

   private void removeSingleChildRoot(Node root)
   throws IOException, StorageException
   {
      int link = root.link(0);

      // release the old root
      root.release();

      // load the new root and delete its current record
      root = getNode(link);
      root.release();

      // now reappend the record to make it the new root
      root.append();
      if (Assert.DEBUG) Assert.post(root.recordId() == ROOT_ID);
   }

   private void remove(Match match,
                       Match[] parents,
                       int parent_index)
   throws IOException, StorageException, UnknownKeyException
   {
      Node node = match.node();
      String key = match.key();
      boolean is_last = match.isLastIndex();

      if (TRACE_FLOW) debug("remove: key " + key);
      if (!node.isRoot()) {
         Match p = parents[parent_index];
         Node  parent = p.node();
         if (TRACE_FLOW) debug("node  : " + node);
         if (TRACE_FLOW) debug("parent: " + parent);
         int index = p.index();
         Node n2 = parent.getChild(index);
         if (Assert.DEBUG) Assert.pre(node == n2);
      }

      node.removeChild(match.index());
      if (TRACE_FLOW) debug("middle node: " + node);
      
      if (is_last && !node.isRoot()) {
         if (TRACE_FLOW) debug("remove: key was last key, patching...");
         updateParentKeys(node.lastKey(), parents, parent_index);
      }

      if (node.isUnderflow()) {
         if (TRACE_FLOW) debug("delete underflow handling begins");

         if (node.isRoot()) {
            if ((node.numKeys() == 1) && !node.isLeaf()) {
               if (TRACE_FLOW) debug("remove: removing root with one child");
               removeSingleChildRoot(node);
            } else {
               node.update();
            }
            if (TRACE_FLOW) debug("underflow delete from root completed");
         } else {
            Node parent    = parents[parent_index].node();
            if (TRACE_FLOW) debug("parent: " + parent);
            int node_index = parents[parent_index].index();
            if (Assert.DEBUG) Assert.pre(node == parent.getChild(node_index));
            int sibling_index = parent.getSibling(node_index);
            if (TRACE_FLOW) debug("sibling_index: " + sibling_index);
            Node sibling = parent.getChild(sibling_index);

            if (sibling.canMerge(node)) {
               if (TRACE_FLOW) debug("handling underflow using merge");
               merge(parent, sibling, sibling_index, node, node_index);
            } else {
               if (TRACE_FLOW) debug("handling underflow using absorb");
               absorb(parent, sibling, sibling_index, node, node_index, parents, parent_index);
            }
            if (TRACE_FLOW) debug("delete underflow handling complete!");
         }
      } else {
         node.update();
      }

      if (CHECK_SANITY) checkTreeSanity(getRoot());
   }

   public synchronized int remove(String key)
   throws IOException, StorageException, UnknownKeyException
   {
      Stack stack = new Stack();
      Match match = find(key, stack, true);
      if (match == null) {
         throw new UnknownKeyException(key);
      }

      int answer = match.link();

      stack.pop();
      Match[] parents = stackToArray(stack);
      remove(match, parents, 0);

      if (CHECK_SANITY) checkTreeSanity(getRoot());

      return answer;
   }

   public synchronized void update(String key,
                                   int record_id)
   throws IOException, StorageException, UnknownKeyException
   {
      Match match = find(key, new Stack(), true);
      if (match == null) {
         throw new UnknownKeyException(key);
      }

      Node node = match.node();
      node.setLink(match.index(), record_id);
      node.update();

      if (CHECK_SANITY) checkTreeSanity(getRoot());
   }

   private class TreeWalker
   implements IndexNavigator {
      TreeWalker(String min_key)
      {
         m_minKey  = min_key;
         m_stack   = null;
         m_current = null;
      }

      TreeWalker(String min_key,
                 Stack stack,
                 Match match)
      {
         m_minKey  = min_key;
         m_stack   = stack;
         m_current = match;
         m_stack.pop();
      }

      public StorageIndex getIndex()
      {
         return BTreeIndex.this;
      }

      public String getKey()
      throws IOException, StorageException
      {
         synchronized (BTreeIndex.this)
         {
            return m_current.key();
         }
      }

      public int getRecordId()
      throws IOException, StorageException
      {
         synchronized (BTreeIndex.this)
         {
            if (TRACE_FLOW) debug("TW.getRecordId returns " + m_current.link());
            return m_current.link();
         }
      }

      public String getMinKey()
      {
         synchronized (BTreeIndex.this)
         {
            return m_minKey;
         }
      }

      public boolean isValid()
      {
         synchronized (BTreeIndex.this)
         {
            return m_current != null;
         }
      }

      public void next()
      throws IOException, StorageException
      {
         synchronized (BTreeIndex.this)
         {
            if (TRACE_FLOW) debug("next: " + m_current);
            if (Assert.DEBUG) Assert.pre(m_current != null);

            if (!m_current.isLastIndex()) {
               m_current.next();
               return;
            }

            if (m_stack.empty()) {
               if (Assert.DEBUG) Assert.pre(m_current.node().isRoot());
               if (Assert.DEBUG) Assert.pre(m_current.node().isLeaf());
               m_current = null;
               m_stack = null;
               return;
            }

            Match parent = (Match)m_stack.pop();
            if (Assert.DEBUG) Assert.pre(!parent.node().isLeaf());

            while (parent.isLastIndex()) {
               if (m_stack.empty()) {
                  m_current = null;
                  m_stack = null;
                  return;
               }
               parent = (Match)m_stack.pop();
               if (Assert.DEBUG) Assert.pre(!parent.node().isLeaf());
            }

            parent.next();
            m_stack.push(parent);

            Node child = getNode(parent.link());
            while (!child.isLeaf()) {
               m_stack.push(new Match(child, 0));
               child = getNode(child.link(0));
            }

            m_current = new Match(child, 0);
         }
      }

      public void close()
      {
         synchronized (BTreeIndex.this)
         {
            m_minKey = null;
            m_stack = null;
            m_current = null;
         }
      }

      private String m_minKey;
      private Stack m_stack;
      private Match m_current;
   }

   public synchronized IndexNavigator navigate(String min_key_prefix)
   throws IOException, StorageException
   {
      Stack stack = new Stack();
      Match match = find(min_key_prefix, stack, false);
      if (match == null) {
         return new TreeWalker(min_key_prefix);
      }
      return new TreeWalker(min_key_prefix, stack, match);
   }

   private static void debug(String msg)
   {
      System.out.println("BTI: " + msg);
   }

   private void checkTreeSanity(Node node)
   {
      node.checkSanity();
      if (!node.isLeaf()) {
         try {
            for (int i = 0; i < node.m_numKeys; i++) {
               checkTreeSanity(node.getChild(i));
            }
         } catch (IOException ex) {
            ex.printStackTrace();
            Assert.impossible(ex);
         }
      }
   }

   private Storage m_storage;
   private int      m_keyLength;
   private int      m_recordLength;
   private LRUCache m_cache;
   private int      m_keysPerNode;

   private static final int ROOT_ID = 0;
   private static final int NOT_FOUND = -2;
   private static final boolean TRACE_FLOW   = false;
   private static final boolean CHECK_SANITY = Assert.DEBUG;
}
