///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility.storage;

import java.io.IOException;
import java.util.Stack;
import org.cooldevtools.utility.Assert;
import org.cooldevtools.utility.LRUCache;

public class BinaryTreeIndex
  implements StorageIndex
{
  public BinaryTreeIndex(Storage storage,
			 int key_length)
    throws IOException, StorageException
    {
      m_storage = storage;
      m_keyLength = key_length;
      m_recordLength = calcRecordLength(key_length);
      m_cache = new LRUCache(256);
    }

  public static int calcRecordLength(int key_length)
    {
      return 4 + 2 * key_length + 12;
    }

  private final class Node
  {
    public Node(int record_id,
		String key,
		int left,
		int right,
		int value)
      {
	m_key      = key;
	m_left     = left;
	m_right    = right;
	m_value    = value;
	m_recordId = record_id;
      }

    public Node(int record_id,
		byte[] bytes)
      throws IOException
      {
	m_recordId = record_id;
	m_key   = Packer.getUString(bytes, 0);
	m_left  = Packer.getInt(bytes, 2 * m_keyLength + 4);
	m_right = Packer.getInt(bytes, 2 * m_keyLength + 8);
	m_value = Packer.getInt(bytes, 2 * m_keyLength + 12);
      }

    public byte[] toByteArray()
      throws IOException
      {
	byte[] bytes = new byte[m_recordLength];
	Packer.putUString(m_key, m_keyLength, bytes, 0);
	Packer.putInt(m_left,  bytes, 2 * m_keyLength + 4);
	Packer.putInt(m_right, bytes, 2 * m_keyLength + 8);
	Packer.putInt(m_value, bytes, 2 * m_keyLength + 12);
	return bytes;
      }

    public String toString()
      {
	return("Node('" + m_key + "'"
	       + ","
	       + m_recordId
	       + ","
	       + m_left
	       + ","
	       + m_right
	       + ","
	       + m_value
	       + ")");
      }

    public String key()
      {
	return m_key;
      }

    public int left()
      {
	return m_left;
      }

    public int right()
      {
	return m_right;
      }

    public int value()
      {
	return m_value;
      }

    public int recordId()
      {
	return m_recordId;
      }

    public void write()
      throws IOException, StorageException
      {
	m_storage.set(m_recordId, toByteArray());
      }

    public Node goLeft()
      throws IOException, StorageException, BadRecordIdException
      {
	if (m_left == Storage.NULL)
	  {
	    return null;
	  }
	return loadNode(m_left);
      }

    public Node goRight()
      throws IOException, StorageException, BadRecordIdException
      {
	if (m_right == Storage.NULL)
	  {
	    return null;
	  }
	return loadNode(m_right);
      }

    public void setLeftChild(Node child)
      throws IOException, StorageException
      {
	m_left = child.recordId();
	write();
      }

    public void setLeftChild(String key,
			     int value)
      throws IOException, StorageException
      {
	Node child = new Node(-1, key, Storage.NULL, Storage.NULL, value);
	m_left = m_storage.put(child.toByteArray());
	write();
      }

    public void setRightChild(Node child)
      throws IOException, StorageException
      {
	m_right = child.recordId();
	write();
      }

    public void setRightChild(String key,
			      int value)
      throws IOException, StorageException
      {
	Node child = new Node(-1, key, Storage.NULL, Storage.NULL, value);
	m_right = m_storage.put(child.toByteArray());
	write();
      }

    public void setValue(int value)
      throws IOException, StorageException
      {
	m_value = value;
	write();
      }

    
    public void release()
      throws IOException, StorageException
      {
	m_left = Storage.NULL;
	m_right = Storage.NULL;
	m_storage.release(m_recordId);
	m_cache.remove(new Integer(m_recordId));
      }

    public void forget(Node child)
      throws IOException, StorageException
      {
	if (m_left == child.recordId())
	  {
	    m_left = Storage.NULL;
	  }
	else if (m_right == child.recordId())
	  {
	    m_right = Storage.NULL;
	  }
	else
	  {
	    Assert.impossible();
	  }
	write();
      }

    private String m_key;
    private int m_left;
    private int m_right;
    private int m_value;
    private int m_recordId;
  }

  private Node cacheNode(Node n)
    {
      m_cache.put(new Integer(n.recordId()), n);
      return n;
    }

  private Node loadNode(int record_id)
    throws IOException, StorageException
    {
      Integer key = new Integer(record_id);
      Object o = m_cache.get(key);
      if (o == null)
	{
	  return cacheNode(new Node(record_id, m_storage.get(record_id)));
	}

      return (Node)o;
    }

  private Node getRoot()
    throws IOException, StorageException
    {
      try
	{
	  return loadNode(ROOT_ID);
	}
      catch (BadRecordIdException ex)
	{
	  Node root = new Node(ROOT_ID, "", Storage.NULL, Storage.NULL, Storage.NULL);
	  int id = m_storage.put(root.toByteArray());
	  if (id != ROOT_ID)
	    {
	      throw new RuntimeException("storage returned non-zero record id");
	    }
	  return cacheNode(root);
	}
    }

  public int getKeyLength()
    {
      return m_keyLength;
    }

  public void close()
    {
      m_storage.close();
    }

  public void commitTo(CommitManager cmgr)
  throws IOException
  {
     m_storage.commitTo(cmgr);
  }

  public void commitWrite()
  throws IOException
  {
     m_storage.commitWrite();
  }

  public void commitFailed()
  {
     m_storage.commitFailed();
     m_cache.clear();
  }

  public void commitSuccessful()
  {
     m_storage.commitSuccessful();
  }
  
  public void commitRecovered()
  {
     m_storage.commitRecovered();
     m_cache.clear();
  }

  public void sync()
    throws IOException, StorageException
    {
      m_storage.sync();
    }

  public void abort()
    {
      m_storage.abort();
      m_cache.clear();
    }

  private Node find(String key)
    throws IOException, StorageException, UnknownKeyException
    {
      Node node = getRoot().goRight();
      while (node != null)
	{
	  int diff = node.key().compareTo(key);
	  debug("find: " + key + " " + node + " diff=" + diff);
	  if (diff == 0)
	    {
	      return node;
	    }

	  if (diff < 0)
	    {
	      node = node.goRight();
	    }
	  else
	    {
	      node = node.goLeft();
	    }
	}

      throw new UnknownKeyException(key);
    }

  public synchronized boolean exists(String key)
    throws IOException, StorageException
    {
      try
        {
          find(key);
          return true;
        }
      catch (UnknownKeyException ex)
        {
          return false;
        }
    }

  public synchronized int get(String key)
    throws IOException, StorageException, UnknownKeyException
    {
      return find(key).value();
    }

  public synchronized void put(String key,
			       int record_id)
    throws IOException, StorageException, DuplicateKeyException
    {
      Node node = getRoot();
      while (true)
	{
	  int diff = node.key().compareTo(key);
	  debug("put scan : " + key + " " + node + " diff=" + diff);
	  if (diff == 0)
	    {
	      throw new DuplicateKeyException(key);
	    }

	  if (diff < 0)
	    {
	      if (node.right() != Storage.NULL)
		{
		  node = node.goRight();
		}
	      else
		{
		  debug("put right: " + key + " " + node + " diff=" + diff);
		  node.setRightChild(key, record_id);
		  break;
		}
	    }
	  else
	    {
	      if (node.left() != Storage.NULL)
		{
		  node = node.goLeft();
		}
	      else
		{
		  debug("put left : " + key + " " + node + " diff=" + diff);
		  node.setLeftChild(key, record_id);
		  break;
		}
	    }
	}
      debug("PUT COMPLETE");
    }

  private void put(Node node,
		   Node child)
    throws IOException, StorageException, DuplicateKeyException
    {
      debug("put2: inserting  " + child + " under " + node);
      while (true)
	{
	  int diff = node.key().compareTo(child.key());
	  debug("put2 scan : " + child + " " + node + " diff=" + diff);
	  if (diff == 0)
	    {
	      throw new DuplicateKeyException(node.key());
	    }

	  if (diff < 0)
	    {
	      if (node.right() != Storage.NULL)
		{
		  node = node.goRight();
		}
	      else
		{
		  debug("put2 right: " + child + " " + node + " diff=" + diff);
		  node.setRightChild(child);
		  break;
		}
	    }
	  else
	    {
	      if (node.left() != Storage.NULL)
		{
		  node = node.goLeft();
		}
	      else
		{
		  debug("put2 left : " + child + " " + node + " diff=" + diff);
		  node.setLeftChild(child);
		  break;
		}
	    }
	}
    }

  private void removeNode(Node parent,
			  Node child)
    throws IOException, StorageException
    {
      debug("RN: parent " + parent + " child " + child);

      parent.forget(child);
      debug("RN: parent after forget " + parent);

      if (child.left() != Storage.NULL)
	{
	  put(parent, child.goLeft());
	}

      if (child.right() != Storage.NULL)
	{
	  put(parent, child.goRight());
	}

      child.release();
    }

  private Node removeNode(String key)
    throws IOException, StorageException, UnknownKeyException
    {
      Node parent = getRoot();
      Node node = parent.goRight();
      while (node != null)
	{
	  int diff = node.key().compareTo(key);
	  if (diff == 0)
	    {
	      removeNode(parent, node);
	      return node;
	    }

	  parent = node;
	  if (diff < 0)
	    {
	      node = parent.goRight();
	    }
	  else
	    {
	      node = parent.goLeft();
	    }
	}

      throw new UnknownKeyException(key);
    }

  public synchronized int remove(String key)
    throws IOException, StorageException, UnknownKeyException
    {
      return removeNode(key).value();
    }

  public synchronized void update(String key,
				  int record_id)
    throws IOException, StorageException, UnknownKeyException
    {
      find(key).setValue(record_id);
    }

  private static class StackFrame
  {
    public StackFrame(Node node,
		      int status)
      {
	this.node = node;
	this.status = status;
      }

    public Node node;
    public int status;

    public static final int GONE_LEFT  = 0;
    public static final int VISITED    = 1;
    public static final int GONE_RIGHT = 2;
  }

  private static Node findNextAfter(Node root,
				    String min_key,
				    Stack stack)
    throws IOException, StorageException, UnknownKeyException
    {
      Node node = root;
      while (node != null)
	{
	  int diff = node.key().compareTo(min_key);
	  if (diff == 0)
	    {
	      stack.push(new StackFrame(node, StackFrame.VISITED));
	      return node;
	    }

	  if (diff < 0)
	    {
	      stack.push(new StackFrame(node, StackFrame.GONE_RIGHT));
	      node = node.goRight();
	    }
	  else
	    {
	      stack.push(new StackFrame(node, StackFrame.GONE_LEFT));
	      node = node.goLeft();
	    }
	}

      while (!stack.empty())
	{
	  StackFrame frame = (StackFrame)stack.pop();
	  if (frame.status == StackFrame.GONE_LEFT)
	    {
	      frame.status = StackFrame.VISITED;
	      stack.push(frame);
	      return frame.node;
	    }
	}

      throw new UnknownKeyException(min_key);
    }

  private class TreeWalker
    implements IndexNavigator
  {
    TreeWalker(String min_key)
      {
	m_minKey = min_key;
	m_stack = null;
	m_node = null;
      }

    TreeWalker(String min_key,
	       Stack stack,
	       Node node)
      {
	m_minKey = min_key;
	m_stack = stack;
	m_node = node;
      }

    public StorageIndex getIndex()
      {
	return BinaryTreeIndex.this;
      }

    public String getKey()
      {
	return m_node.key();
      }

    public int getRecordId()
      {
	return m_node.value();
      }

    public String getMinKey()
      {
	return m_minKey;
      }

    public boolean isValid()
      {
	return m_node != null;
      }

    public void nextLeft()
      {
	while (!m_stack.empty())
	  {
	    StackFrame frame = (StackFrame)m_stack.pop();
	    if (frame.status == StackFrame.GONE_LEFT)
	      {
		debug("nextLeft: visiting node: " + frame.node);
		frame.status = StackFrame.VISITED;
		m_stack.push(frame);
		m_node = frame.node;
		return;
	      }
	    debug("nextLeft: discarded frame with status " + frame.status + ": " + frame.node);
	  }
	m_node = null;
      }

    public void goLeft(Node node)
      throws IOException, StorageException
      {
	while (node.left() != Storage.NULL)
	  {
	    debug("goLeft: pushing node: " + node);
	    m_stack.push(new StackFrame(node, StackFrame.GONE_LEFT));
	    node = node.goLeft();
	  }

	debug("goLeft: visiting node: " + node);
	m_stack.push(new StackFrame(node, StackFrame.VISITED));
	m_node = node;
      }

    public void next()
      throws IOException, StorageException
      {
	if (Assert.DEBUG) Assert.pre(m_node != null);

	StackFrame frame = (StackFrame)m_stack.pop();
	if (Assert.DEBUG) Assert.pre(frame.node == m_node);
	if (Assert.DEBUG) Assert.pre(frame.status == StackFrame.VISITED);

	if (m_node.right() == Storage.NULL)
	  {
	    nextLeft();
	  }
	else
	  {
	    frame.status = StackFrame.GONE_RIGHT;
	    m_stack.push(frame);
	    goLeft(frame.node.goRight());
	  }
      }

    public void close()
      {
	m_minKey = null;
	m_stack = null;
	m_node = null;
      }

    private String m_minKey;
    private Stack m_stack;
    private Node m_node;
  }

  public synchronized IndexNavigator navigate(String min_key_prefix)
    throws IOException, StorageException
    {
      try
	{
	  Stack stack = new Stack();
	  Node node = findNextAfter(getRoot().goRight(), min_key_prefix, stack);
	  return new TreeWalker(min_key_prefix, stack, node);
	}
      catch (UnknownKeyException ex)
	{
	  return new TreeWalker(min_key_prefix);
	}
    }

  private static void debug(String msg)
    {
      //System.out.println("BTI: " + msg);
    }

  private Storage m_storage;
  private int m_keyLength;
  private int m_recordLength;
  private LRUCache m_cache;

  private static final int ROOT_ID = 0;
}
