///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility.storage;

import java.io.*;
import java.util.Random;
import org.cooldevtools.utility.*;

public class BufferedRandomAccessFile
implements LRUPoolFactory, Commitable, BinaryComparator
{
   public BufferedRandomAccessFile(String filename)
   throws IOException
   {
      this(filename, -1);
   }

   public BufferedRandomAccessFile(String filename,
                                   int num_buffers)
   throws IOException
   {
      m_filename = filename;
      m_file     = new RandomAccessFile(filename, "rw");
      m_length   = m_file.length();

      m_input  = new MyInputStream();
      m_output = new MyOutputStream();
   
      if (num_buffers <= 0) {
         num_buffers = calcNumBuffers();
      }

      m_bufferPool = new LRUPool(this, num_buffers);
      m_bufferTable = new SuperHashtable();
      m_sorter = new Sorter(this);
      
      m_buffer = (Buffer)m_bufferPool.getNextAvailable();
      m_buffer.moveTo(0);
   }

   public LRUsable createPoolItem()
   {
      return new Buffer();
   }

   public static void main(String[] args)
   throws Exception
   {
      final String filename = "_test.dat";
      new File(filename).delete();

      BufferedRandomAccessFile file = new BufferedRandomAccessFile(filename);
      byte[] b = new byte[217];
      for (int i = 0; i < b.length; i++) {
         b[i] = (byte)i;
      }

      final int buffers = 11;
      for (int i = 0; i < buffers; i++) {
         file.write(b, 0, b.length);
         file.flush();
      }
      file.flush();

      final int length = buffers * b.length;
      Random r = new Random();
      for (int i = 0; i < length; i++) {
         int loc = Math.abs(r.nextInt()) % length;
         if ((loc + b.length) > length) {
            loc = length - b.length;
         }
         byte[] tmp = new byte[b.length];
         
         System.out.println("seeking position " + loc);
         file.seek(loc);
         int n = file.read(tmp, 0, tmp.length);
         if (n != tmp.length) {
            System.out.println("error: expected to read " + tmp.length + " but read " + n);
         }
         for (int k = 0; k < tmp.length; k++) {
            byte expect = (byte)((loc + k) % b.length);
            if (tmp[k] != expect) {
               System.out.println("error: expected " + expect + " found " + tmp[k]);
            }
         }
      }
   }

   public final class Buffer
   implements LRUsable
   {
      public Buffer()
      {
         m_data    = new byte[BUFFER_SIZE];
	 reset();
      }

      public void reset()
      {
         m_start      = -1;
         m_index      = 0;
         m_length     = 0;
         m_isDirty    = false;
         m_dirtyStart = -1;
         m_dirtyEnd   = -2;
      }

      public boolean isDirty()
      {
         return m_isDirty;
      }

      public void setNotDirty()
      {
         m_isDirty    = false;
         m_dirtyStart = -1;
         m_dirtyEnd   = -2;
      }

      public boolean isAvailableForUse()
      {
         return !m_isDirty;
      }

      private void markRegionDirty(int index,
				   int length)
      {
	 int last = index + length;
	 if (m_isDirty) {
	    if (m_dirtyStart > index) {
	       m_dirtyStart = index;
	    }
	    if (m_dirtyEnd < last) {
	       m_dirtyEnd = last;
	    }
	 } else {
	    m_dirtyStart = index;
	    m_dirtyEnd   = last;
	    m_isDirty = true;
	 }
      }

      public int write(byte[] buffer,
                       int offset,
                       int length)
      {
         if (DEBUG) System.out.println("file " + m_filename + " write " + length + " at position " + getCurrentPosition());

         int nbytes = m_data.length - m_index;
         if (nbytes > length) {
            nbytes = length;
         }

         if (nbytes == 0) {
            return 0;
         }

	 markRegionDirty(m_index, nbytes);
	 if (Assert.DEBUG) Assert.post(m_isDirty);
	 if (Assert.DEBUG) Assert.post(m_dirtyStart <= m_index);
	 if (Assert.DEBUG) Assert.post((m_dirtyEnd - m_dirtyStart) >= nbytes);

         System.arraycopy(buffer, offset, m_data, m_index, nbytes);
         m_index += nbytes;
         if (m_index > m_length) {
            m_length = m_index;
         }

         return nbytes;
      }

      public int read(byte[] buffer,
                      int offset,
                      int length)
      {
         if (DEBUG) System.out.println("read " + length + " at position " + getCurrentPosition());

         int nbytes = m_length - m_index;
         if (nbytes > length) {
            nbytes = length;
         }

         if (nbytes == 0) {
            return 0;
         }

         System.arraycopy(m_data, m_index, buffer, offset, nbytes);
         m_index += nbytes;

         return nbytes;
      }

      public String toString()
      {
         StringBuffer b = new StringBuffer();
         b.append("Buffer{");
         b.append(m_data);
         b.append(",");
         b.append(m_filename);
         b.append(",start=");
         b.append(m_start);
         b.append(",length=");
         b.append(m_length);
         b.append(",dStart=");
         b.append(m_dirtyStart);
         b.append(",dEnd=");
         b.append(m_dirtyEnd);
         b.append("}");
         return b.toString();
      }

      public boolean containsPosition(long pos)
      throws IOException
      {
         // check to see if we have any data at all
         if (m_start < 0) {
            return false;
         }

         // not if the data precedes the start of ours
         if (m_start > pos) {
            return false;
         }

         // not if the data is past the end of ours
         long last = m_start + m_length;
         if (pos > last) {
            return false;
         }

         // yes if the data is in ours
         if (pos < last) {
            return true;
         }

         // otherwise the data is at the end of ours so
         // yes iff we have room to grow and are the last
         // buffer in our owner's chain
         return (m_length < m_data.length) && (pos == length());
      }

      public long getStartPosition()
      {
         return m_start;
      }

      public long getCurrentPosition()
      {
         return m_start + m_index;
      }

      public long getEndPosition()
      {
         return m_start + m_length;
      }

      public void flush()
      throws IOException
      {
         if (m_isDirty) {
	    if (Assert.DEBUG) Assert.pre(m_dirtyEnd > m_dirtyStart);
	    if (Assert.DEBUG) Assert.pre(m_dirtyEnd <= m_length);
	    if (Assert.DEBUG) Assert.pre(m_dirtyStart >= 0);
            if (DEBUG) System.out.println("flush buffer: " + this);
            m_file.seek(m_start + m_dirtyStart);
            m_file.write(m_data, m_dirtyStart, m_dirtyEnd - m_dirtyStart);
            m_isDirty = false;
         }
      }

      public void commitTo(CommitManager cmgr)
      throws IOException
      {
         if (m_isDirty) {
	    if (Assert.DEBUG) Assert.pre(m_dirtyEnd > m_dirtyStart);
	    if (Assert.DEBUG) Assert.pre(m_dirtyEnd <= m_length);
	    if (Assert.DEBUG) Assert.pre(m_dirtyStart >= 0);
            if (DEBUG) System.out.println("flush buffer: " + this);
	    cmgr.addCommitEntry(m_filename, m_start + m_dirtyStart,
				m_data, m_dirtyStart, m_dirtyEnd - m_dirtyStart);
         }
      }

      public void moveTo(long pos)
      throws IOException
      {
         if (containsPosition(pos)) {
            m_index = (int)(pos - m_start);
         } else {
            if (Assert.DEBUG) Assert.pre(!m_isDirty);

            long file_length = length();
            if (pos > file_length) {
               throw new EOFException("file_length=" + file_length + " pos=" + pos);
            }

            long old_start = m_start;
            m_start = pos - (pos % m_data.length);
            if (Assert.DEBUG) Assert.pre(old_start != m_start);

            if (old_start >= 0) {
               m_bufferTable.remove(keyForPosition(old_start));
            }
            m_bufferTable.put(keyForPosition(m_start), this);
            
            if (m_start == file_length) {
               m_index  = 0;
               m_length = 0;
            } else {
               m_file.seek(m_start);
               m_length = m_file.read(m_data);
               m_index  = (int)(pos - m_start);
               if (DEBUG) System.out.println("read " + m_length + " bytes from " + m_filename);
            }
         }

         m_bufferPool.use(this);
         
         if (Assert.DEBUG) Assert.post((m_start % m_data.length) == 0);
         if (Assert.DEBUG) Assert.post((m_start + m_index) == pos);
         if (Assert.DEBUG) Assert.post(m_index <= m_length);
         if (Assert.DEBUG) Assert.post(m_length <= m_data.length);
      }

      private boolean m_isDirty;
      private long    m_start;
      private int     m_index;
      private int     m_length;
      private byte[]  m_data;
      private int     m_dirtyStart;
      private int     m_dirtyEnd;
   }

   private class MyInputStream
   extends InputStream {
      public int available()
      throws IOException
      {
         return BufferedRandomAccessFile.this.available();
      }

      public int read()
      throws IOException
      {
         synchronized (BufferedRandomAccessFile.this)
         {
            int n = BufferedRandomAccessFile.this.read(m_single, 0, 1);
            return (n == 1) ? m_single[0] : -1;
         }
      }

      public int read(byte[] b,
                      int offset,
                      int length)
      throws IOException
      {
         return BufferedRandomAccessFile.this.read(b, offset, length);
      }

      private byte[] m_single = new byte[1];
   }

   private class MyOutputStream
   extends OutputStream {
      public void write(int b)
      throws IOException
      {
         synchronized (BufferedRandomAccessFile.this)
         {
            m_single[0] = (byte)b;
            BufferedRandomAccessFile.this.write(m_single, 0, 1);
         }
      }

      public void write(byte[] b,
                        int offset,
                        int length)
      throws IOException
      {
         BufferedRandomAccessFile.this.write(b, offset, length);
      }

      public void flush()
      throws IOException
      {
         BufferedRandomAccessFile.this.flush();
      }

      public void close()
      throws IOException
      {
         // just flush on close because we don't really close
         // the file until BufferedRandomAccessFile.close() is
         // called explicitly.
         BufferedRandomAccessFile.this.flush();
      }

      private byte[] m_single = new byte[1];
   }

   public InputStream getInput()
   {
      return m_input;
   }

   public OutputStream getOutput()
   {
      return m_output;
   }

   public synchronized int available()
   throws IOException
   {
      return (int)(m_buffer.getEndPosition() - m_buffer.getCurrentPosition());
   }

   public synchronized long getFilePointer()
   {
      return m_buffer.getCurrentPosition();
   }

   private Long keyForPosition(long pos)
   {
      long key_value = pos / BUFFER_SIZE;
      Long key = new Long(key_value);
      if (DEBUG) System.out.println("hashCode for " + pos + "= " + key.hashCode());
      return key;
   }

   public synchronized void seek(long position)
   throws IOException
   {
      if (Assert.DEBUG) Assert.pre(m_buffer != null);
      if (Assert.DEBUG) Assert.pre(m_bufferTable.size() <= m_bufferPool.size());

      if (position > m_length) {
         throw new EOFException();
      }

      if (m_buffer.containsPosition(position)) {
         if (DEBUG) System.out.println("seek(" + position + ") reusing current m_buffer " + m_buffer);
         m_buffer.moveTo(position);
         return;
      }

      m_buffer = (Buffer)m_bufferTable.get(keyForPosition(position));
      if (m_buffer != null) {
         if (DEBUG) System.out.println("seek(" + position + ") reusing buffer from table " + m_buffer);
         if (Assert.DEBUG) Assert.pre(m_buffer.containsPosition(position));
         m_buffer.moveTo(position);
         return;
      }

      m_buffer = (Buffer)m_bufferPool.getNextAvailable();
      if (DEBUG) System.out.println("seek(" + position + ") using buffer from pool " + m_buffer);
      m_buffer.moveTo(position);
   }

   public synchronized void close()
   throws IOException
   {
      flush();
      m_file.close();

      m_file = null;
      m_buffer = null;
      m_bufferTable = null;
   }

   private int calcNumBuffers()
   {
      int num_buffers = (int)(m_length / BUFFER_SIZE / 100);
      if (DEBUG) System.out.println("calcnumbuffers for " + m_filename + " = " + num_buffers);
      return (num_buffers >= 0) ? num_buffers : 1;
   }

   public synchronized void abort()
   {
      // clear the buffer cache and erase all changes
      Buffer[] buffers = getBuffers();
      for (int i = 0; i < buffers.length; i++) {
         buffers[i].reset();
      }
      m_bufferTable.clear();

      // reset the length to the physical length
      // this really should never fail!
      try {
         m_length = m_file.length();
      } catch (IOException ex) {
         ex.printStackTrace();
         throw new RuntimeException(ex.getMessage());
      }
   }

   private Buffer[] scheduleWrites()
   {
      LinkedList writes = new LinkedList();
      Buffer[] buffers = getBuffers();
      for (int i = 0; i < buffers.length; i++) {
         Buffer b = buffers[i];
         if (b.isDirty()) {
	    writes.append(b);
         }
      }

      buffers = new Buffer[writes.size()];
      writes.copyInto(buffers);
      m_sorter.sort(buffers);
      
      return buffers;
   }

   public void commitTo(CommitManager cmgr)
   throws IOException
   {
      Buffer[] buffers = scheduleWrites();
      for (int i = 0; i < buffers.length; i++) {
         buffers[i].commitTo(cmgr);
      }
   }

   public void commitWrite()
      throws IOException
   {
      flush();
   }

   public void commitFailed()
   {
      abort();
   }

   public void commitSuccessful()
   {
      Buffer[] buffers = getBuffers();
      for (int i = 0; i < buffers.length; i++) {
         buffers[i].setNotDirty();
      }

      // this really should never fail!
      try {
         if (Assert.DEBUG) Assert.post(m_length == m_file.length());
      } catch (IOException ex) {
         if (Assert.DEBUG) ex.printStackTrace();
         throw new RuntimeException(ex.getMessage());
      }
   }
   
   public void commitRecovered()
   {
      abort();
   }
   
   public synchronized void flush()
   throws IOException
   {
      Buffer[] buffers = scheduleWrites();
      for (int i = 0; i < buffers.length; i++) {
         buffers[i].flush();
      }
      
      buffers = getBuffers();
      for (int i = 0; i < buffers.length; i++) {
         if (Assert.DEBUG) Assert.post(!buffers[i].isDirty());
      }
      m_bufferPool.setMinSize(calcNumBuffers());
      
      if (Assert.DEBUG) Assert.post(m_file.length() == m_length);
   }

   public synchronized int read(byte[] b,
                                int offset,
                                int length)
   throws IOException
   {
      int total_read = 0;
      while (length > 0) {
         int n = m_buffer.read(b, offset, length);
         total_read += n;

         // Did we finish reading or hit end of file?
         if ((n == length) || (n == 0)) {
            return total_read;
         }

         // If we read less than length but more than zero bytes then we must
         // have drained the buffer but not hit EOF.
         if (Assert.DEBUG) Assert.post(m_buffer.getCurrentPosition() == m_buffer.getEndPosition());

         offset += n;
         length -= n;

         // Reposition to read more of the data.
         seek(m_buffer.getEndPosition());
      }

      if (Assert.DEBUG) Assert.post(length == 0);
      return total_read;
   }

   public synchronized void write(byte[] b,
                                  int offset,
                                  int length)
   throws IOException
   {
      while (length > 0) {
         int n = m_buffer.write(b, offset, length);
         if (m_buffer.getCurrentPosition() > m_length) {
            m_length = m_buffer.getCurrentPosition();
         }
         
         if (n == length) {
            return;
         }

         // If we wrote less than length then we must have filled up the buffer.
         if (Assert.DEBUG) Assert.post(m_buffer.getCurrentPosition() == m_buffer.getEndPosition());

         offset += n;
         length -= n;

         // Reposition to write the rest of the data.
         seek(m_buffer.getEndPosition());
      }

      if (Assert.DEBUG) Assert.post(length == 0);
   }

   public synchronized long length()
   throws IOException
   {
      return m_length;
   }

   private Buffer[] getBuffers()
   {
      Buffer[] answer = new Buffer[m_bufferTable.size()];
      m_bufferTable.copyInto(answer);
      return answer;
   }

   public int compare(Object a,
		      Object b)
   {
      try {
	 Buffer b1 = (Buffer)a;
	 Buffer b2 = (Buffer)b;
	 if (Assert.DEBUG) Assert.inv(b1.isDirty());
	 if (Assert.DEBUG) Assert.inv(b2.isDirty());
	 long pos1 = b1.getStartPosition();
	 long pos2 = b2.getStartPosition();
	 if (pos1 == pos2) {
	    return 0;
	 }
	 return (pos1 < pos2) ? -1 : 1;
      } catch (ClassCastException ex) {
	 if (Assert.DEBUG) Assert.impossible("compare failed");
	 return 0;
      }
   }

   private RandomAccessFile m_file;
   private InputStream      m_input;
   private OutputStream     m_output;
   private Buffer           m_buffer;
   private LRUPool          m_bufferPool;
   private SuperHashtable   m_bufferTable;
   private String           m_filename;
   private long             m_length;
   private Sorter           m_sorter;

   private static final int BUFFER_SIZE = 512;
   private static final boolean DEBUG   = false;
}
