///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility.storage;

import java.io.*;
import org.cooldevtools.utility.*;
import org.cooldevtools.utility.logging.*;
import org.cooldevtools.utility.SuperVector;

public class CommitManager
{
   public CommitManager(String dirname,
                        String filename)
   {
      m_file = new File(dirname, filename);
      m_commitables = null;
      m_scratch = new SuperVector();
      m_test_log = (System.getProperties().getProperty("storage.always_apply_log") != null);
      if (m_test_log) {
	 Logger.s().logSilently(MessagePriority.LOG_INFO, "using log for commit");
      } else {
	 Logger.s().logSilently(MessagePriority.LOG_INFO, "using standard commit logic");
      }
   }

   public synchronized void addCommitable(Commitable c)
   {
      m_scratch.addElement(c);
   }

   public synchronized void initialize()
   throws IOException
   {
      m_commitables = new Commitable[m_scratch.size()];
      m_scratch.copyInto(m_commitables);
      m_scratch = null;

      if (m_file.exists()) {
         if (logFileIsValid()) {
            Logger.s().logSilently(MessagePriority.LOG_NOTICE, "recovering previous commit log");
            applyLogFile();
            for (int i = 0; i < m_commitables.length; i++) {
               m_commitables[i].commitRecovered();
            }
         } else {
            Logger.s().logSilently(MessagePriority.LOG_WARNING, "ignoring corrupt commit log");
            m_file.delete();
         }
      }
   }

   public synchronized void addCommitEntry(String filename,
                                           long offset,
                                           byte[] data,
                                           int start,
                                           int length)
   throws IOException
   {
      m_logOutput.writeInt(NEW_ENTRY);
      m_logOutput.writeUTF(filename);
      m_logOutput.writeLong(offset);
      m_logOutput.writeInt(length);
      m_logOutput.write(data, start, length);
      m_crc = CrcCalc.crc32(data, start, length, m_crc);
   }

   private boolean logFileIsValid()
   {
      try {
         DataInputStream in = new DataInputStream(new BufferedInputStream(new FileInputStream(m_file)));
         try {
            int crc = 0;
            int record_type = in.readInt();
            while (record_type == NEW_ENTRY) {
               String filename = in.readUTF();
               long offset = in.readLong();
               int length = in.readInt();
               byte[] data = new byte[length];
               in.readFully(data);
               crc = CrcCalc.crc32(data, crc);
               record_type = in.readInt();
            }
            int file_crc = in.readInt();
            if (DEBUG) System.out.println("crc=" + crc);
            if (DEBUG) System.out.println("filecrc=" + file_crc);
            return (record_type == END_OF_LOG) && (file_crc == crc);
         } finally {
            in.close();
         }
      } catch (IOException ex) {
         return false;
      }
   }

   private void applyLogFile()
   throws IOException
   {
      RandomAccessFile out = null;
      String prev_filename = null;
      DataInputStream in = new DataInputStream(new BufferedInputStream(new FileInputStream(m_file)));
      try {
         int crc = 0;
         int record_type = in.readInt();
         while (record_type == NEW_ENTRY) {
            String filename = in.readUTF();
            long offset = in.readLong();
            int length = in.readInt();
            byte[] data = new byte[length];
            in.readFully(data);
            crc = CrcCalc.crc32(data, crc);

            if ((out == null) || !prev_filename.equals(filename)) {
               if (out != null) {
                  out.close();
               }
               out = new RandomAccessFile(filename, "rw");
               prev_filename = filename;
            }

            out.seek(offset);
            out.write(data);
            record_type = in.readInt();
         }

         int file_crc = in.readInt();
         if ((record_type != END_OF_LOG) || (file_crc != crc)) {
            throw new IOException("log file corrupted");
         }
      } finally {
         in.close();
         if (out != null) {
            out.close();
         }
      }

      m_file.delete();
   }

   private void applyWrites()
   throws IOException
   {
      if (m_test_log) {
         if (DEBUG) System.out.println("applying log file");
	 applyLogFile();
      } else {
         if (DEBUG) System.out.println("normal commit");
	 for (int i = 0; i < m_commitables.length; i++) {
	    m_commitables[i].commitWrite();
	 }
      }
   }

   public synchronized void precommit()
   throws IOException
   {
      try {
         if (DEBUG) System.out.println("precommit");
         m_logOutput = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(m_file)));
         try {
            m_crc = 0;
            for (int i = 0; i < m_commitables.length; i++) {
               m_commitables[i].commitTo(this);
            }
            m_logOutput.writeInt(END_OF_LOG);
            m_logOutput.writeInt(m_crc);
            if (DEBUG) System.out.println("crc=" + m_crc);
         } finally {
            m_logOutput.flush();
            m_logOutput.close();
            m_logOutput = null;
         }
      } catch (IOException ex) {
         m_file.delete();
         throw ex;
      }
   }

   public synchronized void commit()
   throws IOException
   {
      try {
	 applyWrites();
         for (int i = 0; i < m_commitables.length; i++) {
            m_commitables[i].commitSuccessful();
         }
         m_file.delete();
         if (DEBUG) System.out.println("commit complete");
      } catch (IOException ex) {
         for (int i = 0; i < m_commitables.length; i++) {
            m_commitables[i].commitFailed();
         }
         throw ex;
      }
   }

   public synchronized static CommitManager s()
   {
      return s_instance;
   }

   public synchronized static CommitManager initSingleton(String dirname,
                                                          String log_filename)
   {
      if (Assert.DEBUG) Assert.pre(s_instance == null);
      s_instance = new CommitManager(dirname, log_filename);
      return s_instance;
   }

   public synchronized static void destroySingleton()
   {
      if (Assert.DEBUG) Assert.pre(s_instance != null);
      s_instance = null;
   }

   private File m_file;
   private Commitable[] m_commitables;
   private SuperVector m_scratch;
   private DataOutputStream m_logOutput;

   // working value for crc during precommit phase
   private int m_crc;

   private boolean m_test_log = false;

   private static CommitManager s_instance;

   private static final boolean DEBUG  = false;
   private static final int NEW_ENTRY  = 0x41414141;
   private static final int END_OF_LOG = 0x5a5a5a5a;

   public static final String DEFAULT_FILENAME = "commit.dat";
}
