///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility.storage;

import java.io.*;
import java.util.*;
import org.cooldevtools.utility.RWLock;

public class Database
implements Commitable
{
  protected void setMeta(RecordMeta meta,
	                 IndexMeta[] index_list)
    {
      m_recordMeta = meta;
      m_indexMetaList = new IndexMeta[index_list.length];
      m_indexByName = new Hashtable();

      for (int i = 0; i < index_list.length; i++)
	{
	  m_indexByName.put(index_list[i].getName(), new Integer(i));
	  m_indexMetaList[i] = index_list[i];
	}
    }

  protected void setStorage(Storage data_storage,
	           	    Storage blob_storage,
			    StorageIndex[] indexes)
    {
      m_dataStorage = data_storage;
      m_blobStorage = blob_storage;
      m_indexStorages = indexes;
    }

  public synchronized Record getBlank()
    {
      return new Record(m_recordMeta, m_blobStorage, this);
    }

  public synchronized Record read(int rec_id)
    throws IOException, StorageException
    {
      return new Record(m_recordMeta, rec_id, m_dataStorage.get(rec_id), m_blobStorage, this);
    }

  public synchronized void insert(Record r)
    throws IOException, StorageException
    {
      if (r.getId() != Storage.NULL)
	{
	  throw new BadRecordIdException();
	}

      r.writeBlobs();
      byte[] rec = r.getBytes();
      int rec_id = m_dataStorage.put(rec);
      r.setId(rec_id);

      for (int i = 0; i < m_indexStorages.length; i++)
	{
	  String key = m_indexMetaList[i].computeKey(r);
	  m_indexStorages[i].put(key, rec_id);
	}
    }

  public synchronized void remove(Record r)
    throws IOException, StorageException
    {
      if (r.getId() == Storage.NULL)
	{
	  throw new BadRecordIdException();
	}

      r.deleteBlobs();
      int rec_id = r.getId();
      m_dataStorage.remove(rec_id);
      r.setId(Storage.NULL);

      for (int i = 0; i < m_indexStorages.length; i++)
	{
	  String key = m_indexMetaList[i].computeKey(r);
	  m_indexStorages[i].remove(key);
	}
    }

  public synchronized void update(Record r)
    throws IOException, StorageException
    {
      if (r.getId() == Storage.NULL)
	{
	  throw new BadRecordIdException();
	}

      r.writeBlobs();
      m_dataStorage.set(r.getId(), r.getBytes());
    }

  public synchronized boolean exists(String index_name,
				     Record key_record)
    throws IOException, StorageException
    {
      int index_id = ((Integer)m_indexByName.get(index_name)).intValue();
      return exists(index_id, key_record);
    }

  public synchronized boolean exists(int index_id,
				     Record key_record)
    throws IOException, StorageException
    {
      String key = m_indexMetaList[index_id].computeKey(key_record);
      return m_indexStorages[index_id].exists(key);
    }

  public synchronized boolean exists(String index_name,
				     String key)
    throws IOException, StorageException
    {
      int index_id = ((Integer)m_indexByName.get(index_name)).intValue();
      return m_indexStorages[index_id].exists(key);
    }

  public synchronized boolean exists(int index_id,
			             String key)
    throws IOException, StorageException
    {
      return m_indexStorages[index_id].exists(key);
    }

  public synchronized Record find(String index_name,
				  String key)
    throws IOException, StorageException
    {
      return find(((Integer)m_indexByName.get(index_name)).intValue(), key);
    }

  public synchronized Record find(int index_id,
				  String key)
    throws IOException, StorageException
    {
      int rec_id = m_indexStorages[index_id].get(key);
      return read(rec_id);
    }

  public synchronized Record find(String index_name,
				  Record key_record)
    throws IOException, StorageException
    {
      return find(((Integer)m_indexByName.get(index_name)).intValue(), key_record);
    }

  public synchronized Record find(int index_id,
				  Record key_record)
    throws IOException, StorageException
    {
      String key = m_indexMetaList[index_id].computeKey(key_record);
      int rec_id = m_indexStorages[index_id].get(key);
      return read(rec_id);
    }

  private class DBNavigator
    implements DatabaseNavigator
  {
    DBNavigator(IndexNavigator nav)
      throws IOException, StorageException
      {
	m_nav = nav;
	if (m_nav.isValid())
	  {
	    m_record = read(m_nav.getRecordId());
	  }
      }

    public synchronized Database getDatabase()
      {
	return Database.this;
      }

    public synchronized String getKey()
      throws IOException, StorageException
      {
	return m_nav.getKey();
      }

    public synchronized Record getRecord()
      {
	return m_record;
      }

    public synchronized String getMinKey()
      {
	return m_nav.getMinKey();
      }

    public synchronized boolean isValid()
      {
	return m_nav.isValid();
      }

    public synchronized void next()
      throws IOException, StorageException
      {
	m_nav.next();
	if (m_nav.isValid())
	  {
	    m_record = read(m_nav.getRecordId());
	  }
      }

    public synchronized void close()
      {
	m_nav.close();
      }

    private IndexNavigator m_nav;
    private Record m_record;
  }

  public synchronized DatabaseNavigator navigate(String index_name,
						 String min_key_prefix)
    throws IOException, StorageException
    {
      return navigate(((Integer)m_indexByName.get(index_name)).intValue(), min_key_prefix);
    }

  public synchronized DatabaseNavigator navigate(String index_name,
						 Record key_rec)
    throws IOException, StorageException
    {
      return navigate(((Integer)m_indexByName.get(index_name)).intValue(), key_rec);
    }

  public synchronized DatabaseNavigator navigate(int index_id,
						 String min_key_prefix)
    throws IOException, StorageException
    {
      return new DBNavigator(m_indexStorages[index_id].navigate(min_key_prefix));
    }

  public synchronized DatabaseNavigator navigate(int index_id,
						 Record key_record)
    throws IOException, StorageException
    {
      String key = m_indexMetaList[index_id].computeKey(key_record);
      return new DBNavigator(m_indexStorages[index_id].navigate(key));
    }

  public synchronized RWLock getLocker()
    {
      if (m_lock == null)
	{
	  m_lock = new RWLock();
	}
      return m_lock;
    }

  public synchronized void close()
    {
      m_dataStorage.close();
      m_blobStorage.close();
      for (int i = 0; i < m_indexStorages.length; i++)
	{
	  m_indexStorages[i].close();
	}
    }

  public synchronized void sync()
    throws IOException, StorageException
    {
      m_dataStorage.sync();
      m_blobStorage.sync();
      for (int i = 0; i < m_indexStorages.length; i++)
	{
	  m_indexStorages[i].sync();
	}
    }

  public synchronized void abort()
    {
      m_dataStorage.abort();
      m_blobStorage.abort();
      for (int i = 0; i < m_indexStorages.length; i++)
	{
	  m_indexStorages[i].abort();
	}
    }

  public void commitTo(CommitManager cmgr)
  throws IOException
  {
     m_dataStorage.commitTo(cmgr);
     m_blobStorage.commitTo(cmgr);
     for (int i = 0; i < m_indexStorages.length; i++)
       {
         m_indexStorages[i].commitTo(cmgr);
       }
  }

  public void commitWrite()
  throws IOException
  {
     m_dataStorage.commitWrite();
     m_blobStorage.commitWrite();
     for (int i = 0; i < m_indexStorages.length; i++)
       {
         m_indexStorages[i].commitWrite();
       }
  }

  public void commitFailed()
  {
     m_dataStorage.commitFailed();
     m_blobStorage.commitFailed();
     for (int i = 0; i < m_indexStorages.length; i++)
       {
         m_indexStorages[i].commitFailed();
       }
  }

  public void commitSuccessful()
  {
     m_dataStorage.commitSuccessful();
     m_blobStorage.commitSuccessful();
     for (int i = 0; i < m_indexStorages.length; i++)
       {
         m_indexStorages[i].commitSuccessful();
       }
  }
  
  public void commitRecovered()
  {
     m_dataStorage.commitRecovered();
     m_blobStorage.commitRecovered();
     for (int i = 0; i < m_indexStorages.length; i++)
       {
         m_indexStorages[i].commitRecovered();
       }
  }

  private RWLock m_lock;

  private RecordMeta m_recordMeta;

  private IndexMeta[] m_indexMetaList;
  private Hashtable   m_indexByName;

  private Storage        m_dataStorage;
  private Storage        m_blobStorage;
  private StorageIndex[] m_indexStorages;
}
