///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility.storage;

import java.io.*;
import java.util.Enumeration;
import org.cooldevtools.utility.*;

public class FileStorage
  implements Storage
{
   public FileStorage(String filename)
      throws IOException
   {
      File f = new File(filename);
      if (!f.exists()) {
        throw new FileNotFoundException(filename);
      }

      m_header = new byte[HEADER_LENGTH];
      m_file = new BufferedRandomAccessFile(filename);
      readHeader(false);
   }

   public FileStorage(String filename,
		      int record_length)
      throws IOException
   {
      this(filename, record_length, true);
   }

   public FileStorage(String filename,
		      int record_length,
		      boolean require_length)
      throws IOException
   {
      m_logicalLength  = record_length;
      m_physicalLength = record_length + 1;
      if (m_physicalLength < FREE_REC_LEN) {
	 m_physicalLength = FREE_REC_LEN;
      }
      m_file = new BufferedRandomAccessFile(filename);
      init(require_length);
   }

   private void init(boolean require_length)
      throws IOException
   {
      m_header = new byte[HEADER_LENGTH];

      long file_length = m_file.length();
      if (file_length == 0) {
	 Packer.putInt(MAGIC,            m_header, HDR_MAGIC);
	 Packer.putInt(VERSION,          m_header, HDR_VERSION);
	 Packer.putInt(m_physicalLength, m_header, HDR_REC_LEN);
	 Packer.putInt(NONE_FREE,        m_header, HDR_FREE_PTR);

	 m_file.seek(HEADER_START);
	 m_file.write(m_header, 0, m_header.length);

	 m_nextFree   = NONE_FREE;
	 m_nextAppend = 0;
      } else if (file_length < HEADER_LENGTH) {
	 throw new IOException("invalid file size");
      } else {
	 readHeader(require_length);
      }
   }

   private void readHeader(boolean require_length)
     throws IOException
   {
      long file_length = m_file.length();

      m_file.seek(HEADER_START);
      m_file.read(m_header, 0, m_header.length);

      if (Packer.getInt(m_header, HDR_MAGIC) != MAGIC) {
         throw new IOException("invalid file type");
      }

      if (Packer.getInt(m_header, HDR_VERSION) != VERSION) {
         throw new IOException("unsupported file version");
      }

      if (Packer.getInt(m_header, HDR_REC_LEN) != m_physicalLength) {
	 if (require_length) {
	    throw new InvalidRecordLengthException();
	 }
	 m_physicalLength = Packer.getInt(m_header, HDR_REC_LEN);
	 m_logicalLength = m_physicalLength - 1;
      }

      m_nextFree   = Packer.getInt(m_header, HDR_FREE_PTR);
      m_nextAppend = (int)(file_length - HEADER_LENGTH) / m_physicalLength;
   }

  private void selectMark(int rec_id)
    throws IOException
    {
      m_file.seek(HEADER_LENGTH + (rec_id * m_physicalLength));
    }

  private void selectRecord(int rec_id)
    throws IOException
    {
      m_file.seek(HEADER_LENGTH + (rec_id * m_physicalLength) + 1);
    }

  private void addToFreeList(int record_id)
    throws IOException
    {
      byte[] buffer = new byte[FREE_REC_LEN];

      if (m_nextFree == 0)
	{
	  selectMark(0);
	  m_file.read(buffer, 0, buffer.length);

	  selectMark(record_id);
	  m_file.write(buffer, 0, buffer.length);

	  Packer.putInt(record_id, buffer, 1);
	  selectMark(0);
	  m_file.write(buffer, 0, buffer.length);
	}
      else
	{
	  Packer.putByte(AVAILABLE, buffer, 0);
	  Packer.putInt(m_nextFree, buffer, 1);

	  selectMark(record_id);
	  m_file.write(buffer, 0, buffer.length);
	  m_nextFree = record_id;

	  Packer.putInt(m_nextFree, m_header, HDR_FREE_PTR);
	  m_file.seek(HEADER_START);
	  m_file.write(m_header, 0, m_header.length);
	}
    }

  private int getFromFreeList()
    throws IOException
    {
      if (m_nextFree == NONE_FREE)
	{
	  return extend();
	}

      int record_id = m_nextFree;

      byte[] buffer = new byte[FREE_REC_LEN];

      selectMark(record_id);
      m_file.read(buffer, 0, buffer.length);
      m_nextFree = Packer.getInt(buffer, 1);

      Packer.putByte(RESERVED, buffer, 0);

      selectMark(record_id);
      m_file.write(buffer, 0, buffer.length);

      Packer.putInt(m_nextFree, m_header, HDR_FREE_PTR);
      m_file.seek(HEADER_START);
      m_file.write(m_header, 0, m_header.length);

      return record_id;
    }

  public synchronized int getRecordLength()
    {
      return m_logicalLength;
    }

   public synchronized void close()
   {
      try {
	 m_file.close();
      } catch (IOException ex) {
      }
      m_file = null;
      m_header = null;
   }

  public synchronized void sync()
    throws IOException, StorageException
    {
      m_file.flush();
    }

  public synchronized void abort()
    {
      m_file.abort();
      try {
         readHeader(true);
      } catch (IOException ex) {
         ex.printStackTrace();
         throw new RuntimeException(ex.getMessage());
      }
    }

  public synchronized void commitTo(CommitManager cmgr)
  throws IOException
  {
     m_file.commitTo(cmgr);
  }

  public synchronized void commitWrite()
  throws IOException
  {
     m_file.commitWrite();
  }

  public synchronized void commitFailed()
  {
     m_file.commitFailed();
     try {
        readHeader(true);
     } catch (IOException ex) {
        ex.printStackTrace();
        throw new RuntimeException(ex.getMessage());
     }
  }

  public synchronized void commitSuccessful()
  {
     m_file.commitSuccessful();
  }
  
  public synchronized void commitRecovered()
  {
     m_file.commitRecovered();
     try {
        readHeader(true);
     } catch (IOException ex) {
        ex.printStackTrace();
        throw new RuntimeException(ex.getMessage());
     }
  }

  private int extend()
    throws IOException
    {
      byte[] extend = new byte[m_physicalLength];
      extend[0] = RESERVED;

      selectMark(m_nextAppend);
      m_file.write(extend, 0, extend.length);

      return m_nextAppend++;
    }

  public synchronized int reserve()
    throws IOException, StorageException, OutOfSpaceException
    {
      int record_id = getFromFreeList();
      return record_id;
    }

  private void validateRecordLength(byte[] record)
    throws InvalidRecordLengthException
    {
      if (record.length != m_logicalLength)
	{
	  throw new InvalidRecordLengthException();
	}
    }

  private void validateRecordId(int record_id)
    throws IOException, BadRecordIdException
    {
      byte[] buffer = new byte[1];
      selectMark(record_id);
      int nread = m_file.read(buffer, 0, buffer.length);
      if ((nread != 1) || (buffer[0] != RESERVED))
	{
	  throw new BadRecordIdException();
	}
    }

  public synchronized void release(int record_id)
    throws IOException, StorageException, BadRecordIdException
    {
      validateRecordId(record_id);
      addToFreeList(record_id);
    }

  public synchronized int put(byte[] record)
    throws IOException, StorageException, OutOfSpaceException
    {
      int record_id = reserve();
      set(record_id, record);
      return record_id;
    }

  public synchronized void set(int record_id, byte[] record)
    throws IOException, StorageException, BadRecordIdException
    {
      validateRecordLength(record);
      validateRecordId(record_id);

      selectRecord(record_id);
      m_file.write(record, 0, record.length);
    }

  public synchronized byte[] get(int record_id)
    throws IOException, StorageException, BadRecordIdException
    {
      validateRecordId(record_id);

      byte[] record = new byte[m_logicalLength];

      selectRecord(record_id);
      int nread = m_file.read(record, 0, record.length);
      if (nread != record.length)
	{
          System.out.println("eof expected " + record.length
                             + " bytes but only read " + nread);
	  throw new EOFException();
	}
      return record;
    }

  public synchronized void get(int record_id, byte[] record)
    throws IOException, StorageException, BadRecordIdException
    {
      validateRecordLength(record);
      validateRecordId(record_id);

      selectRecord(record_id);
      int nread = m_file.read(record, 0, record.length);
      if (nread != record.length)
	{
	  throw new EOFException();
	}
    }

  public synchronized byte[] remove(int record_id)
    throws IOException, StorageException, BadRecordIdException
    {
      byte[] answer = get(record_id);
      release(record_id);
      return answer;
    }

  public synchronized void remove(int record_id, byte[] record)
    throws IOException, StorageException, BadRecordIdException
    {
      get(record_id, record);
      release(record_id);
    }

  public synchronized int nextRecord(int start_record)
    throws IOException, StorageException
  {
    int record_id = 0;
    if (start_record != NULL) {
      validateRecordId(start_record);
      record_id = start_record + 1;
    }

    byte[] buffer = new byte[1];
    while (true) {
      selectMark(record_id);
      int nread = m_file.read(buffer, 0, buffer.length);
      if (nread != 1) {
        return NULL;
      }

      if (buffer[0] == RESERVED) {
        return record_id;
      }

      ++record_id;
    }
  }

  private static void simple()
    throws Exception
  {
    PrintStream out = System.out;
    BufferedReader in = new BufferedReader(new InputStreamReader(System.in));

    final int LOOPS = 10;
    int number = 0x100;

    Storage s = new FileStorage("_test.dat", 3);

    out.println("writing records");
    for (int i = 0; i < 2 * LOOPS; i++) {
      byte[] orig = Packer.stringToBytes(Integer.toHexString(number++));
      int rec = s.put(orig);
      byte[] echo = s.get(rec);
      if (echo.length != orig.length) {
        System.out.println("echo length " + echo.length
                           + " expecting " + orig.length);
      }
      for (int k = 0; k < echo.length; k++) {
        if (echo[k] != orig[k]) {
          System.out.println("error at index " + k
                             + ": " + echo[k]
                             + " != " + orig[k]);
        }
      }
    }

    s.sync();
    out.println("Press ENTER to continue...");
    in.readLine();

    out.println("releasing records");
    for (int i = 0; i < 2 * LOOPS; i += 2) {
      s.release(i);
    }

    s.sync();
    out.println("Press ENTER to continue...");
    in.readLine();

    out.println("writing records");
    for (int i = 0; i < 2 * LOOPS; i++) {
      s.put(Packer.stringToBytes(Integer.toHexString(number++)));
    }
    s.close();
  }

  private void verify()
    throws IOException
  {
    SuperHashtable free_table = new SuperHashtable();
    byte[] buffer = new byte[m_physicalLength];
    for (int i = 0; i < m_file.length() / m_physicalLength; ++i) {
      selectMark(i);
      int nread = m_file.read(buffer, 0, buffer.length);
      if (nread != m_physicalLength) {
        System.out.println("Record id: " + i + ": cannot be read");
      } else if (buffer[0] == AVAILABLE) {
        free_table.put(new Integer(i), Boolean.TRUE);
      } else if (buffer[0] == RESERVED) {
        // ignore used records
      } else {
        System.out.println("Record id: " + i + ": bad status code: " + buffer[0]);
      }
    }

    int record_id = m_nextFree;
    while (record_id != NONE_FREE) {
      selectMark(record_id);
      int nread = m_file.read(buffer, 0, buffer.length);
      if (nread != m_physicalLength) {
        System.out.println("Record id: " + record_id + ": cannot be read");
        break;
      } else if (buffer[0] != AVAILABLE) {
        System.out.println("Record id: " + record_id + ": expected AVAILABLE but found: " + buffer[0]);
        break;
      }
      free_table.remove(new Integer(record_id));
      record_id = Packer.getInt(buffer, 1);
    }

    if (free_table.size() != 0) {
      System.out.println("THE FOLLOWING RECORDS WERE MISSING FROM FREE LIST:");
      Enumeration en = free_table.keys();
      while (en.hasMoreElements()) {
        record_id = ((Integer)en.nextElement()).intValue();
        System.out.println("  " + record_id);
      }
    }
  }

  private static void verifyFile(String filename)
    throws Exception
  {
    FileStorage file = new FileStorage(filename);
    file.verify();
  }

  public static void main(String args[])
    throws Exception
  {
    if (args[0].equals("simple")) {
      simple();
      return;
    }

    if (args[0].equals("verify")) {
      verifyFile(args[1]);
      return;
    }

    System.out.println("usage: <class> simple|verify filename");
  }

  private BufferedRandomAccessFile m_file;
  private int    m_logicalLength;
  private int    m_physicalLength;
  private int    m_nextAppend;
  private int    m_nextFree;
  private byte[] m_header;

  private static final byte AVAILABLE = (byte)0x00;
  private static final byte RESERVED  = (byte)0xee;

  private static final int HEADER_START  = 0;
  private static final int HEADER_LENGTH = 16;
  private static final int HDR_MAGIC     = 0;
  private static final int HDR_VERSION   = 4;
  private static final int HDR_REC_LEN   = 8;
  private static final int HDR_FREE_PTR  = 12;

  private static final int   MAGIC   = 0xda1aba5e;
  private static final short VERSION = 0x0001;

  private static final int POS_FREE_LIST = 0;
  private static final int FREE_REC_LEN  = 5;
  private static final int NONE_FREE     = -1;
}
