///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility.storage;

import java.io.*;
import java.util.Vector;
import org.cooldevtools.hdl.*;

public class HdlSchema
implements Schema
{
   public HdlSchema()
   {
   }

   public HdlSchema(String path,
                    String filename)
   throws IOException, HDLException
   {
      load(path, filename);
   }

   public HdlSchema(BufferedReader in)
   throws IOException, HDLException
   {
      load(in);
   }

   public HdlSchema(Statement root)
   throws IOException, HDLException
   {
      m_root = root;
      SyntaxChecker check = new SyntaxChecker(SYNTAX);
      check.checkSyntax(m_root);
   }

   public synchronized void load(String path,
                                 String filename)
   throws IOException, HDLException
   {
      File f = new File(path, filename);
      BufferedReader in = new BufferedReader(new FileReader(f));
      try {
         load(in);
      } finally {
         in.close();
      }
   }

   public synchronized void load(BufferedReader in)
   throws IOException, HDLException
   {
      Parser p = new Parser();
      m_root = p.parse(in);
      SyntaxChecker check = new SyntaxChecker(SYNTAX);
      check.checkSyntax(m_root);
   }

   private Statement findDB(String dbname)
   throws StorageException, IOException
   {
      int num_children = m_root.numChildren();
      for (int i = 0; i < num_children; i++) {
         Statement child = m_root.child(i);
         if (child.name().strValue().equals("database") && child.argument(0).strValue().equals(dbname)) {
            return child;
         }
      }
      throw new UnknownKeyException(dbname);
   }

   public synchronized RecordMeta getRecordMetaFor(String dbname)
   throws StorageException, IOException
   {
      Statement db = findDB(dbname);
      int num_children = db.numChildren();
      Vector vec = new Vector(num_children);
      for (int i = 0; i < num_children; i++) {
         Statement child = db.child(i);
         if (child.name().strValue().equals("field")) {
            String name = child.argument(0).strValue();
            int type    = FIELD_TYPES[child.argument(1).intValue()];
            int len     = (child.numArguments() > 2) ? child.argument(2).intValue() : -1;
            vec.addElement(new FieldMeta(name, len, type));
         }
      }

      return new RecordMeta(FieldMeta.toArray(vec));
   }

   public synchronized IndexMeta[] getIndexMetaFor(String dbname,
                                                   RecordMeta rec)
   throws StorageException, IOException
   {
      Statement db = findDB(dbname);
      int num_children = db.numChildren();
      Vector vec = new Vector(num_children);
      for (int i = 0; i < num_children; i++) {
         Statement child = db.child(i);
         if (child.name().strValue().equals("index")) {
            String name = child.argument(0).strValue();
            int child_nc = child.numChildren();
            String[] columns = new String[child_nc];
            for (int k = 0; k < child_nc; k++) {
               columns[k] = child.child(k).argument(0).strValue();
            }
            vec.addElement(new IndexMeta(name, rec, columns));
         }
      }

      return IndexMeta.toArray(vec);
   }

   private Statement m_root;

   private static final int[] FIELD_TYPES = {
      FieldMeta.TYPE_STRING,
      FieldMeta.TYPE_USTRING,
      FieldMeta.TYPE_SHORT,
      FieldMeta.TYPE_INT,
      FieldMeta.TYPE_LONG,
      FieldMeta.TYPE_BLOB,
   };
   private static final String SYNTAX = 
      "begin language schema;\n" +
      "  begin block database;\n" +
      "    args 1 1 string;\n" +
      "    begin statement field;\n" +
      "      args 2 3 string table int;\n" +
      "      table id string\n" +
      "            id ustring\n" +
      "            id short\n" +
      "            id int\n" +
      "            id long\n" +
      "            id blob;\n" +
      "    end;\n" +
      "    begin block index;\n" +
      "      args 2 2 string id btree;\n" +
      "      begin statement field;\n" +
      "        args 1 1 string;\n" +
      "      end;\n" +
      "    end;\n" +
      "  end;\n" +
      "end;\n";
}

