///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility.storage;

import java.io.*;
import java.util.Random;

public class Packer
{
   public static byte[] stringToBytes(String s)
   throws IOException
   {
      int length = s.length();
      char[] chars = s.toCharArray();
      byte[] bytes = new byte[length];
      for (int i = 0; i < length; i++) {
         bytes[i] = (byte)chars[i];
      }
      return bytes;
    }

   public static String stringFromBytes(byte[] buffer,
                                        int offset)
   throws IOException
   {
      int length = buffer.length - offset;
      char[] chars = new char[length];
      for (int i = 0; i < length; i++) {
         chars[i] = (char)(buffer[i] & 0xff);
      }
      return new String(chars);
   }

   /**
      Encode the string into the buffer.
      The string will occupy 4 + 2 * length bytes in the buffer.
      The extra 4 bytes are used to record the string's length.
    */
   public static void putUString(String s,
                                 int length,
                                 byte[] buffer,
                                 int offset)
   {
      int strlen = s.length();
      if (strlen < length) {
         length = strlen;
      }

      putInt(length, buffer, offset);
      offset += 4;

      char[] tmp = new char[length];
      s.getChars(0, length, tmp, 0);

      for (int i = 0; i < length; i++) {
         putShort((short)tmp[i], buffer, offset);
         offset += 2;
      }
   }

   /**
      Encode the string into the buffer.
      The string will occupy 4 + length bytes in the buffer.
      The extra 4 bytes are used to record the string's length.
      Only the low order 8 bits are converted, so never use this for
      unicode data!
    */
   public static void putString(String s,
                                int length,
                                byte[] buffer,
                                int offset)
   {
      int strlen = s.length();
      if (strlen < length) {
         length = strlen;
      }

      putInt(length, buffer, offset);
      offset += 4;

      char[] tmp = s.toCharArray();
      for (int i = 0; i < length; i++) {
         buffer[offset + i] = (byte)tmp[i];
      }
   }

   public static String getString(byte[] buffer,
                                  int offset)
   {
      int strlen = getInt(buffer, offset);
      offset += 4;

      char[] tmp = new char[strlen];
      for (int i = 0; i < strlen; i++) {
         tmp[i] = (char)(((short)buffer[offset + i]) & 0x00ff);
      }
      return new String(tmp);
   }

   public static String getUString(byte[] buffer,
                                   int offset)
   {
      int strlen = getInt(buffer, offset);
      offset += 4;

      char[] tmp = new char[strlen];
      for (int i = 0; i < strlen; i++) {
         tmp[i] = (char)getShort(buffer, offset);
         offset += 2;
      }
      return new String(tmp);
   }

   public static void putInt(int value,
                             byte[] buffer,
                             int offset)
   {
      buffer[offset    ] = (byte)(value >>> 24);
      buffer[offset + 1] = (byte)((value >>> 16) & 0x000000ff);
      buffer[offset + 2] = (byte)((value >>>  8) & 0x000000ff);
      buffer[offset + 3] = (byte)(value & 0x000000ff);
   }

   public static void putLong(long value,
                              byte[] buffer,
                              int offset)
   {
      putInt((int)(value >>> 32), buffer, offset);
      putInt((int)(value & 0x00000000ffffffff), buffer, offset + 4);
   }

   public static void putByte(byte value,
                              byte[] buffer,
                              int offset)
   {
      buffer[offset] = value;
   }

   public static int getInt(byte[] buffer,
                            int offset)
   {
      int value
      = (buffer[offset] << 24)
        | ((buffer[offset + 1] << 16) & 0x00ff0000)
        | ((buffer[offset + 2] <<  8) & 0x0000ff00)
        | (buffer[offset + 3]         & 0x000000ff);
      return value;
   }

   public static long getLong(byte[] buffer,
                              int offset)
   {
      long value1 = getInt(buffer, offset);
      long value2 = getInt(buffer, offset + 4);
      return (value1 << 32) | (value2 & 0x00000000ffffffff);
   }

   public static void putShort(short value,
                               byte[] buffer,
                               int offset)
   {
      buffer[offset    ] = (byte)(value >>> 8);
      buffer[offset + 1] = (byte)(value & 0xff);
   }

   public static short getShort(byte[] buffer,
                                int offset)
   {
      short value = (short)((buffer[offset] << 8)
                            | (buffer[offset + 1] & 0x000000ff));
      return value;
   }

   private static void verifyInt(int value)
   {
      byte[] buffer = new byte[4];
      putInt(value, buffer, 0);
      if (value != getInt(buffer, 0)) {
         System.out.println("expected " + Integer.toHexString(value)
                            + " found " + Integer.toHexString(getInt(buffer, 0)));
      }
   }

   private static void verifyLong(int int_value)
   {
      byte[] buffer = new byte[8];
      long value = (int_value << 32) | int_value;
      putLong(value, buffer, 0);
      if (value != getLong(buffer, 0)) {
         System.out.println("expected " + Long.toHexString(value)
                            + " found " + Long.toHexString(getLong(buffer, 0)));
      }
   }

   private static void verifyShort(short value)
   {
      byte[] buffer = new byte[4];
      putShort(value, buffer, 0);
      if (value != getShort(buffer, 0)) {
         System.out.println("expected " + Integer.toHexString(value)
                            + " found " + Integer.toHexString(getShort(buffer, 0)));
      }
   }

   private static void verifyString(String value)
   {
      byte[] buffer = new byte[1024];
      putString(value, 8, buffer, 0);
      if (!value.equals(getString(buffer, 0))) {
         System.out.println("expected '" + value
                            + "' found '" + getString(buffer, 0)
                            + "'");
      }
   }

   private static void verifyUString(String value)
   {
      byte[] buffer = new byte[1024];
      putUString(value, 8, buffer, 0);
      if (!value.equals(getUString(buffer, 0))) {
         System.out.println("expected '" + value
                            + "' found '" + getUString(buffer, 0)
                            + "'");
      }
   }

   public static void main(String[] args)
   {
      final int LOOPS = 1000;
      Random r = new Random(1000);
      for (int i = 0; i < LOOPS; i++) {
         verifyLong(r.nextInt());
      }
      for (int i = 0; i < LOOPS; i++) {
         verifyInt(r.nextInt());
      }
      for (int i = 0; i < LOOPS; i++) {
         verifyShort((short)r.nextInt());
      }
      verifyString("");
      verifyString("Short");
      verifyString("This is a string");
      verifyUString("");
      verifyUString("Short");
      verifyUString("This is a string");
   }
}
