///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility.storage;

import java.util.Vector;
import java.io.IOException;
import org.cooldevtools.utility.Assert;

public class Record {
   Record(RecordMeta meta,
          Storage blobs,
          Database owner)
   {
      this(meta, Storage.NULL, new byte[meta.getLength()], blobs, owner);
   }

   Record(RecordMeta meta,
          int record_id,
          byte[] buffer,
          Storage blobs,
          Database owner)
   {
      m_meta = meta;
      m_id   = record_id;
      m_blobs = blobs;
      setBytes(buffer);
      m_blobMods = new byte[m_meta.getNumFields()][];
      m_owner = owner;
   }

   public Database getDB()
   {
      return m_owner;
   }

   public void reset()
   {
      m_id = Storage.NULL;
      setBytes(new byte[m_meta.getLength()]);
      m_blobMods = new byte[m_meta.getNumFields()][];
   }

   public byte[] getBytes()
   {
      return m_buffer;
   }

   public void setBytes(byte[] bytes)
   {
      m_buffer = bytes;
   }

   void setId(int id)
   {
      m_id = id;
   }

   int getId()
   {
      return m_id;
   }

   public RecordMeta getMeta()
   {
      return m_meta;
   }

   public String getString(String name)
   {
      return m_meta.getStringFromBuffer(m_meta.getFieldIndex(name), m_buffer);
   }

   public short getShort(String name)
   {
      return m_meta.getShortFromBuffer(m_meta.getFieldIndex(name), m_buffer);
   }

   public int getInt(String name)
   {
      return m_meta.getIntFromBuffer(m_meta.getFieldIndex(name), m_buffer);
   }

   public long getLong(String name)
   {
      return m_meta.getLongFromBuffer(m_meta.getFieldIndex(name), m_buffer);
   }

   public String getString(int index)
   {
      return m_meta.getStringFromBuffer(index, m_buffer);
   }

   public short getShort(int index)
   {
      return m_meta.getShortFromBuffer(index, m_buffer);
   }

   public int getInt(int index)
   {
      return m_meta.getIntFromBuffer(index, m_buffer);
   }

   public long getLong(int index)
   {
      return m_meta.getLongFromBuffer(index, m_buffer);
   }

   public void setString(String name,
                         String value)
   {
      setString(m_meta.getFieldIndex(name), value);
   }

   public void setShort(String name,
                        short value)
   {
      setShort(m_meta.getFieldIndex(name), value);
   }

   public void setInt(String name,
                      int value)
   {
      setInt(m_meta.getFieldIndex(name), value);
   }

   public void setLong(String name,
                       long value)
   {
      setLong(m_meta.getFieldIndex(name), value);
   }

   public void setString(int index,
                         String value)
   {
      if (Assert.DEBUG) Assert.pre(m_meta.getField(index).getType() == FieldMeta.TYPE_STRING);
      if ((m_id != Storage.NULL) && m_meta.isIndexed(index)) {
         throw new RuntimeException("cannot modify indexed field");
      }
      m_meta.setStringInBuffer(value, index, m_buffer);
   }

   public void setShort(int index,
                        short value)
   {
      if (Assert.DEBUG) Assert.pre(m_meta.getField(index).getType() == FieldMeta.TYPE_SHORT);
      if ((m_id != Storage.NULL) && m_meta.isIndexed(index)) {
         throw new RuntimeException("cannot modify indexed field");
      }
      m_meta.setShortInBuffer(value, index, m_buffer);
   }

   public void setInt(int index,
                      int value)
   {
      if (Assert.DEBUG) Assert.pre(m_meta.getField(index).getType() == FieldMeta.TYPE_INT);
      if ((m_id != Storage.NULL) && m_meta.isIndexed(index)) {
         throw new RuntimeException("cannot modify indexed field");
      }
      m_meta.setIntInBuffer(value, index, m_buffer);
   }

   public void setLong(int index,
                       long value)
   {
      if (Assert.DEBUG) Assert.pre(m_meta.getField(index).getType() == FieldMeta.TYPE_LONG);
      if ((m_id != Storage.NULL) && m_meta.isIndexed(index)) {
         throw new RuntimeException("cannot modify indexed field");
      }
      m_meta.setLongInBuffer(value, index, m_buffer);
   }

   public byte[] getBlob(String name)
   throws IOException, StorageException
   {
      return getBlob(m_meta.getFieldIndex(name));
   }

   public byte[] getBlob(int index)
   throws IOException, StorageException
   {
      if (Assert.DEBUG) Assert.pre(m_meta.getField(index).getType() == FieldMeta.TYPE_BLOB);
      if (m_blobMods[index] != null) {
         return m_blobMods[index];
      }

      return m_meta.getBlobFromBuffer(index, m_buffer, m_blobs);
   }

   public String getBlobAsString(int index)
   throws IOException, StorageException
   {
      return Packer.stringFromBytes(getBlob(index), 0);
   }

   public String getBlobAsString(String name)
   throws IOException, StorageException
   {
      return Packer.stringFromBytes(getBlob(name), 0);
   }

   public void setBlob(String name,
                       byte[] value)
   {
      setBlob(m_meta.getFieldIndex(name), value);
   }

   public void setBlob(int index,
                       byte[] value)
   {
      if (Assert.DEBUG) Assert.pre(m_meta.getField(index).getType() == FieldMeta.TYPE_BLOB);
      m_blobMods[index] = value;
   }

   public void setBlob(String name,
                       String value)
   throws IOException
   {
      setBlob(name, Packer.stringToBytes(value));
   }

   public void setBlob(int index,
                       String value)
   throws IOException
   {
      setBlob(index, Packer.stringToBytes(value));
   }

   void deleteBlobs()
   throws IOException, StorageException
   {
      m_meta.deleteBlobsForBuffer(m_buffer, m_blobs);
   }

   void writeBlobs()
   throws IOException, StorageException
   {
      for (int i = 0; i < m_blobMods.length; i++) {
         if (m_blobMods[i] != null) {
            m_meta.setBlobInBuffer(m_blobMods[i], i, m_buffer, m_blobs);
         }
      }
   }

   public static Record[] toArray(Vector v)
   {
      synchronized (v) {
         Record[] answer = new Record[v.size()];
         v.copyInto(answer);
         return answer;
      }
   }

   private int        m_id;
   private RecordMeta m_meta;
   private byte[]     m_buffer;
   private Storage    m_blobs;
   private byte[][]   m_blobMods;
   private Database   m_owner;
}
