///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility.storage;

import java.io.*;

public class SimpleDatabase
extends Database
{
   public SimpleDatabase(RecordMeta meta,
                         IndexMeta[] index_list,
                         String path,
                         String dbname)
   throws IOException, StorageException
   {
      initializeDatabase(meta, index_list, path, dbname);
   }

   public SimpleDatabase(Schema schema,
                         String path,
                         String dbname)
   throws IOException, StorageException
   {
      RecordMeta rm   = schema.getRecordMetaFor(dbname);
      IndexMeta[] ims = schema.getIndexMetaFor(dbname, rm);
      initializeDatabase(rm, ims, path, dbname);
   }

   private void initializeDatabase(RecordMeta meta,
                                   IndexMeta[] index_list,
                                   String path,
                                   String dbname)
   throws IOException, StorageException
   {
      if (!path.endsWith(File.separator)) {
         path = path + File.separator;
      }

      final int BLOB_SIZE   = 1024;
      final int BTREE_ORDER = 10;
      m_datafile = new FileStorage(path + dbname + ".data", meta.getLength());
      m_blobfile = new VariableStorage(path + dbname + "_blobs.data", BLOB_SIZE);
      m_indexes  = new BTreeIndex[index_list.length];
      for (int i = 0; i < m_indexes.length; i++) {
         String filename = path + dbname + "_" + index_list[i].getName() + ".index";
         int rec_len = BTreeIndex.calcRecordLength(index_list[i].getKeyLength(), BTREE_ORDER);
         FileStorage store = new FileStorage(filename, rec_len);
         m_indexes[i] = new BTreeIndex(store, index_list[i].getKeyLength(), BTREE_ORDER);
      }
      
      setMeta(meta, index_list);
      setStorage(m_datafile, m_blobfile, m_indexes);
   }

   private static void walk(String index_name,
                            DatabaseNavigator nav)
   throws Exception
   {
      PrintStream out = System.out;
      out.println("Index: " + index_name);

      for (int i = 1; nav.isValid(); i++) {
         out.println("record #" + i + ": key = '" + nav.getKey() + "'");
         Record r = nav.getRecord();
         out.println("  GroupName : " + r.getString("GroupName"));
         out.println("  NextNumber: " + r.getString("NextNumber"));
         out.println("        Body: " + new String(r.getBlob("Body")));
         nav.next();
      }
   }

   public static void main(String[] args)
   throws Exception
   {
      FieldMeta[] fms = {
         new FieldMeta("GroupName", 48, FieldMeta.TYPE_STRING),
         new FieldMeta("NextNumber", FieldMeta.TYPE_SHORT),
         new FieldMeta("Body", FieldMeta.TYPE_BLOB)
      };
      RecordMeta rm = new RecordMeta(fms);
      String[] i1_fields = { "GroupName"};
      String[] i2_fields = { "NextNumber"};
      String[] i3_fields = { "GroupName", "NextNumber"};
      IndexMeta[] ims = {
         new IndexMeta("GroupName", rm, i1_fields),
         new IndexMeta("NextNumber", rm, i2_fields),
         new IndexMeta("Both", rm, i3_fields)
      };
      Database db = new SimpleDatabase(rm, ims, ".", "_test");
      byte[] blob = Packer.stringToBytes("Hello");
      byte[] blob2 = Packer.stringToBytes("Olleh");
      Record r = db.getBlank();
      r.reset();
      r.setString(0, "clnews.development");
      r.setInt(1, 18);
      r.setBlob(2, blob);
      db.insert(r);
      r.reset();
      r.setString(0, "clnews.announce");
      r.setInt(1, 4);
      r.setBlob(2, blob);
      db.insert(r);
      r.reset();
      r.setString(0, "clnews.support");
      r.setInt(1, 176);
      r.setBlob(2, blob);
      db.insert(r);
      r.reset();
      r.setString(0, "clnews.discuss");
      r.setInt(1, 43);
      r.setBlob(2, blob);
      db.insert(r);
      r.setBlob(2, blob2);
      db.update(r);
      for (int i = 0; i < ims.length; i++) {
         walk(ims[i].getName(), db.navigate(i, ""));
      }
      db.close();
   }

   private FileStorage       m_datafile;
   private VariableStorage   m_blobfile;
   private BTreeIndex[] m_indexes;
}
