///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility.storage;

import java.io.IOException;

public interface Storage
extends Commitable
{
  /**
     Returns the record length of the data store.
   */
  public int getRecordLength();

  /**
     Closes the storage object.
     Once called, the storage object cannot be used again.
     Implicitly calls sync().
   */
  public void close();

  /**
     Synchronize the in core image with the permanent store.
     Used for data stores that defer writes for higher performance.
   */
  public void sync()
    throws IOException, StorageException;

  /**
     Throw out any unsaved changes and return to the state
     at the last call to sync().  No exceptions are allowed
     since this is generally used as the last resort when
     handling other exceptions.
   */
  public void abort();

  /**
     Reserve a record and return it's record id.
     Doesn't actually store anything in the record.
   */
  public int reserve()
    throws IOException, StorageException, OutOfSpaceException;

  /**
     Returns the next reserved record id after the specified one.
     If record_id == NULL, the first in-use record id in the file
     is returned.  If none is found, NULL is returned.
   */
  public int nextRecord(int start_record)
    throws IOException, StorageException;

  /**
     Release (unreserve) a record.
   */
  public void release(int record_id)
    throws IOException, StorageException, BadRecordIdException;

  /**
     Reserve a record, store the data into it, and return the record id.
     Equiv. to calling reserve() followed by set().
   */
  public int put(byte[] record)
    throws IOException, StorageException, OutOfSpaceException;

  /**
     Put the data into a previously reserved record.
   */
  public void set(int record_id, byte[] record)
    throws IOException, StorageException, BadRecordIdException;

  /**
     Reads the data from a previously reserved record and returns
     it as a byte array.
   */
  public byte[] get(int record_id)
    throws IOException, StorageException, BadRecordIdException;

  /**
     Reads the data from a previously reserved record into the specified
     byte array.
   */
  public void get(int record_id, byte[] record)
    throws IOException, StorageException, BadRecordIdException;

  /**
     Reads the data from a record and releases the record.
     Equiv. to calling get() followed by release().
   */
  public byte[] remove(int record_id)
    throws IOException, StorageException, BadRecordIdException;

  /**
     Reads the data from a record and releases the record.
     Equiv. to calling get() followed by release().
   */
  public void remove(int record_id, byte[] record)
    throws IOException, StorageException, BadRecordIdException;

  /**
     Value returned by getRecordLength() for variable length
     storage objects.
   */
  public static final int VARIABLE = -1;

  public static final int NULL = -1;
}
