///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// http://www.cooldevtools.com
// $Id$
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//    http://www.cooldevtools.com/qpl.html
//

package org.cooldevtools.utility.storage;

import java.io.IOException;
import java.util.Enumeration;
import org.cooldevtools.utility.*;

public class VariableStorage implements Storage
{
   public VariableStorage(Storage store)
   {
      m_store  = store;
      m_length = m_store.getRecordLength() - 8;
      m_sorter = new Sorter(new SortNodes());
   }

   public VariableStorage(String filename)
      throws IOException, StorageException
   {
      this(new FileStorage(filename));
   }

   public VariableStorage(String filename,
			  int record_length)
      throws IOException, StorageException
   {
      this(new FileStorage(filename, record_length, false));
   }

   private class Node
   {
      public Node()
      {
	 this(Storage.NULL, 0);
      }

      public Node(int _record_id, int _length)
      {
	 next      = Storage.NULL;
	 data      = new byte[m_length];
	 length    = _length;
	 record_id = _record_id;
      }

      public String toString()
      {
	 return "Node(" + record_id + "," + next + "," + length + "," + Misc.prettyBytes(data, length, 50) + ")";
      }

      public int    record_id;
      public int    length;
      public int    next;
      public byte[] data;
   }

   private class SortNodes implements BinaryComparator
   {
      public int compare(Object a, Object b)
      {
	 Node na = (Node)a;
	 Node nb = (Node)b;
	 return na.record_id - nb.record_id;
      }
   }

   private Node read(int record_id)
      throws IOException, StorageException
   {
      Node n = new Node();
      byte[] rec = m_store.get(record_id);
      n.record_id = record_id;
      n.length    = Packer.getInt(rec, 0);
      n.next      = Packer.getInt(rec, 4);
      n.data      = new byte[n.length];
      System.arraycopy(rec, 8, n.data, 0, n.length);
      return n;
   }

   private void write(Node n)
      throws IOException, StorageException
   {
      byte[] rec = new byte[m_store.getRecordLength()];
      Packer.putInt(n.length, rec, 0);
      Packer.putInt(n.next, rec, 4);
      System.arraycopy(n.data, 0, rec, 8, n.length);
      m_store.set(n.record_id, rec);
   }

   public synchronized int getRecordLength()
   {
      return Storage.VARIABLE;
   }

   public synchronized void close()
   {
      m_store.close();
      m_store = null;
   }

   public synchronized void sync()
      throws IOException, StorageException
   {
      m_store.sync();
   }

   public synchronized void abort()
   {
      m_store.abort();
   }

   public synchronized void commitTo(CommitManager cmgr)
      throws IOException
   {
      m_store.commitTo(cmgr);
   }

   public synchronized void commitWrite()
      throws IOException
   {
      m_store.commitWrite();
   }

   public synchronized void commitFailed()
   {
      m_store.commitFailed();
   }

   public synchronized void commitSuccessful()
   {
      m_store.commitSuccessful();
   }

  public synchronized int nextRecord(int start_record)
    throws IOException, StorageException
  {
    throw new StorageException("not implemented");
  }

   public synchronized void commitRecovered()
   {
      m_store.commitRecovered();
   }
  
   public synchronized int reserve()
      throws IOException, StorageException, OutOfSpaceException
   {
      return put(new byte[0]);
   }

   public synchronized void release(int record_id)
      throws IOException, StorageException, BadRecordIdException
   {
      while (record_id != Storage.NULL) {
	 Node n = read(record_id);
	 m_store.release(record_id);
	 record_id = n.next;
      }
   }

   private Node[] reserveNodes(int length)
      throws IOException, StorageException
   {
      if (length == 0) {
	 return new Node[] { new Node(m_store.reserve(), length) };
      }

      int extra = length % m_length;
      int num_nodes = (length / m_length) + (extra > 0 ? 1 : 0);
      Node[] nodes = new Node[num_nodes];

      int last_index = num_nodes - 1;
      for (int i = 0; i <= last_index; i++) {
	 nodes[i] = new Node(m_store.reserve(), m_length);
      }

      m_sorter.sort(nodes);

      for (int i = 0; i < last_index; i++) {
	 nodes[i].next   = nodes[i+1].record_id;
      }
      nodes[last_index].next   = Storage.NULL;
      if (extra > 0) {
	 nodes[last_index].length = extra;
      }

      return nodes;
   }

   public synchronized int put(byte[] record)
      throws IOException, StorageException, OutOfSpaceException
   {
      Node[] nodes = reserveNodes(record.length);
      int offset = 0;

      for (int i = 0; i < nodes.length; i++) {
	 Node n = nodes[i];

	 System.arraycopy(record, offset, n.data, 0, n.length);
	 write(n);

	 offset += n.length;
      }

      if (Assert.DEBUG) Assert.post(offset == record.length);
      if (Assert.DEBUG) Assert.post(nodes[nodes.length - 1].next == Storage.NULL);

      return nodes[0].record_id;
   }

   public synchronized void set(int record_id, byte[] record)
      throws IOException, StorageException, BadRecordIdException
   {
      throw new NotImplementedException();
   }

   public synchronized byte[] get(int record_id)
      throws IOException, StorageException, BadRecordIdException
   {
      Node n = read(record_id);
      if (n.next == Storage.NULL) {
	 return n.data;
      }

      Accumulator acc = new Accumulator();
      acc.add(n.data);
      while (n.next != Storage.NULL) {
	 n = read(n.next);
	 acc.add(n.data);
      }
      return acc.sum();
   }

   public synchronized void get(int record_id, byte[] record)
      throws IOException, StorageException, BadRecordIdException
   {
      byte[] data = get(record_id);
      if (data.length > record.length) {
	 System.arraycopy(data, 0, record, 0, record.length);
      } else {
	 System.arraycopy(data, 0, record, 0, data.length);
      }
   }

   public synchronized byte[] remove(int record_id)
      throws IOException, StorageException, BadRecordIdException
   {
      Node n = read(record_id);
      m_store.release(record_id);
      if (n.next == Storage.NULL) {
	 return n.data;
      }

      Accumulator acc = new Accumulator();
      acc.add(n.data);
      m_store.release(record_id);
      while (n.next != Storage.NULL) {
	 record_id = n.next;
	 n = read(record_id);

	 acc.add(n.data);
	 m_store.release(record_id);
      }

      return acc.sum();
   }

   public synchronized void remove(int record_id, byte[] record)
      throws IOException, StorageException, BadRecordIdException
   {
      byte[] data = remove(record_id);
      System.arraycopy(data, 0, record, 0, Math.min(record.length, data.length));
   }

  private void verify()
    throws IOException
  {
    SuperVector tails = new SuperVector();
    SuperHashtable by_dst = new SuperHashtable();

    for (int record_id = m_store.nextRecord(NULL);
         record_id != NULL;
         record_id = m_store.nextRecord(record_id)) {
      Integer i = new Integer(record_id);
      Node n = read(record_id);
      if (n.next != Storage.NULL) {
        by_dst.put(new Integer(n.next), i);
      } else {
        tails.addElement(i);
      }
    }

    Enumeration en = tails.elements();
    while (en.hasMoreElements()) {
      Integer src = (Integer)en.nextElement();
      SuperVector v = new SuperVector();
      while (src != null) {
        v.addElement(src);
        src = (Integer)by_dst.get(src);
      }
      src = (Integer)v.elementAt(v.size() - 1);
      System.out.print("Blob nodes:");
      for (int i = v.size() - 1; i >= 0; --i) {
        Integer dst = (Integer)v.elementAt(i);
        System.out.print(" ");
        System.out.print(dst.intValue());
      }
      System.out.println();
    }
  }

  public static void main(String[] args)
    throws IOException
  {
    VariableStorage store = new VariableStorage(args[0]);
    store.verify();
  }

   private Storage m_store;
   private int m_length;
   private Sorter m_sorter;
}
